/* FILE: display.cc                 -*-Mode: c++-*-
 *
 * C++ code for display windows
 *
 */

#include "display.h"
#include "misc.h"
#include "constants.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

//////////////////////////////////////////////////////////////////////////
// Colormap class
const ClassDoc DisplayColorMapClass::class_doc("DisplayColorMapClass",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","15-Dec-1996");
DisplayColorMapClass::DisplayColorMapClass(void)
{
#define MEMBERNAME "DisplayColorMapClass"
  if(TKCOLORLENGTH!=6)
    FatalError(-1,STDDOC,"Current code only supports TKCOLORLENGTH==6");
  if(COLORMAPSIZE!=256)
    FatalError(-1,STDDOC,"Current code only supports COLORMAPSIZE==256");
  int i,r,g,b;
  for(i=0;i<COLORMAPSIZE;i++) {
    g=0;
    if(i<COLORMAPSIZE/2) { b=0; r=COLORMAPSIZE-2-2*i; }
    else             { r=0; b=2*(i-COLORMAPSIZE/2); }
    if(r>255) r=255; if(g>255) g=255; if(b>255) b=255; // Safety
    sprintf(TkColor[i],"#%02X%02X%02X",r,g,b);
  }
#undef MEMBERNAME
}

//////////////////////////////////////////////////////////////////////////
// CoordinateSystem helper function
INT4m CoordinateAngleDifference(CoordinateSystem cs1,CoordinateSystem cs2,
				double &angle)
{ // Calculates the difference (in degrees, from 0 to 360) of cs1 - cs2, if
  // both are CS_DISPLAY_* coordinates, sets the difference in 'angle' and
  // returns 0;  If cs1 is not a CS_DISPLAY_* c.s., then the return value
  // is 1; if cs2 is not a CS_DISPLAY_* c.s., then the return value is 2.
  // In these latter two cases angle is 0.0 on return.
  double ang1,ang2;
  angle=0.0;
  switch(cs1) {
  case CS_DISPLAY_STANDARD: ang1=  0; break;
  case CS_DISPLAY_ROT_90:   ang1= 90; break;
  case CS_DISPLAY_ROT_180:  ang1=180; break;
  case CS_DISPLAY_ROT_270:  ang1=270; break;
  default: return 1;  // Illegal coordinates
    // break;
  }
  switch(cs2) {
  case CS_DISPLAY_STANDARD: ang2=  0; break;
  case CS_DISPLAY_ROT_90:   ang2= 90; break;
  case CS_DISPLAY_ROT_180:  ang2=180; break;
  case CS_DISPLAY_ROT_270:  ang2=270; break;
  default: return 2;  // Illegal coordinates
    // break;
  }
  angle=ang1-ang2;
  if(angle<0) angle+=360;
  return 0;
}

//////////////////////////////////////////////////////////////////////////
// Display frame class
const ClassDoc DisplayFrame::class_doc("DisplayFrame",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","14-Dec-1996");
// Recommended display window width, height and cellsize, in pixels:
const REAL4m DisplayFrame::default_width=640;
const REAL4m DisplayFrame::default_height=480;
const CoordinateSystem DisplayFrame::default_coords=CS_DISPLAY_STANDARD;
const REAL4m DisplayFrame::default_cellsize=20;
const REAL4m DisplayFrame::default_arrow_ratio=0.75;
const REAL4m DisplayFrame::margin_ratio=1.0;

INT4m DisplayFrame::ChangeCoordinates(CoordinateSystem new_coords)
{
  if(display_coords==new_coords) return 0;
  ChangeDisplayBoxCoords(display_coords,new_coords);
  ChangeDisplayListCoords(display_coords,new_coords);
  ChangeBoundaryListCoords(display_coords,new_coords);
  display_coords=new_coords;
  return 0;
}

REAL4m DisplayFrame::GetSampleStepSize(void)
{
  REAL4m step_size=cellsize*mesh_per_pixel;
  REAL4m mesh_cellsize=mesh->GetApproximateCellWidth();
  if(mesh_cellsize>0.) step_size/=mesh_cellsize;
  return step_size;
}


INT4m DisplayFrame::SetMeshSampleSize(REAL4m step_request)
{ // Fills display_list from *mesh using requested step size.
  // Copies mesh_box to display_box
  // Use step_request=0 to get _all_ points in mesh.
  // NOTE: Changes display_coords to CS_CALCULATION_STANDARD.  It is
  //       the responsibility of the calling routine to save
  //       display_coords, if desired, _before_ calling this routine.
  // NOTE: Samples mesh in terms of *mesh units*, as opposed to
  //       *mesh cells* (cf. SetSampleStepSize in client interface).
  // NOTE: mesh_per_pixel must be set before calling this routine.

#define MEMBERNAME "SetMeshSampleSize"
  if(step_request<0)
    FatalError(-1,STDDOC,"%s: step_request=%f",ErrBadParam,step_request);
  if(mesh==(Mesh *)NULL) FatalError(-1,STDDOC,"Mesh pointer not set.");

  // Save current coordinate system
  CoordinateSystem orig_coords=display_coords;

  // Get mesh bounding box
  mesh->GetRange(mesh_box);

  // Determine stepping rates.  The code below is awkward; It is a
  // result of legacy code/design (or non-design, as the case may be).
  REAL4m mesh_cellsize=mesh->GetApproximateCellWidth();
  if(mesh_cellsize<=0.0) mesh_cellsize=1.0; // Safety
  Vec3f mesh_celldim=mesh->GetApproximateCellDimensions();
  if(mesh_celldim.x<=0.0) mesh_celldim.x=1.0; //Safety
  if(mesh_celldim.y<=0.0) mesh_celldim.y=1.0; //Safety

  // Get list of vectors to display
  REAL4m xstep=mesh_celldim.x*step_request/mesh_cellsize;
  REAL4m ystep=mesh_celldim.y*step_request/mesh_cellsize;
  REAL4m zstep=0.;
  mesh->GetDisplayList(xstep,ystep,zstep,mesh_box,display_list);
  xstep*=mesh_cellsize/mesh_celldim.x;
  ystep*=mesh_cellsize/mesh_celldim.y;

  // Setup display for new cellsize
  mesh_sample_size=MAX(xstep,ystep);
  step_request=mesh_sample_size;
  REAL4m grit=mesh->GetSubsampleGrit()*mesh_cellsize;
  if(0.<step_request && step_request<grit/2.) step_request=grit;

  arrowsize=MAX(step_request,mesh_cellsize);
  arrowsize*=default_arrow_ratio*arrowmag/mesh_per_pixel;

  cellsize=step_request/mesh_per_pixel;

  if(display_list.GetSize()<=0) margin=0.0;
  else {
    REAL4m arrow_space;
    arrow_space=MAX(step_request,mesh_cellsize)
      *MAX(1.0,default_arrow_ratio)*arrowmag;   // Mesh units
    margin=MAX(step_request,mesh_cellsize)*margin_ratio;
    margin=MAX(margin,arrow_space/2.);
    margin/=mesh_per_pixel;                     // Pixels
  }

  display_box=mesh_box;
  mesh->GetBoundaryList(boundary_list);
  display_coords=CS_CALCULATION_STANDARD; // reset coordinate system

  // Convert coordinates back to original coordinates
  ChangeCoordinates(orig_coords);

  return 0;
#undef MEMBERNAME
}

INT4m DisplayFrame::SetSampleStepSize(REAL4m step_request)
{ 
  // Fills display_list from *mesh using requested step size.
  // Copies mesh_box to display_box
  // Use step_request=0 to get _all_ points in mesh.
  // NOTE: Changes display_coords to CS_CALCULATION_STANDARD.  It is
  //       the responsibility of the calling routine to save
  //       display_coords, if desired, _before_ calling this routine.
  // NOTE: Samples mesh in terms of *mesh cells*, as opposed to
  // *mesh units* (cf. SetMeshSampleSize in internal interface).
#define MEMBERNAME "SetSampleStepSize"
  if(step_request<0)
    FatalError(-1,STDDOC,"%s: step_request=%f",ErrBadParam,step_request);
  if(mesh==(Mesh *)NULL) FatalError(-1,STDDOC,"Mesh pointer not set.");

  // Convert step_request from mesh cell size to mesh units.
  if(mesh->GetApproximateCellWidth()>0.)
    step_request*=mesh->GetApproximateCellWidth();

  return SetMeshSampleSize(step_request);
#undef MEMBERNAME
}

INT4m DisplayFrame::SetupFrame(REAL4m requested_width,
			       REAL4m requested_height,
			       CoordinateSystem requested_coords,
			       REAL4m step_request,
			       REAL4m new_arrowmag,
			       REAL4m new_scale)
{ // Adjusts mesh_per_pixel and mesh_sample_size as requested,
  // and fills display_list by calling SetMeshSampleSize() and
  // ChangeCoordinates(requested_coords).
  //   Notes: requested_width & requested_height are in pixels,
  // step_request is in mesh _cell_ units, new_arrowmag is a
  // value relative to the "standard" arrowsize, and new_scale
  // is in default_cellsize units per mesh cell unit.
  //   There are 4 cases, depending on whether each of step_request
  // and new_scale >0.  A value <=0 indicates a request for this
  // routine to calculate appropriate values.
  //   If new_scale<=0, then a scaling is determined that allows
  // the mesh+margin to just fit inside requested_width and
  // requested_height.  This is complicated by the fact that the
  // margin size depends upon step_request.
  //   If step_request<=0, then a value is computed which allows
  // for approximately default_cellsize pixels per display cell.
  // This, of course, depends upon the scaling.
  //   If neither new_scale nor step_request are positive, then
  // we calculate 2 sets of values for each, one where we assume
  // the scaling is restricted by requested_width, and the other
  // where we assume the scaling is restricted by requested_height.
  // Each set contains enough information to set both scaling
  // and step request, and then the smaller of the 2 scalings (and
  // the corresponding step request) is selected.
  //   NOTE: step_request has some special cases: step_request=0.
  // means "draw all points", and in general, if step_request<1,
  // then arrows are sized as though step_request==1.
#define MEMBERNAME "SetupFrame"
  if(mesh==(Mesh *)NULL) FatalError(-1,STDDOC,"Mesh pointer not set.");

  // Get mesh bounding box and boundary vertex list
  mesh->GetRange(mesh_box);
  Vec3f mesh_diag=mesh_box.GetDiag();

  // First handle 1 special case: When mesh_diag has a non-positive
  // dimension, there is no display, so just setup dummy values as
  // needed for step_request and new_scale.
  if(mesh_diag.x<=0. || mesh_diag.y<=0.) {
    if(step_request<0.) step_request=0.;
    if(new_scale<=0.)   new_scale=1.0;
    if(mesh_diag.x<=0.) mesh_diag.x=1.0;  // Safety
    if(mesh_diag.y<=0.) mesh_diag.y=1.0;  // Safety
  }

  // Calculations will be done in CS_CALCULATION_STANDARD
  // and CS_DISPLAY_STANDARD coordinates, so swap width
  // and height if the import values are for rotated coords.
  if(requested_coords==CS_DISPLAY_ROT_90 
     || requested_coords==CS_DISPLAY_ROT_270) {
    REAL4m temp=requested_width;
    requested_width=requested_height;
    requested_height=temp;
  }

  REAL4m mesh_per_cell=mesh->GetApproximateCellWidth();
  if(mesh_per_cell<=0.) mesh_per_cell=1.0;  // Safety
  mesh_sample_size=mesh_per_cell;

  // Convert input values to "sensible" units (pixels & mesh units)!
  if(step_request>0) step_request*=mesh_per_cell;
  /// step_request is now in mesh units.

  //  if(new_scale>0) new_scale*=(default_cellsize/mesh_per_cell);
  if(new_scale>0) new_scale/=mesh_per_cell;
  /// new_scale is now in pixels per mesh unit

  // Set arrowmag
  if(new_arrowmag>=0 && new_arrowmag<1/EPSILON) arrowmag=new_arrowmag;
  else                                          arrowmag=1.0;  // Default

  // Handle automatic valuation requests
  if(step_request<0. && new_scale>0.) {
    // Compute a value which allows for approximately default_cellsize
    // pixels per display cell. This, of course, depends upon the
    // scaling.
    step_request=default_cellsize/new_scale;
    step_request=mesh->GetPreferredStepSize(step_request);
    if(step_request<=0.) step_request=mesh_per_cell;
  }
  else if (step_request>=0. && new_scale<=0.) {
    // The margin size can be calculated in mesh units, so we
    // can directly calculate appropriate x & y scaling factors
    // to allow mesh+margin to _just_ fit inside the requested
    // region.
    REAL4m arrow_space,margin_work;  // In mesh units
    arrow_space=MAX(1.0,default_arrow_ratio)
      *MAX(step_request,mesh_per_cell)*arrowmag;
    /// Note that arrow size doesn't scale down below mesh_per_cell.

    margin_work=MAX(mesh_per_cell,step_request)*margin_ratio;
    margin_work=MAX(margin_work,arrow_space/2.);

    REAL4m xscale=requested_width/(2*margin_work+mesh_diag.x);
    REAL4m yscale=requested_height/(2*margin_work+mesh_diag.y);
    new_scale=MIN(xscale,yscale);
  }
  else if (step_request<0. && new_scale<=0.) {
    // Last case: calculate both step_request and new_scale.
    // We calculate 2 sets of values for each, one where we assume
    // the scaling is restricted by requested_width, and the other
    // where we assume the scaling is restricted by requested_height.
    // Each set contains enough information to set both scaling
    // and step request, and then the smaller of the 2 scalings (and
    // the corresponding step request) is selected.
    //   We also want to adjust step_request to meet
    //     step_request=mesh->GetPreferredStepSize(step_request);

    // Guess at margin size relative to step_request
    REAL4m margin_adjust=MAX(margin_ratio,
			     MAX(1.0,default_arrow_ratio)*arrowmag/2.);

    REAL4m work_width,work_height;
    work_width=(requested_width/default_cellsize)-2*margin_adjust;
    work_height=(requested_height/default_cellsize)-2*margin_adjust;

    // If either work_width or work_height is <=0, then there is
    // not enough room along that dimension to display margins at
    // default_cellsize*margin_adjust.  In this case, we'll just punt,
    // and *pretend* requested_width is larger than it really is...
    if(work_width<=0.) work_width=1.;
    if(work_height<=0.) work_height=1.;

    // Take first pass at calculating step_request's:
    REAL4m stepx=mesh_diag.x/work_width;
    REAL4m stepy=mesh_diag.y/work_height;

    // Adjust step_request's to a preferred value.
    stepx=mesh->GetPreferredStepSize(stepx);
    stepy=mesh->GetPreferredStepSize(stepy);
    if(stepx<=0.) stepx=mesh_per_cell;
    if(stepy<=0.) stepy=mesh_per_cell;

    // Calculate margins.
    REAL4m arrow_spacex,arrow_spacey;  // In mesh units
    arrow_spacex=MAX(1.0,default_arrow_ratio)
      *MAX(stepx,mesh_per_cell)*arrowmag;
    arrow_spacey=MAX(1.0,default_arrow_ratio)
      *MAX(stepy,mesh_per_cell)*arrowmag;

    REAL4m marginx=MAX(stepx*margin_ratio,arrow_spacex/2.);
    REAL4m marginy=MAX(stepy*margin_ratio,arrow_spacey/2.);

    // Calculate scalings (pixels per mesh)
    REAL4m xscale=requested_width/(mesh_diag.x+2*marginx);
    REAL4m yscale=requested_height/(mesh_diag.y+2*marginy);

    // Decide!
    if(xscale<=yscale) {
      new_scale=xscale;
      step_request=stepx;
    }
    else {
      new_scale=yscale;
      step_request=stepy;
    }
  }
  if(new_scale<EPSILON*EPSILON)     // Safety
    new_scale=EPSILON*EPSILON;
  else if(new_scale>1./(EPSILON*EPSILON))
    new_scale=1./(EPSILON*EPSILON);
  

  // Okay!  At this point, step_request and new_scale are defined,
  // and are in mesh units and pixels per mesh unit, respectively.

  // Calculate mesh_per_pixel
  mesh_per_pixel=1./new_scale;

  // Fill display_list
  display_coords=CS_CALCULATION_STANDARD;
  SetMeshSampleSize(step_request);
  ChangeCoordinates(requested_coords);
  return 0;
#undef MEMBERNAME
}

void DisplayFrame::ConvertMeshPtToDisplayPt(Vec3f &point,
					    CoordinateSystem new_coords)
{ // This routine assumes point is using the CS_CALCULATION_STANDARD
  // coordinate system, and display is using CS_DISPLAY_* coordinate
  // system specified by *this->display_coords.
  // NOTE: This routine assumes that display_box is already converted to
  //       "new_coords"
#define MEMBERNAME "ConvertMeshPtToDisplayPt"
#if DEBUGLVL > 9
  if(mesh_per_pixel<=0.0)
    FatalError(-1,STDDOC,"mesh_per_pixel (=%g) not initialized.",
	       mesh_per_pixel);
#endif
  // First convert to CS_DISPLAY_STANDARD coordinate system
  Vec3f mesh_min,display_min,display_max,vtemp;
  mesh_box.GetMinPt(mesh_min);
  display_box.GetExtremes(display_min,display_max);
  point-=mesh_min;
  point*=(1.0/mesh_per_pixel);
  vtemp.x=ROUND(point.x)+REAL4m(margin)+display_min.x;
  vtemp.y=display_max.y-ROUND(point.y)-REAL4m(margin);
  vtemp.z=display_min.z+ROUND(point.z);

  // Next convert from CS_DISPLAY_STANDARD to new_coords
  point.z=vtemp.z; // This axis is fixed
  switch(new_coords) {
  case CS_DISPLAY_STANDARD:
    point.x=vtemp.x; // Nothing to do.
    point.y=vtemp.y;
    break;
  case CS_DISPLAY_ROT_90:
    point.x=vtemp.y;
    point.y=display_max.y-vtemp.x;
    break;
  case CS_DISPLAY_ROT_180:
    point.x=display_max.x-vtemp.x;
    point.y=display_max.y-vtemp.y;
    break;
  case CS_DISPLAY_ROT_270:
    point.x=display_max.x-vtemp.y;
    point.y=vtemp.x;
    break;
  default:
    FatalError(-1,STDDOC,"%s: Illegal new_coords (%d)",
	       ErrProgramming,new_coords);
    break;
  }
#undef MEMBERNAME
}

void DisplayFrame::ChangeDisplayBoxCoords(CoordinateSystem old_coords,
					  CoordinateSystem new_coords)
{ // Changes display_box coordinates
#define MEMBERNAME "ChangeDisplayListCoords"
  if(old_coords==new_coords) return;
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    if(mesh_per_pixel<=0.0)
      FatalError(-1,STDDOC,"mesh_per_pixel (=%g) not initialized.",
		 mesh_per_pixel);
    Vec3f disp_diag=(1./mesh_per_pixel)*display_box.GetDiag();
    // Introduce margins
    disp_diag+=Vec3f(REAL4m(2*margin),REAL4m(2*margin),0.);
    display_box.Set(Vec3f(0.,0.,0.),disp_diag);
    // NOTE: display_box is zero-based (as opposed to, e.g., "-margin"-based).
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    ChangeDisplayBoxCoords(old_coords,CS_DISPLAY_STANDARD);
    ChangeDisplayBoxCoords(CS_DISPLAY_STANDARD,new_coords);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    double angle_diff;
    if(CoordinateAngleDifference(new_coords,old_coords,angle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    if(angle_diff==90 || angle_diff==270) { // Swap x & y display ranges
      REAL4m temp;
      Vec3f display_min,display_max;
      display_box.GetExtremes(display_min,display_max);
      temp=display_max.x;
      display_max.x=display_max.y;  display_max.y=temp;
      temp=display_min.x;
      display_min.x=display_min.y;      display_min.y=temp;
      // z coordinate is not changed by these rotations
      display_box.Set(display_min,display_max);
    }
  }
#undef MEMBERNAME
}

void DisplayFrame::ChangeDisplayListCoords(CoordinateSystem old_coords,
					   CoordinateSystem new_coords)
{
  // NOTE: You must call ChangeDisplayBoxCoords() *before* calling this
  // routine, as otherwise the display_box won't be set correcty.

#define MEMBERNAME "ChangeDisplayListCoords"
  if(old_coords==new_coords) return;
  LocatedVector *display_vector;
  Vec3f vtemp;
  Vec3f display_min,display_max;
  display_box.GetExtremes(display_min,display_max);
  // Code below may assume display_min == (0,0,0)

  // Branch through cases
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    if(mesh_per_pixel<=0.0)
      FatalError(-1,STDDOC,"mesh_per_pixel = %g <= 0.0",mesh_per_pixel);
    REAL4m pixel_per_mesh=1.0/mesh_per_pixel;
    Vec3f mesh_base;
    mesh_box.GetExtremes(mesh_base,vtemp);
    Vec3f left_offset,right_offset;
    // Make margins
    left_offset=display_min-Vec3f(REAL4m(margin),REAL4m(margin),0.);
    right_offset=display_max-Vec3f(REAL4m(margin),REAL4m(margin),0.);
    for(display_vector=display_list.GetFirst();
	display_vector!=(LocatedVector *)NULL;
	display_vector=display_list.GetNext()) {
      // The "->location" section of this could be implemented
      // by a call to ConvertMeshPtToDisplayPt, but this was written
      // before that routine was, and this should be faster for
      // a long list.
      vtemp=display_vector->location;
      vtemp-=mesh_base;  vtemp*=pixel_per_mesh;
      display_vector->location.x = vtemp.x-left_offset.x;
      display_vector->location.y = right_offset.y-vtemp.y;
      display_vector->location.z = right_offset.z-vtemp.z;
      display_vector->value.y*=-1;
    }
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    if(Verbosity>=100) {
      NonFatalError(STDDOC,
	    "Requested coordinate transformation (%d->%d) not yet coded.\n"
	    "  Using 2-step transform (through CS_DISPLAY_STANDARD) instead.",
	    int(old_coords),int(new_coords));
    }
    ChangeDisplayBoxCoords(new_coords,CS_DISPLAY_STANDARD);
    ChangeDisplayListCoords(old_coords,CS_DISPLAY_STANDARD);
    ChangeDisplayBoxCoords(CS_DISPLAY_STANDARD,new_coords);
    ChangeDisplayListCoords(CS_DISPLAY_STANDARD,new_coords);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    double dangle_diff;
    if(CoordinateAngleDifference(new_coords,old_coords,dangle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    INT4m angle_diff=INT4m(ROUND(dangle_diff));
    for(display_vector=display_list.GetFirst();
	display_vector!=(LocatedVector *)NULL;
	display_vector=display_list.GetNext()) {
      Vec3f *loc=&(display_vector->location);
      Vec3f *dir=&(display_vector->value);
      switch(angle_diff) {
      case 90:
	vtemp = *loc;
	loc->x = vtemp.y;
	loc->y = display_max.y - vtemp.x;
	vtemp = *dir;
	dir->x =  vtemp.y;
	dir->y = -vtemp.x;
	break;
      case 180:
	loc->x = display_max.x - loc->x;
	loc->y = display_max.y - loc->y;
	dir->x *= -1;
	dir->y *= -1;
	break;
      case 270:
	vtemp = *loc;
	loc->x = display_max.x - vtemp.y;
	loc->y = vtemp.x;
	vtemp = *dir;
	dir->x = -vtemp.y;
	dir->y =  vtemp.x;
	break;
      default:
	FatalError(-1,STDDOC,"%s: Illegal angle_diff (%d)",
		   ErrProgramming,angle_diff);
	break;
      }
    }
  }
#undef MEMBERNAME
}

void DisplayFrame::ChangeBoundaryListCoords(CoordinateSystem old_coords,
					    CoordinateSystem new_coords)
{ // Convert boundary vertex list
  // NOTE: You must call ChangeDisplayBoxCoords() *before* calling this
  // routine, as otherwise the display_box won't be set correcty.
#define MEMBERNAME "ChangeBoundaryListCoords"
  if(old_coords==new_coords) return;
  Vec3f *bdry;  // Pointer to boundary_list vertices
  // Branch through cases
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    for(bdry=boundary_list.GetFirst(); bdry!=(Vec3f *)NULL;
	bdry=boundary_list.GetNext()) {
      ConvertMeshPtToDisplayPt(*bdry,new_coords);
         // This is presumably a short list
    }
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    // Do 2-step change through CS_DISPLAY_STANDARD
    ChangeDisplayBoxCoords(new_coords,CS_DISPLAY_STANDARD);
    ChangeBoundaryListCoords(old_coords,CS_DISPLAY_STANDARD);
    ChangeDisplayBoxCoords(CS_DISPLAY_STANDARD,new_coords);
    ChangeBoundaryListCoords(CS_DISPLAY_STANDARD,new_coords);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    double dangle_diff;
    if(CoordinateAngleDifference(new_coords,old_coords,dangle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    INT4m angle_diff=INT4m(ROUND(dangle_diff));
    Vec3f vtemp,display_min,display_max;
    display_box.GetExtremes(display_min,display_max);
    for(bdry=boundary_list.GetFirst(); bdry!=(Vec3f *)NULL;
	bdry=boundary_list.GetNext()) {
      switch(angle_diff) { // Note: bdry->z doesn't change
      case 90:
	vtemp = *bdry;
	bdry->x = vtemp.y;
	bdry->y = display_max.y - vtemp.x;
	break;
      case 180:
	bdry->x = display_max.x - bdry->x;
	bdry->y = display_max.y - bdry->y;
	break;
      case 270:
	vtemp = *bdry;
	bdry->x = display_max.x - vtemp.y;
	bdry->y = vtemp.x;
	break;
      default:
	FatalError(-1,STDDOC,"%s: Illegal angle_diff (%d) (2nd switch())",
		   ErrProgramming,angle_diff);
	break;
      }
    }
  }
#undef MEMBERNAME
}

INT4m DisplayFrame::DrawArrow(const char *canvas_name,
			      const LocatedVector &locvec)
{
#define MEMBERNAME "DrawArrow"
  const Vec3f *postemp;
  const Vec3f *vtemp;
  REAL4m x1,y1,x2,y2,delx,dely,vec_mag,arrow_len,proj_len;
  INT4m r0,c0,r1,c1,r2,c2,ioff,a1,a2,a3;
  INT4m color;
  CHAR *pcolor1,*pcolor2;

  postemp=&(locvec.location);
  vtemp=&(locvec.value);

  // Line components
  delx=vtemp->x*arrowsize;    dely=vtemp->y*arrowsize;
  vec_mag=vtemp->Mag();
  arrow_len=vec_mag*arrowsize;
  if(arrow_len<SQRT_EPSILON) return TCL_OK;  // This is a zero vector
  proj_len=sqrt(delx*delx+dely*dely);

  // Arrowhead components
  a1=(int)ROUND(proj_len*0.3);
  a2=(int)ROUND(proj_len*0.4);
  a3=(int)ROUND(arrow_len*0.15);
  if(a3<5) a1=a2;

  // Attributes
  int linewidth=(int)ROUND(a3*0.6);
  if(linewidth<1) linewidth=1;
  color=(int)ROUND(COLORMAPSIZE*((((vtemp->z)/vec_mag)+1.0)/2.));
  if(color<0)                  color=0;
  else if(color>=COLORMAPSIZE) color=COLORMAPSIZE-1;
  pcolor1=colormap.TkColor[color];

  // NOTE: Timing tests on Linux 2.0.24 on a i486-DX2/66 indicate that
  //       a substantial time loss is involved inside sprintf doing
  //       %f conversions.  It is 6 or 7 times faster to use %d's instead
  //      of sprintf(scratch_buf,"%s create line %6.2f %6.2f %6.2f %6.2f"
  //	             "-arrow last -fill %s",canvas_name,x1,y1,x2,y2,
  //                 colormap.TkColor[color]);

  if(proj_len>=arrow_len*0.15) { // Standard arrow
    x1=(postemp->x)-delx/2;   x2=x1+delx;
    r1=int(ROUND(x1));     r2=int(ROUND(x2));
    y1=(postemp->y)-dely/2;   y2=y1+dely;
    c1=int(ROUND(y1));     c2=int(ROUND(y2));
    sprintf(scratch_buf,
	    "%s create line %d %d %d %d -arrow last -arrowshape {%d %d %d}"
	    "  -width %d -fill %s",canvas_name,r1,c1,r2,c2,a1,a2,a3,
	    linewidth,pcolor1);
  }
  else { // Small projection with undefined direction.  Draw a dot
    pcolor2=colormap.TkColor[COLORMAPSIZE/2];
    r0=(int)ROUND(postemp->x); c0=(int)ROUND(postemp->y);
    if(vtemp->z<0) { // Arrow pointing into-plane
      ioff=(int)ROUND(REAL4m(a3)/2.);
      r1=r0-ioff;  c1=c0-ioff;
      r2=r0+ioff;  c2=c0+ioff;
      sprintf(scratch_buf,"%s create rect %d %d %d %d -fill %s -outline %s",
	      canvas_name,r1,c1,r2,c2,pcolor1,pcolor1);
      // Note: The return quantity of sprintf is OS dependent.
      if(a3>6) {
	char *bufptr=strchr(scratch_buf,'\0');
	ioff=(int)ROUND(REAL4m(a3)/6.);
	r1=r0-ioff;  c1=c0-ioff;
	r2=r0+ioff;  c2=c0+ioff;
	sprintf(bufptr,"; %s create rect %d %d %d %d -fill %s -width 0",
		canvas_name,r1,c1,r2,c2,pcolor2);
      }
    } else { // Arrow pointing out-of-plane
      ioff=(int)ROUND(REAL4m(a3)*(SQRT2/2.));
      r1=r0-ioff;  c1=c0-ioff;
      r2=r0+ioff;  c2=c0+ioff;
      sprintf(scratch_buf,
	      "%s create poly %d %d %d %d %d %d %d %d"
	      " -fill %s -outline %s -width %d",
	      canvas_name,r1,c0,r0,c2,r2,c0,r0,c1,
	      pcolor1,pcolor1,linewidth/3);
      // Note: The return quantity of sprintf is OS dependent.
      if(a3>6) {
	char *bufptr=strchr(scratch_buf,'\0');
	sprintf(bufptr,"; %s create line %d %d %d %d -fill %s -width %d"
		"; %s create line %d %d %d %d -fill %s -width %d",
		canvas_name,r1,c0,r2,c0,pcolor2,linewidth/3,
		canvas_name,r0,c2,r0,c1,pcolor2,linewidth/3);
      }
    }
  }
  if(Tcl_Eval(tcl_interp,scratch_buf) != TCL_OK) return TCL_ERROR;
  return TCL_OK;
#undef MEMBERNAME
}

#ifdef TCL_OBJ_TEST_CODE
INT4m DisplayFrame::DrawArrowObj(const char *canvas_name,
				 const LocatedVector &locvec)
{
#define MEMBERNAME "DrawArrowObj"
  const Vec3f *postemp;
  const Vec3f *vtemp;
  double x1,y1,x2,y2,delx,dely,vec_mag,arrow_len,proj_len;
  double ioff,a1,a2,a3;
  INT4m color;
  CHAR *pcolor1,*pcolor2;

#ifdef SILLY
fprintf(stderr,"Pre-check\n"); fflush(stderr);
double sillyval;
Tcl_Obj* sillyobj;
sillyobj=Tcl_NewDoubleObj(7.42);
Tcl_IncrRefCount(sillyobj);
Tcl_GetDoubleFromObj(tcl_interp,sillyobj,&sillyval);
fprintf(stderr,"Silly value=%lf, refcount=%d, shared=%d\n",
	sillyval,sillyobj->refCount,Tcl_IsShared(sillyobj));
Tcl_DecrRefCount(sillyobj);
fprintf(stderr,"Post-check\n"); fflush(stderr);
#endif // SILLY

  postemp=&(locvec.location);
  vtemp=&(locvec.value);

  // Line components
  delx=vtemp->x*arrowsize;    dely=vtemp->y*arrowsize;
  vec_mag=vtemp->Mag();
  arrow_len=vec_mag*arrowsize;
  if(arrow_len<SQRT_EPSILON) return TCL_OK;  // This is a zero vector
  proj_len=sqrt(delx*delx+dely*dely);

  // Arrowhead components
  a1=proj_len*0.3;
  a2=proj_len*0.4;
  a3=arrow_len*0.15;
  if(a3<5) a1=a2;

  // Attributes
  double linewidth=a3*0.6;
  if(linewidth<1.) linewidth=1.;
  color=(int)ROUND(COLORMAPSIZE*((((vtemp->z)/vec_mag)+1.0)/2.));
  if(color<0)                  color=0;
  else if(color>=COLORMAPSIZE) color=COLORMAPSIZE-1;
  pcolor1=colormap.TkColor[color];

  // NOTE: Timing tests on Linux 2.0.24 on a i486-DX2/66 indicate that
  //       a substantial time loss is involved inside sprintf doing
  //       %f conversions.  It is 6 or 7 times faster to use %d's instead
  //      of sprintf(scratch_buf,"%s create line %6.2f %6.2f %6.2f %6.2f"
  //	             "-arrow last -fill %s",canvas_name,x1,y1,x2,y2,
  //                 colormap.TkColor[color]);

  if(proj_len>=arrow_len*0.15) { // Standard arrow
    x1=(postemp->x)-delx/2;   x2=x1+delx;
    y1=(postemp->y)-dely/2;   y2=y1+dely;

#ifdef OLD_CODE
    sprintf(scratch_buf,
	    "%s create line %f %f %f %f -arrow last -arrowshape {%f %f %f}"
	    "  -width %f -fill %s",canvas_name,x1,y1,x2,y2,a1,a2,a3,
	    linewidth,pcolor1);
#else
    static Tcl_Obj *parrowshape[3];
    static Tcl_Obj *pcmd[15],*cmd;

    parrowshape[0]=Tcl_NewDoubleObj(a1);
    parrowshape[1]=Tcl_NewDoubleObj(a2);
    parrowshape[2]=Tcl_NewDoubleObj(a3);
    for(int i=0;i<=2;i++) Tcl_IncrRefCount(parrowshape[i]);

    strcpy(scratch_buf,canvas_name);
    pcmd[0]=Tcl_NewStringObj(scratch_buf,-1);
    pcmd[1]=Tcl_NewStringObj("create",6);
    pcmd[2]=Tcl_NewStringObj("line",4);
    pcmd[3]=Tcl_NewDoubleObj(x1);
    pcmd[4]=Tcl_NewDoubleObj(y1);
    pcmd[5]=Tcl_NewDoubleObj(x2);
    pcmd[6]=Tcl_NewDoubleObj(y2);
    pcmd[7]=Tcl_NewStringObj("-arrow",6);
    pcmd[8]=Tcl_NewStringObj("last",4);
    pcmd[9]=Tcl_NewStringObj("-arrowshape",11);
    pcmd[10]=Tcl_NewListObj(3,parrowshape);
    pcmd[11]=Tcl_NewStringObj("-width",6);
    pcmd[12]=Tcl_NewDoubleObj(linewidth);
    pcmd[13]=Tcl_NewStringObj("-fill",5);
    pcmd[14]=Tcl_NewStringObj(pcolor1,-1);
    for(int i=0;i<=14;i++) Tcl_IncrRefCount(pcmd[i]);
    cmd=Tcl_NewListObj(15,pcmd);
    Tcl_IncrRefCount(cmd);

fprintf(stderr,"RefA count pcmd[1]=%d\n",pcmd[1]->refCount); fflush(stderr);
    int eval_results=Tcl_EvalObj(tcl_interp,cmd);
fprintf(stderr,"RefB count pcmd[1]=%d\n",pcmd[1]->refCount); fflush(stderr);
    Tcl_DecrRefCount(cmd);
fprintf(stderr,"RefC count pcmd[1]=%d\n",pcmd[1]->refCount); fflush(stderr);

    for(int i=0;i<=14;i++) Tcl_IncrRefCount(pcmd[i]);

    if(eval_results != TCL_OK) return TCL_ERROR;
    return TCL_OK;
    
#endif // OLD_CODE
  }
  else { // Small projection with undefined direction.  Draw a dot
    pcolor2=colormap.TkColor[COLORMAPSIZE/2];
    double x0=postemp->x;
    double y0=postemp->y;
    if(vtemp->z<0) { // Arrow pointing into-plane
      ioff=a3/2.;
      x1=x0-ioff;  y1=y0-ioff;
      x2=x0+ioff;  y2=y0+ioff;
      sprintf(scratch_buf,"%s create rect %f %f %f %f -fill %s -outline %s",
	      canvas_name,x1,y1,x2,y2,pcolor1,pcolor1);
      // Note: The return quantity of sprintf is OS dependent.
      if(a3>6.) {
	char *bufptr=strchr(scratch_buf,'\0');
	ioff=a3/6.;
	x1=x0-ioff;  y1=y0-ioff;
	x2=x0+ioff;  y2=y0+ioff;
	sprintf(bufptr,"; %s create rect %f %f %f %f -fill %s -width 0",
		canvas_name,x1,y1,x2,y2,pcolor2);
      }
    } else { // Arrow pointing out-of-plane
      ioff=a3*(SQRT2/2.);
      x1=x0-ioff;  y1=y0-ioff;
      x2=x0+ioff;  y2=y0+ioff;
      sprintf(scratch_buf,
	      "%s create poly %f %f %f %f %f %f %f %f"
	      " -fill %s -outline %s -width %f",
	      canvas_name,x1,y0,x0,y2,x2,y0,x0,y1,
	      pcolor1,pcolor1,linewidth/3.);
      // Note: The return quantity of sprintf is OS dependent.
      if(a3>6.) {
	char *bufptr=strchr(scratch_buf,'\0');
	sprintf(bufptr,"; %s create line %f %f %f %f -fill %s -width %f"
		"; %s create line %f %f %f %f -fill %s -width %f",
		canvas_name,x1,y0,x2,y0,pcolor2,linewidth/3.,
		canvas_name,x0,y2,x0,y1,pcolor2,linewidth/3.);
      }
    }
  }
  if(Tcl_Eval(tcl_interp,scratch_buf) != TCL_OK) return TCL_ERROR;
  return TCL_OK;
#undef MEMBERNAME
}
#endif // TCL_OBJ_TEST_CODE

INT4m DisplayFrame::DrawLine(const char *canvas_name,
			     const Vec3f &Pt1,const Vec3f &Pt2)
{
  sprintf(scratch_buf,
	  "%s create line %6.2f %6.2f %6.2f %6.2f",
	  canvas_name,Pt1.x,Pt1.y,Pt2.x,Pt2.y);
  if(Tcl_Eval(tcl_interp,scratch_buf) != TCL_OK) return TCL_ERROR;
  return TCL_OK;
}

INT4m DisplayFrame::Show(const char *canvas_name)
{
#define MEMBERNAME "Show"
  // Initialize canvas
  if(tcl_interp==NULL)
    FatalError(-1,STDDOC,"tcl_interp not initialized.");
  if(mesh==NULL)
    FatalError(-1,STDDOC,"mesh not initialized.");

  // Check that the display_list is in a displayable coordinate system
  if(display_coords!=CS_DISPLAY_STANDARD &&
     display_coords!=CS_DISPLAY_ROT_90   &&
     display_coords!=CS_DISPLAY_ROT_180  &&
     display_coords!=CS_DISPLAY_ROT_270    ) {
    FatalError(-1,STDDOC,
	       "Can't draw display list in current (%d) coordinate system",
	       display_coords);
  }

  // Draw arrows
  LocatedVector *display_vector;
  for(display_vector=display_list.GetFirst();
      display_vector!=(LocatedVector *)NULL;
      display_vector=display_list.GetNext()) {
    if(DrawArrow(canvas_name,*display_vector) != TCL_OK)
      return TCL_ERROR;
  }

  // Draw mesh edge box
  Vec3f *bdry1,*bdry2;
  if((bdry1=boundary_list.GetFirst())!=(Vec3f *)NULL) {
    while((bdry2=boundary_list.GetNext())!=(Vec3f *)NULL) {
      if(DrawLine(canvas_name,*bdry1,*bdry2) != TCL_OK) return TCL_ERROR;
      bdry1=bdry2;
    }
  }

  return TCL_OK;
#undef MEMBERNAME
}


//   SetDisplayBox changes display_box to bb, reducing extents
// as necessary to keep the current aspect ratio.  If any of the
// extents are negative, then those dimensions will be scaled by
// the others.
//   The new values for display_box are copied into bb upon return.
INT4m DisplayFrame::SetDisplayBox(BoundingBox& bb)
{
  REAL4m curwidth =display_box.GetWidth();
  REAL4m curheight=display_box.GetHeight();
  REAL4m curdepth =display_box.GetDepth();
  REAL4m newwidth =ROUND(bb.GetWidth());
  REAL4m newheight=ROUND(bb.GetHeight());
  REAL4m newdepth =ROUND(bb.GetDepth());

  REAL8m xscale,yscale,zscale; // 8-byte float needed to store BIG_VALUE
  if(curwidth>0.  && newwidth>0.)   xscale=newwidth/curwidth;
  else                              xscale=BIG_VALUE;
  if(curheight>0. && newheight>0.)  yscale=newheight/curheight;
  else                              yscale=BIG_VALUE;
  if(curdepth>0.  && newdepth>0.)   zscale=newdepth/curdepth;
  else                              zscale=BIG_VALUE;

  REAL8m scale=0.0;
  if(xscale<yscale) {
    if(xscale<zscale) scale=xscale;
    else              scale=zscale;
  } else {
    if(yscale<zscale) scale=yscale;
    else              scale=zscale;
  }
  if(scale<=0. || scale>=BIG_VALUE || mesh_per_pixel/scale<=0.) {
    // Illegal invalid; return unchanged
    bb=display_box;
    return 1;
  }

  // Rescale mesh-to-pixel ratio & reload display list
  // (which automatically updates display_box, margin,
  // arrowsize and cellsize).
  mesh_per_pixel/=scale;
  SetMeshSampleSize(mesh_sample_size);
  bb=display_box;
  return 0;
}

// NOTE: "scale" in the next 2 routines is in units of "pixels per
//       approx cell size".
REAL4m DisplayFrame::GetScale(void) const
{
  REAL4m mesh_per_cell=mesh->GetApproximateCellWidth();
  if(mesh_per_cell<=0.) mesh_per_cell=1.0;  // Safety
  return mesh_per_cell/mesh_per_pixel;
}

INT4m DisplayFrame::SetScale(REAL4m newscale)
{
  if(newscale<EPSILON || newscale>1/EPSILON) return 1; // No change
  REAL4m mesh_per_cell=mesh->GetApproximateCellWidth();
  if(mesh_per_cell<=0.) mesh_per_cell=1.0;  // Safety
  mesh_per_pixel=mesh_per_cell/newscale;
  // Reload display list, which automatically updates display_box,
  // margin, arrowsize and cellsize.
  SetMeshSampleSize(mesh_sample_size);
  return 0;
}

INT4m DisplayFrame::SetArrowMag(REAL4m _arrowmag)
{
  if(_arrowmag<0. || arrowmag>1./EPSILON) return 1; // No change
  arrowmag=_arrowmag;
  // Reload display list, which automatically updates display_box,
  // margin, arrowsize and cellsize.
  SetMeshSampleSize(mesh_sample_size);
  return 0;
}
