/* FILE: mesh.cc                -*-Mode: c++-*-
 *
 * Abstract mesh class and children
 *
 */

#include <stdio.h>

#include "fileio.h"
#include "mesh.h"
#include "vector.h"
#include "constants.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building


//////////////////////////////////////////////////////////////////////////
// Abstract Mesh class
void Mesh::SetFilename(const char* _Filename) {
  if(Filename!=NULL) delete[] Filename;
  if(_Filename==NULL) Filename=NULL;
  else {
    Filename=new char[strlen(_Filename)+1];
    strcpy(Filename,_Filename);
  }
}
void Mesh::SetTitle(const char* _Title) {
  if(Title!=NULL) delete[] Title;
  if(_Title==NULL) Title=NULL;
  else {
    Title=new char[strlen(_Title)+1];
    strcpy(Title,_Title);
  }
}
void Mesh::SetDescription(const char* _Description) {
  if(Description!=NULL) delete[] Description;
  if(_Description==NULL) Description=NULL;
  else {
    Description=new char[strlen(_Description)+1];
    strcpy(Description,_Description);
  }
}
const char* Mesh::GetName() const { 
  if(Filename==NULL) return "";
  return Filename; 
}
const char* Mesh::GetTitle() const {
  if(Title==NULL) return "";
  return Title; 
}
const char* Mesh::GetDescription() const {
  if(Description==NULL) return "";
  return Description; 
}

//////////////////////////////////////////////////////////////////////////
// Regular 3D grid of Vec3f's
const ClassDoc GridVec3f::class_doc("GridVec3f",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","11-Dec-1996");

// The following constructor should be called from file input functions.
GridVec3f::GridVec3f(const char *filename,
		     const char* title,const char* desc,
		     INT4m newi,INT4m newj,INT4m newk,
		     const Vec3f &basept,const Vec3f &gridstep,
		     const BoundingBox &newrange,
		     const List<Vec3f> &newboundary)
  :Mesh("GridVec3f",filename,title,desc)
{
  SetSize(newi,newj,newk);
  coords_base=basept;
  coords_step=gridstep;
  if(coords_step.x==0.) coords_step.x=1.;
  if(coords_step.y==0.) coords_step.y=1.;
  if(coords_step.z==0.) coords_step.z=1.;
  range=newrange;
  boundary=newboundary;
}

// Problem <-> Nodal Coordinates Conversion Routines
// The conversion from nodes to problem coordinates is
//    (i,j,k) -> base+(i*step.x,j*step.y,k*step.z).
// NOTE 1: The grid vector access functions, e.g., operator(), access
//   via the integral "nodal" coordinates, *not* in terms of the
//   (real valued) problem coordinates.
// NOTE 2: The values in coords_step are guaranteed to be !=0,
//   but may be negative!  However, GetStdStep() (below), returns
//   a value >0.
// NOTE 3: The coords conversion routines don't preserve order
//   if step.? is negative, e.g., say (min.x,min.y) < (max.x,max.y)
//   going in, but step.x is negative.  Then after conversion,
//   min.y~ will still be < max.y~, but now min.x~>max.x~.

Vec3f GridVec3f::NodalToProblemCoords(Vec3f const &w_nodal) const
{
  Vec3f v(w_nodal);
  v.x*=coords_step.x;
  v.y*=coords_step.y;
  v.z*=coords_step.z;
  v+=coords_base;
  return v;
}

Vec3f GridVec3f::ProblemToNodalCoords(Vec3f const &v_problem) const
{
  Vec3f w(v_problem);
  w-=coords_base;
  w.x/=coords_step.x;
  w.y/=coords_step.y;
  w.z/=coords_step.z;
  return w;
}


INT4m GridVec3f::GetDisplayList(REAL4m &xstep_request,
				REAL4m &ystep_request,
				REAL4m &zstep_request,
				const BoundingBox &dsprange,
				List<LocatedVector> &display_list)
{ // Fills display_list with points from mesh inside "dsprange", with
  // step size close to ?step_request, which is reset to value actually
  // used.
  //   Use ?step_request=0 to remove constraint on that axis, e.g.,
  // if xstep_request=ystep_request=zstep_request=0, then this
  // routine places all points inside "dsprange" into display_list.
  //   NOTE: Non-integral values for step_request lead to unattractive
  // banding in the output display list, and so are currently
  // overridden.
#define MEMBERNAME "GetDisplayList"
  if(xstep_request<0 || ystep_request<0 || zstep_request<0)
    FatalError(-1,STDDOC,"%s: xstep=%f, ystep=%f, zstep=%d",
	       ErrBadParam,xstep_request,ystep_request,zstep_request);

  display_list.Clear();
  INT4m isize,jsize,ksize;
  GetDimens(isize,jsize,ksize);
  if(isize<1 || jsize<1 || ksize<1)
    return display_list.GetSize(); // Empty grid
  Vec3f rangemin,rangemax;
  dsprange.GetExtremes(rangemin,rangemax);

  // Convert dsprange & step requests from problem coords to nodal coords
  rangemin=ProblemToNodalCoords(rangemin);
  rangemax=ProblemToNodalCoords(rangemax);
  if(coords_step.x<0) swap(rangemin.x,rangemax.x); // Adjust for ordering
  if(coords_step.y<0) swap(rangemin.y,rangemax.y); // across coord transform.
  if(coords_step.z<0) swap(rangemin.z,rangemax.z);

  REAL4m nxstep=xstep_request/fabs(coords_step.x);
  REAL4m nystep=ystep_request/fabs(coords_step.y);
  REAL4m nzstep=zstep_request/fabs(coords_step.z);

  INT4m ifirst,ilast,jfirst,jlast,kfirst,klast;
  ifirst=INT4m(ceil(rangemin.x)); ilast=INT4m(floor(rangemax.x))+1;
  jfirst=INT4m(ceil(rangemin.y)); jlast=INT4m(floor(rangemax.y))+1;
  kfirst=INT4m(ceil(rangemin.z)); klast=INT4m(floor(rangemax.z))+1;

  if(ifirst<0 || ilast>isize || jfirst<0 || jlast>jsize ||
     kfirst<0 || klast>ksize) {
#if DEBUGLVL > 999
    NonFatalError(STDDOC,"Range box out-of-bounds: (%d,%d,%d) "
		  "to (%d,%d,%d), not in (0,0,0) to (%d,%d,%d)",
		  ifirst,jfirst,kfirst,ilast,jlast,klast,
		  isize,jsize,ksize);
#endif
    if(ifirst<0) ifirst=0; if(ilast>isize) ilast=isize;
    if(jfirst<0) jfirst=0; if(jlast>jsize) jlast=jsize;
    if(kfirst<0) kfirst=0; if(klast>ksize) klast=ksize;
  }

  // Calculate step sizes
  // Note: Non-integer step size produce ugly banding in output sampling
  nxstep=ROUND(nxstep);
  nystep=ROUND(nystep);
  nzstep=ROUND(nzstep);

  REAL4m max_step=MAX(MAX(isize,jsize),ksize);
  if(nxstep>max_step) nxstep=max_step;
  if(nystep>max_step) nystep=max_step;
  if(nzstep>max_step) nzstep=max_step;

  INT4m istep=INT4m(nxstep);
  INT4m jstep=INT4m(nystep);
  INT4m kstep=INT4m(nzstep);

  if(istep<1) istep=1;  if(jstep<1) jstep=1;  if(kstep<1) kstep=1;

  // Determine offsets
  INT4m ioff=((ilast-ifirst-1)%istep)/2;
  INT4m joff=((jlast-jfirst-1)%jstep)/2;
  INT4m koff=((klast-kfirst-1)%kstep)/2;

  INT4m i,j,k;
  INT4m istart,jstart,kstart;
  Vec3f position;  // Position, in "Problem" coords
  REAL4m xstart,ystart,zstart,xstep,ystep,zstep;

  istart=ifirst+ioff;
  xstart=coords_base.x+coords_step.x*REAL4m(istart);
  xstep=coords_step.x*REAL4m(istep);

  jstart=jfirst+joff;
  ystart=coords_base.y+coords_step.y*REAL4m(jstart);
  ystep=coords_step.y*REAL4m(jstep);

  kstart=kfirst+koff;
  zstart=coords_base.z+coords_step.z*REAL4m(kstart);
  zstep=coords_step.z*REAL4m(kstep);

  for(k=kstart,position.z=zstart;k<klast;k+=kstep,position.z+=zstep) {
    for(j=jstart,position.y=ystart;j<jlast;j+=jstep,position.y+=ystep) {
      for(i=istart,position.x=xstart;i<ilast;i+=istep,position.x+=xstep) {
	display_list.Append(LocatedVector(position,GridVec(i,j,k)));
      }
    }
  }
  // Export actual step sizes, in problem units
  xstep_request=nxstep*fabs(coords_step.x);
  ystep_request=nystep*fabs(coords_step.y);
  zstep_request=nzstep*fabs(coords_step.z);

  return display_list.GetSize();
#undef MEMBERNAME
}

//////////////////////////////////////////////////////////////////////////
// General (non-regular) 3D mesh of Vec3f's
const ClassDoc GeneralMesh3f::class_doc("GeneralMesh3f",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","1-Sep-1997");

// Generic Mesh functions
INT4m GeneralMesh3f::GetDisplayList(REAL4m &xstep_request,
				    REAL4m &ystep_request,
				    REAL4m & /* zstep_request */,
				    const BoundingBox & range_request,
				    List<LocatedVector> &display_list)
{
#define MEMBERNAME "GetDisplayList"
  display_list.Clear();
  mesh.ClearSelectCounts();
  if(xstep_request<SQRT_EPSILON || ystep_request<SQRT_EPSILON) {
    // Return whole list
    mesh.GetWholeList(display_list);
  }
  else {
    // Otherwise, center a maximal regular grid, with lattice
    // steps xstep_request, ystep_request, inside working mesh box
    // region "work_range", and then use mesh.GetClosest2D to collect
    // samples from the mesh closest to the overlaid regular grid.
    BoundingBox work_range=IntersectBoundingBoxes(data_range,range_request);
    Vec3f minpt,maxpt;
    work_range.GetExtremes(minpt,maxpt);
    const REAL4m fudge=1.001;  // Fudge factor, used to help
    /// against display "banding" caused by rounding.

    REAL4m xextent=maxpt.x-minpt.x;
    REAL4m xstep=xstep_request;
    INT4m  ixcount=int(floor((xextent*fudge)/xstep))+1;
    REAL4m xmargin=(xextent-xstep*(ixcount-1))/(2.*fudge);

    REAL4m yextent=maxpt.y-minpt.y;
    REAL4m ystep=ystep_request;
    INT4m  iycount=int(floor((yextent*fudge)/ystep))+1;
    REAL4m ymargin=(yextent-ystep*(iycount-1))/(2.*fudge);

    // If margin is close to an integral multiple of step_hints,
    // then round it to the multiple.  This should help remove
    // some display "banding" when sampling from grids that are
    // in actual fact regular.
    if(step_hints.x>EPSILON && xmargin>EPSILON) {
      REAL4m mult=xmargin/step_hints.x;
      REAL4m whole=ROUND(mult);
      REAL4m frac=mult-whole;
      if(whole>EPSILON && fabs(frac)<0.1)
	xmargin=step_hints.x*whole;
    }
    if(step_hints.y>EPSILON && ymargin>EPSILON) {
      REAL4m mult=ymargin/step_hints.y;
      REAL4m whole=ROUND(mult);
      REAL4m frac=mult-whole;
      if(whole>EPSILON && fabs(frac)<0.1)
	ymargin=step_hints.y*whole;
    }

    int ix,iy; REAL4m x,y;
    LocatedVector *lvp;  INT4m select_count;  Vec3f pos(0.,0.,0.);
    for(ix=0, x=minpt.x+xmargin; ix<ixcount; ix++, x+=xstep) {
      pos.x=x;
      for(iy=0, y=minpt.y+ymargin; iy<iycount; iy++, y+=ystep) {
	pos.y=y;
	mesh.GetClosest2D(pos,lvp,select_count);
	if(select_count==1) display_list.Append(*lvp);
      }
    }
  }
  return display_list.GetSize();
#undef MEMBERNAME
}

void GeneralMesh3f::SetApproximateStepHints()
{ // We should have a sophisticated algorithm here that computes
  // and returns the average nearest neighbor distance, but for
  // now we'll just calculate some "average" cell dimensions.
  // NOTE: This functions overwrites any pre-existing value in
  //       member variable step_hints, which may override a file
  //       specified value.  A client may want to check to see
  //       if step_hints.x==0 before calling this routine.
#define MEMBERNAME "SetApproximateStepHints"
  step_hints.Set(0.,0.,0.);
  INT4m point_count=GetSize();
  if(point_count<1) return;  // Meaningless case
  REAL4m total_area=fabs(data_range.GetWidth()*data_range.GetHeight());
  /// fabs() for safety.

  REAL4m avecell_area=total_area/point_count;  // "Cells" here
  /// really mean "cells" for display purposes, which are centered
  /// about the sample nodes.

  step_hints.x=sqrt(avecell_area);
  step_hints.y=step_hints.x;
  step_hints.z=0.;
  return;
#undef MEMBERNAME
}

void GeneralMesh3f::SetBoundaryList(const List<Vec3f> &boundary_list)
{
#define MEMBERNAME "SetBoundaryList"
  boundary=boundary_list;
  for(Vec3f *vp=boundary.GetFirst();vp!=NULL;
      vp=boundary.GetNext()) {
    range.ExpandWith(*vp);
  }
#undef MEMBERNAME
}

REAL4m GeneralMesh3f::GetApproximateCellWidth() const
{ // Uses step_hints.  If step_hints.x==0, then first
  // calls SetApproximateStepHints.
  // NOTE: This function is "conceptually const" as opposed to
  //  "bitwise const"; cf. p76-78 of Meyer's book, "Effective C++."
  //  As a result, we need to do some ugly un-const casting.
  //  This is dangerous, since it looses compiler checks against
  //  changes to ::SetApproximateStepHints that may destroy
  //  conceptual const-ness of this function.
#define MEMBERNAME "GetApproximateCellWidth"
  GeneralMesh3f * const localThis = (GeneralMesh3f *)this;
  if(step_hints.x==0) localThis->SetApproximateStepHints();
  if(step_hints.x==0) return 1.0;  // Nonsense case
  return sqrt(SQ(step_hints.x)+SQ(step_hints.y));
  /// We use the diagonal length of a the cell because a simple average
  /// of step_hints.x and step_hints.y doesn't "fill up enough whitespace"
  /// on irregular grids.
#undef MEMBERNAME
}

Vec3f GeneralMesh3f::GetApproximateCellDimensions() const
{
  // See note in :::GetApproximateCellWidth() about
  // "conceptually const".
#define MEMBERNAME "GetApproximateCellDimensions"
  GeneralMesh3f * const localThis = (GeneralMesh3f *)this;
  if(step_hints.x==0) localThis->SetApproximateStepHints();
  if(step_hints.x==0) return Vec3f(1.,1.,1.);  // Nonsense case
  return step_hints;
#undef MEMBERNAME
}

// GeneralMesh3f specific functions
void GeneralMesh3f::ResetMesh()
{
#define MEMBERNAME "ResetMesh"
  mesh.ClearList();
  data_range.Reset();
  boundary.Clear();
  range.Reset();
  step_hints.Set(0.,0.,0.);
#undef MEMBERNAME
}

void GeneralMesh3f::ResetMesh(const char* newFilename)
{
#define MEMBERNAME "ResetMesh(const char*)"
  SetFilename(newFilename);
  ResetMesh();
#undef MEMBERNAME
}

void GeneralMesh3f::ResetMesh(const char* newFilename,
			      const char* newTitle,
			      const char* newDescription)
{
#define MEMBERNAME "ResetMesh(const char*,const char*)"
  SetTitle(newTitle);
  SetDescription(newDescription);
  ResetMesh(newFilename);
#undef MEMBERNAME
}

void GeneralMesh3f::AddPoint(const LocatedVector& lv)
{
#define MEMBERNAME "AddPoint"
  mesh.AddPoint(lv);
  data_range.ExpandWith(lv.location);
  range.ExpandWith(lv.location);
#undef MEMBERNAME
}

void GeneralMesh3f::SortPoints()
{
#define MEMBERNAME "SortPoints"
  // Guess at cell dimensions, if necessary
  if(step_hints.x==0) SetApproximateStepHints();

  // Make BoxList mesh bounding box big enough so that
  // refinement boxes will include all sampling points
  // we may ask of it.
  if(!range.IsEmpty()) {
    Vec3f min,max;
    range.GetExtremes(min,max);
    mesh.ExpandBoxRegion(min.x,min.y,max.x,max.y);
  }
  mesh.InflateBoxRegion(1.2,1.2);   // Add a 10% margin on all sides

  // OK, now do refinement
  // mesh.Refine(6);
  mesh.Refine(1024,2.,15.);


#if DEBUGLVL>9
  Message("Mesh refinement stats (cf. GeneralMesh3f::SortPoints()"
	  " in mesh.cc)\n"
	  "  Number of subboxes: %d\n"
	  "  Average number of points per box: %.2f\n"
	  "  Average box list length:  %.2f\n"
	  "  Memory waste: %d\n",
	  mesh.GetBoxCount(),mesh.GetAveInCount(),mesh.GetAveListCount(),
	  mesh.GetSpaceWaste());
#endif // DEBUGLVL>9
#undef MEMBERNAME
}



