/* FILE: mmsInit.cc                 -*-Mode: c++-*-
 *
 * Initialization routines for a Tcl/Tk loadable module which adds
 * a set of Tcl commands to an interpreter suitable for controlling
 * a 3D micromagnetic simulation.
 *
 */

#include <tcl.h>
#include "mmsolve.h"
#include "misc.h"
#include "mifio.h"
#include "grid.h"

/* End includes */

/*
 *----------------------------------------------------------------------
 *
 * DllEntryPoint --
 *
 *      This wrapper function is used by Windows to invoke the
 *      initialization code for the DLL.  If we are compiling
 *      with Visual C++, this routine will be renamed DllMain.
 *
 * Results:
 *      Returns TRUE;
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

#ifdef FUTURE_CODE
#ifdef __WIN32__
BOOL APIENTRY
DllEntryPoint(HINSTANCE hInst, DWORD reason, LPVOID reserved)
//    HINSTANCE hInst;          /* Library instance handle. */
//    DWORD reason;             /* Reason this function is being called. */
//    LPVOID reserved;          /* Not used. */
{
    return TRUE;
}
#endif
#endif // FUTURE_CODE

/*
 *----------------------------------------------------------------------
 *
 * Mmsolve_Init --
 *
 *      This procedure initializes the mmSolve extension.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

EXPORT(int,Mmsolve_Init)(Tcl_Interp *interp)
{
  AutoBuf ab;
  Tcl_CreateCommand(interp,ab("mms_mif"), Mms_MifCmd, NULL, NULL);
  Tcl_CreateCommand(interp,ab("mms_grid2D"), Mms_Grid2DCmd, NULL, NULL);
  return Tcl_PkgProvide(interp,ab("mmsolve"),AutoBuf(MMSOLVE_VERSION));
}

/*
 *----------------------------------------------------------------------
 *
 * Mmsolve_SafeInit --
 *
 *      This procedure initializes the example command for a safe
 *      interpreter.  You would leave out of this procedure any commands
 *      you deemed unsafe.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

EXPORT(int,Mmsolve_SafeInit)(Tcl_Interp *interp)
{
    return Mmsolve_Init(interp);
}

/*
 * Support macros
 */

// Macros which take a list of conversion functions and return an argument
// list where those functions are applied to elements of argv.
#define ARGN0() ()
#define ARGN1(a1) (a1(argv[2])) 
#define ARGN2(a1,a2) (a1(argv[2]), a2(argv[3]))
#define ARGN3(a1,a2,a3) (a1(argv[2]), a2(argv[3]), a3(argv[4]))
#define ARGN4(a1,a2,a3,a4) (a1(argv[2]), a2(argv[3]), a3(argv[4]), a4(argv[5]))
#define ARGN5(a1,a2,a3,a4,a5) (a1(argv[2]), a2(argv[3]), a3(argv[4]), \
			       a4(argv[5]), a5(argv[6])) 

#define ARGE0() (errorcode)
#define ARGE1(a1) (a1(argv[2]), errorcode) 
#define ARGE2(a1,a2) (a1(argv[2]), a2(argv[3]), errorcode)
#define ARGE3(a1,a2,a3) (a1(argv[2]), a2(argv[3]), a3(argv[4]), errorcode)
#define ARGE4(a1,a2,a3,a4) (a1(argv[2]), a2(argv[3]), a3(argv[4]), \
			    a4(argv[5]), errorcode)
#define ARGE5(a1,a2,a3,a4,a5) (a1(argv[2]), a2(argv[3]), a3(argv[4]), \
			    a4(argv[5]), a5(argv[6]), errorcode)

#define ARG(N,LIST,ERROR) ARG##ERROR##N LIST

// Conversion routines
#define NOP(x) x
#define VOIDC(func) (func,"")    // Evaluate "void" return type function
/// func, and return the Tcl "void" return value of "".
static const char* Vec3Dtoa(Vec3D v)
{ // Converts a Vec3D to an "ascii" character string.
  static char buf[256]; // Should be big enough!
  sprintf(buf,"%.10g %.10g %.10g",v.GetX(),v.GetY(),v.GetZ());
  return buf;
}
static char atoc(const char *str)
{
  return str[0];
}

// A single macro that expands into a C function that wraps an instance
// method of a C++ class, providing a Tcl interface.  It is parameterized
// by several arguments, and several 'global' definitions assumed to be
// #define-d above and #undefine-d below uses of this macro.  The 'global'
// definitions help reduce the number of arguments.
//	Uses globals:
//		EXTENSION_TCL_PREFIXSTR
//		CLASS_C_NAME
//		CLASS_TCL_NAME
//	Uses arguments:
//		METHOD - name of method
//		NUMARGS - number of arguments which method takes
//		USAGE - a string describing the arguments to the 
//			methods, suitable for a usage message
//		CVT - a parenthesized list of NUMARGS conversion 
//			routines to be applied to strings to generate
//			arguments of the proper type for the C++ class
//			method
//		RCVT - a conversion routine to apply to the return value
//			of the method.
//		RFMT - an sprintf format string for converting the return
//			value of RCVT to a string.
//              ERROR - whether or not the C++ class method returns an error
//                     code.  Set ERROR to E if it does, N if not.
//
//
#define MAKE_INSTANCE_NAME1(x,y) x##Instance##y
#define MAKE_INSTANCE_NAME(x,y) MAKE_INSTANCE_NAME1(x,y)
#define STRINGIFY1(x) #x
#define STRINGIFY(x) STRINGIFY1(x)
/// We need 2 levels of replacement to get token pasting and
//  "stringification" to work properly.


#define INSTANCE_METHOD(METHOD,NUMARGS,USAGE,CVT,RCVT,RFMT,ERROR)       \
static int                                                              \
MAKE_INSTANCE_NAME(CLASS_C_NAME,METHOD)                                 \
(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)      \
{                                                                       \
    int i;                                                              \
    if (argc != NUMARGS + 2) {                                          \
        Tcl_AppendResult(interp, "\n'<",                                \
                STRINGIFY(EXTENSION_TCL_PREFIX),                        \
                "_", STRINGIFY(CLASS_TCL_NAME), " instance> ",          \
		argv[1],"' usage:\n\t$instance ",                       \
                argv[1], USAGE,                                         \
                "\n\nInvalid call:\n\t$instance",                       \
                (char *) NULL);                                         \
        for (i=1; i<argc; i++) {                                        \
            Tcl_AppendResult(interp, " ", argv[i], (char *) NULL);      \
        }                                                               \
        return TCL_ERROR;                                               \
    }                                                                   \
    int errorcode=0;                                                    \
    CLASS_C_NAME *classPtr = (CLASS_C_NAME *) clientData;               \
    sprintf(interp->result, #RFMT,                                      \
            RCVT(classPtr->METHOD ARG(NUMARGS,CVT,ERROR)));             \
    if(errorcode>0) {                                                   \
        /* Error detected.  Overwrite return value with error       */  \
        /* message and return TCL_ERROR.  The static portion of     */  \
        /* interp->result is only 200 bytes wide, so we use         */  \
        /* Tcl_SetResult() in case of longer error message strings. */  \
        String errmsg;                                                  \
        MessageLocker::GetMessage(errmsg);                              \
        Tcl_ResetResult(interp);                                        \
	Tcl_AppendResult(interp,errmsg.GetStr(),(char *)NULL);          \
        MessageLocker::Reset();                                         \
        return TCL_ERROR;                                               \
    } else if(errorcode<0) {                                            \
       /* Warning message detected.  Should write to some sort of   */  \
       /* log channel, but for now just pitch into bit bucket.      */  \
       MessageLocker::Reset();                                          \
       return TCL_OK;                                                   \
    }                                                                   \
    return TCL_OK;                                                      \
}

#define EXTENSION_TCL_PREFIX  mms
#define         CLASS_C_NAME  MIF
#define       CLASS_TCL_NAME  mif
INSTANCE_METHOD(GetA,0,"",(),NOP,%.10g,N)
INSTANCE_METHOD(GetFieldStepIndex,0,"",(),NOP,%d,N);
INSTANCE_METHOD(GetFieldStepTotal,0,"",(),NOP,%d,N);
INSTANCE_METHOD(GetK1,0,"",(),NOP,%.10g,N);
INSTANCE_METHOD(GetLogLevel,0,"",(),NOP,%d,N);
INSTANCE_METHOD(GetMaxIterationCount,0,"",(),NOP,%d,N);
INSTANCE_METHOD(GetMs,0,"",(),NOP,%.10g,N);
INSTANCE_METHOD(GetOutBaseName,0,"",(),NOP,%s,N);
INSTANCE_METHOD(GetRandomizerSeed,0,"",(),NOP,%d,N);
INSTANCE_METHOD(GetUserInteractionLevel,0,"",(),NOP,%d,N);
INSTANCE_METHOD(GetConvergeTorqueValue,0,"",(),NOP,%.10g,N);
INSTANCE_METHOD(GetUserReportCode,0,"",(),NOP,%d,N);
INSTANCE_METHOD(ResetField,0,"",(),VOIDC,%s,N);
INSTANCE_METHOD(SetFieldStepIndex,1,"<new field index>",(atoi),NOP,%d,N);
INSTANCE_METHOD(StepField,0,"",(),NOP,%d,E);
INSTANCE_METHOD(Read,1,"<filename>",(NOP),VOIDC,%s,E);
#undef EXTENSION_TCL_PREFIX
#undef CLASS_C_NAME
#undef CLASS_TCL_NAME


#define EXTENSION_TCL_PREFIX  mms
#define         CLASS_C_NAME  Grid2D
#define       CLASS_TCL_NAME  grid2d
INSTANCE_METHOD(BinDumph,2,"<filename note>",(NOP,NOP),VOIDC,%s,N);
INSTANCE_METHOD(BinDumpm,2,"<filename note>",(NOP,NOP),VOIDC,%s,N);
INSTANCE_METHOD(WriteOvf,5,"<type filename note>",\
		(atoc,NOP,NOP,NOP,NOP),VOIDC,%s,N);
INSTANCE_METHOD(GetAveMag,0,"",(),Vec3Dtoa,%s,N);
INSTANCE_METHOD(GetCellSize,0,"",(),NOP,%.10g,N);
INSTANCE_METHOD(GetEnergy,0,"",(),NOP,%.10g,N);
INSTANCE_METHOD(GethUpdateCount,0,"",(),NOP,%d,N);
INSTANCE_METHOD(GetMaxNeighborAngle,0,"",(),NOP,%.10g,N);
INSTANCE_METHOD(GetMaxTorque,0,"",(),NOP,%.10g,N);
INSTANCE_METHOD(GetStepSize,0,"",(),NOP,%.10g,N);
INSTANCE_METHOD(GetTimeStep,0,"",(),NOP,%.10g,N);
INSTANCE_METHOD(MarkhInvalid,0,"",(),VOIDC,%s,N);
INSTANCE_METHOD(GetNomBApplied,0,"",(),Vec3Dtoa,%s,N);
INSTANCE_METHOD(ResetODE,0,"",(),VOIDC,%s,N);
INSTANCE_METHOD(RoughenBoundaryMoments,0,"",(),VOIDC,%s,N);
INSTANCE_METHOD(StepODE,0,"",(),NOP,%.10g,E);
#undef EXTENSION_TCL_PREFIX
#undef CLASS_C_NAME
#undef CLASS_TCL_NAME


////////////////////////////////////////////////////////////////////////
//
//  Special Cases
//
static int
MIFInstanceGetNomExternalField
(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    if (argc < 2 || argc >3) {
        Tcl_AppendResult(interp, "\n'<mms_mif instance> ", argv[1],
                "' usage:\n\t$instance ", argv[1],
		" GetNomExternalField [range]",
                "\n\nInvalid call:\n\t$instance", (char *) NULL);
        for (i=1; i<argc; i++) {
            Tcl_AppendResult(interp, " ", argv[i], (char *) NULL);
        }
        return TCL_ERROR;
    }
    MIF *classPtr = (MIF *) clientData;

    double* field=NULL;
    if(argc<3) field=classPtr->GetNomExternalField();
    else       field=classPtr->GetNomExternalField(atoi(argv[2]));

#if MIFVECDIM != 3
#error "Printing code INCORRECTLY assumes MIFVECDIM vector dimension is 3"
#endif
    sprintf(interp->result,"%.10g %.10g %.10g",field[0],field[1],field[2]);

    return TCL_OK;
}

static int
Grid2DInstanceGetEnergies
(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    if (argc != 2 ) {
        Tcl_AppendResult(interp, "\n'<mms_mif instance> ", argv[1],
                "' usage:\n\t$instance ", argv[1]," GetEnergies",
                "\n\nInvalid call:\n\t$instance",
		"\n\nReturns: Eexch Eanis Edemag Ezeeman Etotal\n",
	        (char *) NULL);
        for (i=1; i<argc; i++) {
            Tcl_AppendResult(interp, " ", argv[i], (char *) NULL);
        }
        return TCL_ERROR;
    }
    Grid2D *classPtr = (Grid2D *) clientData;

    double exch,anis,demag,zeeman,total;
    exch=anis=demag=zeeman=total=0.0;  // Just to keep compiler quiet.
    classPtr->GetEnergies(exch,anis,demag,zeeman,total);

    sprintf(interp->result,"%10.10g %10.10g %10.10g %10.10g %10.10g",
	    exch,anis,demag,zeeman,total);

    return TCL_OK;
}

static int
Grid2DInstanceGetDimens
(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int i;
    if (argc != 2 ) {
        Tcl_AppendResult(interp, "\n'<mms_mif instance> ", argv[1],
                "' usage:\n\t$instance ", argv[1]," GetDimens",
                "\n\nInvalid call:\n\t$instance",
		"\n\nReturns: Nx Ny\n",
	        (char *) NULL);
        for (i=1; i<argc; i++) {
            Tcl_AppendResult(interp, " ", argv[i], (char *) NULL);
        }
        return TCL_ERROR;
    }
    Grid2D *classPtr = (Grid2D *) clientData;

    int Nx,Ny;
    Nx=Ny=0;  // Just to keep compiler quiet.
    classPtr->GetDimens(Nx,Ny);

    sprintf(interp->result,"%4d %4d",Nx,Ny);

    return TCL_OK;
}

static int
MIFInstancedelete
(ClientData /*clientData*/, Tcl_Interp *interp, int /*argc*/, char **argv)
{
    static char buf[256]; // Should be big enough
    sprintf(buf,"rename %s {}",argv[0]);
    return Tcl_GlobalEval(interp, buf);
}

static int
Grid2DInstancedelete
(ClientData /*clientData*/, Tcl_Interp *interp, int /*argc*/, char **argv)
{
    static char buf[256]; // Should be big enough
    sprintf(buf,"rename %s {}",argv[0]);
    return Tcl_GlobalEval(interp, buf);
}

#define MMSFUNCLIST(NAME,METHOD) { #METHOD , NAME##Instance##METHOD }
struct funcmap { const char* methodname; TCLCMD* methodfunc; };
struct funcmap MifMethods[] = {
         MMSFUNCLIST(MIF,GetA),
	 MMSFUNCLIST(MIF,GetFieldStepIndex),
	 MMSFUNCLIST(MIF,GetFieldStepTotal),
         MMSFUNCLIST(MIF,GetK1),
         MMSFUNCLIST(MIF,GetLogLevel),
         MMSFUNCLIST(MIF,GetMaxIterationCount),
         MMSFUNCLIST(MIF,GetMs),
         MMSFUNCLIST(MIF,GetNomExternalField),
         MMSFUNCLIST(MIF,GetOutBaseName),
         MMSFUNCLIST(MIF,GetRandomizerSeed),
         MMSFUNCLIST(MIF,GetUserInteractionLevel),
         MMSFUNCLIST(MIF,GetUserReportCode),
         MMSFUNCLIST(MIF,GetConvergeTorqueValue),
         MMSFUNCLIST(MIF,Read),
         MMSFUNCLIST(MIF,ResetField),
         MMSFUNCLIST(MIF,SetFieldStepIndex),
         MMSFUNCLIST(MIF,StepField),
         { "delete" , MIFInstancedelete },
         { (const char *)NULL, (TCLCMD *)NULL }
};
struct funcmap Grid2DMethods[] = {
         MMSFUNCLIST(Grid2D,BinDumph),
         MMSFUNCLIST(Grid2D,BinDumpm),
         MMSFUNCLIST(Grid2D,WriteOvf),
         MMSFUNCLIST(Grid2D,GetAveMag),
         MMSFUNCLIST(Grid2D,GetCellSize),
         MMSFUNCLIST(Grid2D,GetDimens),
         MMSFUNCLIST(Grid2D,GetEnergies),
         MMSFUNCLIST(Grid2D,GetEnergy),
         MMSFUNCLIST(Grid2D,GethUpdateCount),
         MMSFUNCLIST(Grid2D,GetMaxNeighborAngle),
         MMSFUNCLIST(Grid2D,GetMaxTorque),
         MMSFUNCLIST(Grid2D,GetStepSize),
         MMSFUNCLIST(Grid2D,GetTimeStep),
         MMSFUNCLIST(Grid2D,MarkhInvalid),
         MMSFUNCLIST(Grid2D,GetNomBApplied),
         MMSFUNCLIST(Grid2D,ResetODE),
         MMSFUNCLIST(Grid2D,RoughenBoundaryMoments),
         MMSFUNCLIST(Grid2D,StepODE),
         { "delete" , Grid2DInstancedelete },
         { (const char *)NULL, (TCLCMD *)NULL }
};


// Macro that expands into the instance method invoking command
//	NAME and
//	TCLNAME
#define MMS_INSTANCE(NAME,TCLNAME)                                      \
static int                                                              \
Mms_##NAME##InstanceCmd                                                 \
(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)      \
{                                                                       \
    int i;                                                              \
    if (argc < 2) {                                                     \
        Tcl_AppendResult(interp, "no value given for parameter \"",     \
                "method\" to \"", argv[0], "\"", (char *) NULL);        \
        return TCL_ERROR;                                               \
    }                                                                   \
    for(i=0 ; NAME##Methods[i].methodname!=NULL ; i++) {                \
        if(strcmp(argv[1],##NAME##Methods[i].methodname)==0) {          \
            return NAME##Methods[i].methodfunc(clientData,interp,       \
                    argc,argv);                                         \
        }                                                               \
    }                                                                   \
    Tcl_AppendResult(interp, "\n'<mms_##TCLNAME instance>' usage:\n\t", \
		     "$instance <instance_method> ?args?\n\n",          \
		     "Invalid call:\n\t$instance", (char *) NULL);      \
    for (i=1; i<argc; i++) {                                            \
      Tcl_AppendResult(interp, " ", argv[i], (char *) NULL);            \
    }                                                                   \
    Tcl_AppendResult(interp, "\n\nInvalid instance method: ", argv[1],  \
		     "\n\nValid instance methods:\n\t", (char *) NULL); \
    for(i=0 ; NAME##Methods[i].methodname!=NULL ; i++) {                \
        Tcl_AppendResult(interp, " ", NAME##Methods[i].methodname,      \
                (char *) NULL);                                         \
    }                                                                   \
    return TCL_ERROR;                                                   \
}     

MMS_INSTANCE(Mif,mif);
MMS_INSTANCE(Grid2D,grid2D);


// Macro that expands into a cleanup command to be called when the
// corresponding instance command is deleted in a Tcl interpreter.
#define MMS_CLASSDELETEPROC(NAME,CLASS)                                 \
void                                                                    \
Mms##NAME##DeleteProc(ClientData clientData)                            \
{                                                                       \
    CLASS *classPtr = (CLASS *) clientData;                             \
    delete classPtr;                                                    \
}

MMS_CLASSDELETEPROC(Mif,MIF);
MMS_CLASSDELETEPROC(Grid2D,Grid2D);


// Macro that expands into a TCLCMD for the 'new' proc of a Tcl class
// which takes no arguments.  It links a Tcl instance command to an
// instance of the C++ class CLASS.
#define MMS_CLASSNEWPROC_0P(NAME,CLASS)                                 \
static int                                                              \
Mms##NAME##New                                                          \
(ClientData /* clientData */, Tcl_Interp *interp, int argc, char **argv)\
{                                                                       \
    int i;                                                              \
    if (argc > 2) {                                                     \
        Tcl_AppendResult(interp, "\n'", argv[0], " ", argv[1],          \
                "' usage:\n\t", argv[0], " ", argv[1],                  \
                "\n\nInvalid call:\n\t", argv[0], (char *) NULL);       \
        for (i=1; i<argc; i++) {                                        \
            Tcl_AppendResult(interp, " ", argv[i], (char *) NULL);      \
        }                                                               \
        return TCL_ERROR;                                               \
    }                                                                   \
    static int nextinstance = 0;                                        \
    CLASS *classPtr = new CLASS##();                                    \
    sprintf(interp->result, "_%s%d", argv[0], nextinstance);            \
    Tcl_CreateCommand(interp, interp->result, Mms_##NAME##InstanceCmd,  \
            (ClientData) classPtr, Mms##NAME##DeleteProc);              \
    nextinstance += 1;                                                  \
    return TCL_OK;                                                      \
}

MMS_CLASSNEWPROC_0P(Mif,MIF);


// Macro that expands into a TCLCMD for the 'new' proc of a Tcl class
// which takes 1 arguments.  It links a Tcl instance command to an
// instance of the C++ class CLASS.
//	USAGE_STR is the call specific usage message
//	CNVT_FUNC1 converts the string argument into the type needed
//		by the CLASS constructor.
#define MMS_CLASSNEWPROC_1P(NAME,CLASS,USAGE_STR,CNVT_FUNC1)            \
static int                                                              \
Mms##NAME##New                                                          \
(ClientData /* clientData */, Tcl_Interp *interp, int argc, char **argv)\
{                                                                       \
    int i;                                                              \
    if (argc != 3) {                                                    \
        Tcl_AppendResult(interp, "\n'", argv[0], " ", argv[1],          \
                "' usage:\n\t", argv[0], " ", argv[1], " " USAGE_STR,   \
                "\n\nInvalid call:\n\t", argv[0], (char *) NULL);       \
        for (i=1; i<argc; i++) {                                        \
            Tcl_AppendResult(interp, " ", argv[i], (char *) NULL);      \
        }                                                               \
        return TCL_ERROR;                                               \
    }                                                                   \
    static int nextinstance = 0;                                        \
    CLASS *classPtr = new CLASS##(##CNVT_FUNC1##(argv[2]));             \
    sprintf(interp->result, "_%s%d", argv[0], nextinstance);            \
    Tcl_CreateCommand(interp, interp->result, Mms_##NAME##InstanceCmd,  \
            (ClientData) classPtr, Mms##NAME##DeleteProc);              \
    nextinstance += 1;                                                  \
    return TCL_OK;                                                      \
}


// Macro that expands into a function that returns the CLASS reference
// associated with the Tcl class instance command 'name' in 'interp'
#define MMS_CLASS_REF(NAME,CLASS)                                       \
CLASS &                                                                 \
Mms##NAME##Ref(Tcl_Interp *interp, char *name)                          \
{                                                                       \
    Tcl_CmdInfo info;                                                   \
    if (!Tcl_GetCommandInfo(interp, name, &info)) {                     \
        panic(AutoBuf("No object corresponding to name %s"), name);     \
    }                                                                   \
    return (CLASS &) (*((CLASS *) info.clientData));                    \
}


// Macro that adds the interp argument to a function call
#define MMS_ADDINTERP(ARG) (interp, ARG)

MMS_CLASS_REF(Mif,MIF)
MMS_CLASSNEWPROC_1P(Grid2D,Grid2D,"<mms_mif instance>",
		    MmsMifRef MMS_ADDINTERP);


// Macro that expands into a TCLCMD which implements the Tcl command
// indicated by NAME.  This command serves as the Tcl interface for
// a C++ class of the same name (modulo capitalization).
#define MMS_CLASS(NAME)                                                 \
int                                                                     \
Mms_##NAME##Cmd                                                         \
(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)      \
{                                                                       \
    char c;                                                             \
    int i;                                                              \
    if (argc < 2) {                                                     \
        Tcl_AppendResult(interp, "no value given for parameter \"",     \
                "proc\" to \"", argv[0], "\"", (char *) NULL);          \
        return TCL_ERROR;                                               \
    }                                                                   \
    c = argv[1][0];                                                     \
    if ((c == 'n') && (strcmp(argv[1], "new") == 0)) {                  \
        return Mms##NAME##New(clientData, interp, argc, argv);          \
    }                                                                   \
    Tcl_AppendResult(interp, "\n'", argv[0], "' usage:\n\t",            \
           argv[0], " <class_proc> ?args?\n\nInvalid call:\n\t",        \
           argv[0], (char *) NULL);                                     \
    for (i=1; i<argc; i++) {                                            \
        Tcl_AppendResult(interp, " ", argv[i], (char *) NULL);          \
    }                                                                   \
    Tcl_AppendResult(interp, "\n\nInvalid class proc: ", argv[1],       \
            "\n\nValid class procs:\n\tnew", (char *) NULL);            \
    return TCL_ERROR;                                                   \
}

MMS_CLASS(Mif);
MMS_CLASS(Grid2D);

