/* FILE: vector.h           -*-Mode: c++-*-
 *
 * This file contains class declarations for vector classes.
 *
 */

#ifndef _VECTOR_H
#define _VECTOR_H

#include <stdio.h>

/* End includes */

/*
 * Notes on vector classes:
 *
 * 1) Since the number of vectors may be large, and we may want to allow
 * external processing, the array memory is allocated external to the
 * class and a pointer is sent to each realization.  In particular, the
 * array memory is _not_ owned by the vector, but must be allocated and
 * freed by the calling program.  This is dangerous, since the calling
 * program can delete the memory and leave the Vector instance with an
 * invalid memory pointer.  But at this point in time my judgement is
 * that the benefits outweigh the risks...
 *
 * 2) In view of 1), and as a crude check of initial allocation, the
 * Vector/UnitVector initialization sequences initialize the array
 * memory to 0's (Vector) or (1 0 0 ...) (UnitVector), unless the
 * initialization routine is called without array pointer, in which
 * case arr is set to point to worksp, which is then _not_ initialized.
 * (This helps out in the case where worksp is being used as temporary
 * storage for dynamically allocated (usually temporary) vectors.)
 *
 * 3) One is tempted at first to make UnitVector a class derived from
 * Vector, but the member functions have to all be rewritten anyway,
 * and the overhead associated with virtual function calls probably does
 * not justify their expense.  If it appears in the future that a common
 * interface to Vector and UnitVector is worthwhile, I can always define
 * a virtual class Array and derive both Vector and UnitVector from it.
 *
 * 4) The overhead involved in dynamic allocation of arrays of indeterminate
 * size (needed for simple operators like `+') seems to necessitate fixed
 * length vectors (ThreeVector and ThreeVector) rather than more general
 * beasts.
 *
 * 5) The use of a common workspace array can cause all types of aliasing
 * problems, especially in compound calls to vector operators.  See, for
 * example, the cross product (^) operator.  For this reason the workspace
 * array ThreeVector::worksp has been eliminated, and the use of static
 * arrays inside functions must be _closely_ analyzed.
 *
 */
#define VECDIM 3  // This is silly, but may make modifications easier

class ThreeVector {
protected:
  double*  arr;
  static double zeroarr[VECDIM];
  static ThreeVector zero;
public:
  void     Init(double* array); // Inits array to (0,0,0), so array != NULL
  void     Set(double* array) { arr=array; } // Does not access array.
           ThreeVector(void);
           ThreeVector(double* array);
           ThreeVector& Set(double a,double b,double c); // *this=<a,b,c>
  void     ReleaseArray(void); // Resets arr to point to NULL
          ~ThreeVector(void);
  static const ThreeVector& ZeroVector(void) { return zero; }
  static int Alloc2DArray(int dim1,int dim2,ThreeVector** &v);
  static void Release2DArray(int dim1,ThreeVector** &v);
  void     Dup(const double* array);

  inline double&  operator[](int n);
  inline const double&  operator[](int n) const;
  inline ThreeVector&  operator=(const ThreeVector& u);
  inline ThreeVector&  operator=(double c);  // *this = <c,c,c>
  operator double*() const { return arr; } // Use this sparingly!!!

  ThreeVector&  operator^=(const ThreeVector& v);
  inline ThreeVector&  operator*=(double a);
  inline ThreeVector&  operator+=(const ThreeVector& v);
  inline ThreeVector&  operator-=(const ThreeVector& v);

  friend inline double operator*(const ThreeVector& u,
				 const ThreeVector& v); // Dot product
  inline double MagSq(void) const; // Magnitude squared
  double   Mag(void) const;   // Magnitude
  void     Random(double magnitude); // Fills arr with "random" vector of
                                     // given magnitude.
  void     Perturb(double max_mag); // Perturbs with a random vector no bigger
                                    // than max_mag.
  void     Scale(double new_mag);
  inline void FastAdd(double a,const ThreeVector& v,double b,
		      const ThreeVector& w);
                                                  // *this = a.v+b.w 
  inline void FastAdd(const ThreeVector& v,double b,const ThreeVector& w);
                                                  // *this = v+b.w 
  inline void Accum(double a,const ThreeVector& v);
  void     SignAccum(ThreeVector& accumpos,ThreeVector& accumneg);
                   // Accumulate positive components of *this into
                   // accumpos, negative components into accumneg.
  void     Print(FILE* fout);

  // The following functions are pretty broken.  Avoid using them,
  // and especially don't use in compound statements (you're likely
  // to get aliasing errors on the returned reference).
  ThreeVector&  operator*(double a);
  ThreeVector&  operator^(const ThreeVector& v); // Cross product
  ThreeVector&  operator+(const ThreeVector& v);
  ThreeVector&  operator-(const ThreeVector& v);
};

inline double& ThreeVector::operator[](int n)
{
#ifdef CODE_CHECKS
  if(n<0 || n>=VECDIM)
    PlainError(1,"Error in ThreeVector::operator[]: %s",ErrBadParam);
#endif // CODE_CHECKS
  return arr[n];
}

inline const double& ThreeVector::operator[](int n) const
{
#ifdef CODE_CHECKS
  if(n<0 || n>=VECDIM)
    PlainError(1,"Error in ThreeVector::operator[] const: %s",ErrBadParam);
#endif // CODE_CHECKS
  return arr[n];
}

inline double operator*(const ThreeVector& u,const ThreeVector& v)
{ // Dot product
  int i=0; double dot=0.0,*uptr=u.arr,*vptr=v.arr;
  while( i++ < VECDIM ) dot+=(*(uptr++))*(*(vptr++));
  return dot;
}

inline ThreeVector& ThreeVector::operator=(const ThreeVector& v)
{
  for(int i=0;i<VECDIM;i++) arr[i]=v.arr[i];
  return *this;
}

inline ThreeVector& ThreeVector::operator=(double c)
{ // *this = <c,c,c>
  for(int i=0;i<VECDIM;i++) arr[i]=c;
  return *this;
}

inline double ThreeVector::MagSq(void) const
{
  return SQ(arr[0])+SQ(arr[1])+SQ(arr[2]);
}

inline ThreeVector& ThreeVector::operator*=(double a)
{
  for(int i=0;i<VECDIM;i++) arr[i]*=a;
  return *this;
}

inline ThreeVector& ThreeVector::operator+=(const ThreeVector& v)
{
  for(int i=0;i<VECDIM;i++) arr[i]+=v.arr[i];
  return *this;
}

inline ThreeVector& ThreeVector::operator-=(const ThreeVector& v)
{
  for(int i=0;i<VECDIM;i++) arr[i]-=v.arr[i];
  return *this;
}

inline void ThreeVector::FastAdd(double a,const ThreeVector& v,
				 double b,const ThreeVector& w)
{ // *this = a.v+b.w
  int i;
  for(i=0;i<VECDIM;i++) arr[i]=a*v.arr[i]+b*w.arr[i];
}

inline void ThreeVector::FastAdd(const ThreeVector& v,double b,
				 const ThreeVector& w)
{ // *this = v+b.w
  int i;
  for(i=0;i<VECDIM;i++) arr[i]=v.arr[i]+b*w.arr[i];
}

inline void ThreeVector::Accum(double a,const ThreeVector& v)
{
  int i;
  for(i=0;i<VECDIM;i++) arr[i]+=a*v.arr[i];
}

/////////////////// Array2D Template Declaration ////////////////////
template<class T> class Array2D {
private:
  int dim1,dim2;  // I would *like* these private
  T *data,**row;  // I would *like* these private
public:
  Array2D() { dim1=dim2=0; data=(T*)NULL; row=(T**)NULL; }
  Array2D(Array2D<T> &src); // Copy constructor
  ~Array2D() { Deallocate(); }
  void Allocate(int newdim1,int newdim2);
  void Deallocate();
  typedef T* T_ptr; // Some compilers aren't too smart
  typedef const T* cT_ptr; // Some compilers aren't too smart
  T_ptr operator[](int index1);  // Returns pointer to row (or column,
  //  depending on how you think about these things. ;-)
  cT_ptr operator[](int index1) const;
  void Swap(Array2D<T> &other); // Interchanges *this & other
  int GetSize(int &d1,int &d2) { d1=dim1; d2=dim2; return d1*d2; }
  int GetSize() { return dim1*dim2; }
  typedef T** T_ptr_ptr; // Some compilers aren't too smart
  operator T_ptr_ptr() { return row; }
};

template<class T> Array2D<T>::Array2D(Array2D<T> &src)
{ // Copy constructor
  dim1=dim2=0; data=(T*)NULL; row=(T**)NULL;
  Allocate(src.dim1,src.dim2);
  int i,j;
  T *tp,*tsrcp;
  for(i=0,tp=data,tsrcp=src.data;i<dim1;i++) {
/*************************** BUG ALERT!!! *********************/
//  The next line looks like a bug to me... -mjd 12/Dec/96  /**/
//    row[i]=src.row[i]; // Copy row pointers               /**/
/**************************************************************/
    for(j=0;j<dim2;j++) *(tp++)=*(tsrcp++); // Copy data
  }
}

template<class T> void Array2D<T>::Allocate(int newdim1,int newdim2)
{ // IMPORTANT NOTE: This routine allocates everything as one dimensional
  //   arrays, and then sets up pointers into that array.
  int i;
  if(data!=(T*)NULL) Deallocate();
  dim1=newdim1; dim2=newdim2;
  if(dim1<1 || dim2<1 || dim1*dim2<1) // The last is an overflow check
    PlainError(1,"Error in Array2D::Allocate: %s",ErrBadParam);

  // Allocate one big chunk of memory for the T's
  if(!(data=new T[dim1*dim2])) PlainError(1,"Error in Array2D::Allocate: %s",ErrNoMem);
  // Allocate row pointers
  if(!(row=new T*[dim1])) PlainError(1,"Error in Array2D::Allocate: %s",ErrNoMem);

  // Initialize row pointers
  for(i=0;i<dim1;i++) row[i]=data+i*dim2;
}

template<class T> void Array2D<T>::Deallocate(void)
{
  if(data!=(T*)NULL && dim1>0 && dim2>0) {
    delete[] row;
    delete[] data;
  }
  data=(T*)NULL; row=(T**)NULL;
  dim1=dim2=0;
}

template<class T> T* Array2D<T>::operator[](int index1)
{
#ifdef CODE_CHECKS
  if(index1<0 || index1>dim1)
    PlainError(1,"Error in Array2D::operator[]: %s",ErrArrOutOfBounds);
#endif // CODE_CHECKS  
  return(row[index1]);
}

template<class T> const T* Array2D<T>::operator[](int index1) const
{
#ifdef CODE_CHECKS
  if(index1<0 || index1>dim1)
    PlainError(1,"Error in Array2D::operator[]: %s",ErrArrOutOfBounds);
#endif // CODE_CHECKS  
  return(row[index1]);
}

template<class T> void Array2D<T>::Swap(Array2D<T> &other)
{ // Interchanges *this & other
  int id1,id2;
  T *tptr,**tpptr;
  id1=dim1;   dim1=other.dim1; other.dim1=id1;
  id2=dim2;   dim2=other.dim2; other.dim2=id2;
  tptr=data;  data=other.data; other.data=tptr;
  tpptr=row;  row=other.row;   other.row=tpptr;
}

/////////////////// Replacement Vec3D Declaration ////////////////////
class Vec3D {
protected:
  double x,y,z;
public:
  Vec3D() { x=y=z=0.; }
  Vec3D(double _x,double _y,double _z) { x=_x; y=_y; z=_z; }
  Vec3D(double *arr) { x=arr[0]; y=arr[1]; z=arr[2]; }
  Vec3D(const ThreeVector &v) { x=v[0]; y=v[1]; z=v[2]; }
  Vec3D& operator=(const ThreeVector &v) {
    x=v[0]; y=v[1]; z=v[2]; return *this;
  }
  inline double GetX() { return x; }
  inline double GetY() { return y; }
  inline double GetZ() { return z; }
  inline void FastAdd(double a,const ThreeVector& v,double b,
		      const ThreeVector& w);      // *this = a.v+b.w 
  inline void FastAdd(const ThreeVector& v,double b,const ThreeVector& w);
                                                  // *this = v+b.w 
  inline void FastAdd(double a,const Vec3D& v,double b,
		      const Vec3D& w);      // *this = a.v+b.w 
  inline void FastAdd(const Vec3D& v,double b,const Vec3D& w);
                                            // *this = v+b.w 
  inline void Accum(double a,const Vec3D& v);
  inline void Accum(double a,const ThreeVector& v);
  inline Vec3D& operator^=(const Vec3D& v);
  inline Vec3D& operator*=(double a) { x*=a; y*=a; z*=a; return *this; }
  inline Vec3D& operator+=(const Vec3D& v) { 
    x+=v.x; y+=v.y; z+=v.z; return *this;
  }
  inline Vec3D& operator+=(const ThreeVector& v) { 
    x+=v[0]; y+=v[1]; z+=v[2]; return *this;
  }
  inline Vec3D& operator-=(const Vec3D& v) {
    x-=v.x; y-=v.y; z-=v.z; return *this;
  }
  inline double MagSq(void) const { return x*x+y*y+z*z; }
  // Dot product
  friend inline double operator*(const Vec3D& u,const Vec3D& v);
  // src -> dest
  friend inline void Convert(ThreeVector& dest,const Vec3D& src);
  friend inline void Convert(Vec3D& dest,const ThreeVector& src);
  void Normalize(); // Adjusts size to 1
  void Random();    // Makes a random unit vector

  // Coordinate transformations.  The 2D grid/solver uses x+z in-plane
  // (width and height), with y pointing into the screen (thickness).
  //  For reference this is called the "xzy" coordinate system.  All
  // outside OOMMF code (including the MIF class) uses x+y in-plane
  // (width and height), with z pointing out of the screen (thickness).
  // This is referred to as the "xyz" coordinate system.  This functions
  // should not be used outside of the Grid2D class.
  void ConvertXyzToXzy() { double temp= y;  y= -z;  z=  temp; }
  void ConvertXzyToXyz() { double temp= y;  y=  z;  z= -temp; }
};

#if VECDIM!=3
#error All lot of this code _assumes_ VECDIM == 3
#endif

inline void Vec3D::FastAdd(double a,const ThreeVector& v,
		    double b,const ThreeVector& w)
{ // *this = a.v+b.w
  x=a*v[0]+b*w[0];  y=a*v[1]+b*w[1];  z=a*v[2]+b*w[2];
}

inline void Vec3D::FastAdd(const ThreeVector& v,
		    double b,const ThreeVector& w)
{ // *this = v+b.w
  x=v[0]+b*w[0];  y=v[1]+b*w[1];  z=v[2]+b*w[2];
}

inline void Vec3D::FastAdd(double a,const Vec3D& v,
		    double b,const Vec3D& w)
{ // *this = a.v+b.w
  x=a*v.x+b*w.x;  y=a*v.y+b*w.y;  z=a*v.z+b*w.z;
}

inline void Vec3D::FastAdd(const Vec3D& v,
		    double b,const Vec3D& w)
{ // *this = v+b.w
  x=v.x+b*w.x;  y=v.y+b*w.y;  z=v.z+b*w.z;
}

inline void Vec3D::Accum(double a,const Vec3D& v)
{ // *this += a.v
  x+=a*v.x; y+=a*v.y; z+=a*v.z;
}

inline void Vec3D::Accum(double a,const ThreeVector& v)
{ // *this += a.v
  x+=a*v[0]; y+=a*v[1]; z+=a*v[2];
}

inline Vec3D& Vec3D::operator^=(const Vec3D& v) // Cross product
{
  double tx,ty,tz;

  tx=x; ty=y; tz=z;  // tz not really needed
  x=ty*v.z-tz*v.y;
  y=tz*v.x-tx*v.z;
  z=tx*v.y-ty*v.x;

  return *this;
}

// Dot product
inline double operator*(const Vec3D& u,const Vec3D& v)
{ return u.x*v.x+u.y*v.y+u.z*v.z; }

inline void Convert(ThreeVector& dest,const Vec3D& src)
{ // src -> dest
  dest[0]=src.x; dest[1]=src.y; dest[2]=src.z;
}

inline void Convert(Vec3D& dest,const ThreeVector& src)
{ // src -> dest
  dest.x=src[0]; dest.y=src[1]; dest.z=src[2];
}

#endif // _VECTOR_H
