/* FILE: zeeman.h           -*-Mode: c++-*-
 *
 * This file contains class declarations for micromagnetic zeeman fields
 *
 */

#ifndef ZEEMAN_H
#define ZEEMAN_H

#include "vector.h"

/* End includes */

class Grid2D;

//////////////////////// Zeeman Declaration //////////////////////////
class Zeeman {
protected:
  int Nx,Ny;  // Grid2D extents
  double Ms;
  double nomhfieldmag;
  double nomhfieldarr[VECDIM];
  ThreeVector nomhfield; // Nominal applied field
public:
  Zeeman(int argc,const char* argp[]);
  virtual ~Zeeman(void);
  static Zeeman* MakeZeeman(int argc,const char* argp[]);
  // static member MakeZeeman is the field type selection routine
  virtual int SetCoords(Grid2D* grid,int imax,int jmax);
  virtual int SetNomField(double newMs,const ThreeVector& newBfield);
  // Note: Some subclasses will ignore SetNomField()
  int SetNomFieldVec3D(double newMs,const Vec3D& newBfield) {
    double arr[3]; ThreeVector vtemp(arr);
    Convert(vtemp,newBfield);
    return SetNomField(newMs,vtemp);
  }
  virtual double GetNomBFieldMag();
  virtual Vec3D  GetNomBField();
  virtual int GetLocalhField(int i,int j,ThreeVector& localfield) const =0;
  virtual double GetLocalhFieldMag(int i,int j) const;
  virtual int GetLocalBField(int i,int j,ThreeVector& localfield) const;
};

// Uniform external field
class UniformZeeman : public Zeeman {
private:
public:
  UniformZeeman(int argc,const char* argp[]);
  UniformZeeman(double Ms,double Bx,double By,double Bz);
  int GetLocalhField(int,int,ThreeVector& localfield) const
     { localfield=nomhfield; return 0; }
  double GetLocalhFieldMag(int,int) const { return nomhfieldmag; }  
};

// Field due to a "ribbon" of charge
// This type is insensitive to changes in the "nominal" field (SetNomField())
class RibbonZeeman : public Zeeman {
private:
  ThreeVector** h;      // h field vector array
  double relcharge;     // Charge magnitude, relative to Ms
  double ribx0,riby0,ribx1,riby1;  // Endpoints of charge ribbon
  double ribheight;                // Height of ribbon
  void FieldCalc(double dpar,double riblength,double dperp,
		 double &mx,double &my);
public:
  RibbonZeeman(int argc,const char* argp[]);
  // argp's are charge & ribbon endpoints:
  //    relcharge ribx0 riby0 ribx1 riby1 ribheight
  // The charge is relative to Ms.
  ~RibbonZeeman(void);
  int SetCoords(Grid2D* grid,int imax,int jmax);
  int GetLocalhField(int i,int j,ThreeVector& localfield) const;
};

// "Fixed" magnetization line
// This type is ignores calls to SetNomField(); hnomfield is setup
// in constructor, and is never changed.
class TieZeeman : public Zeeman {
private:
  ThreeVector** h;      // h field vector array
  double ribx0,riby0,ribx1,riby1; // Endpoints of fixed field
  double ribwidth;  // width of ribon
public:
  TieZeeman(int argc,const char* argp[]);
  // argp's are relative field, ribbon endpoints & ribbon width:
  //        rf0 rf1 rf2 ribx0 riby0 ribx1 riby1 ribwidth
  // The field strength is relative to Ms.
  ~TieZeeman(void);
  int SetCoords(Grid2D* grid,int imax,int jmax);
  int GetLocalhField(int i,int j,ThreeVector& localfield) const;
  int SetNomField(double,const ThreeVector&) { return 0; }
};

// Multiple type element
class MultiZeeman : public Zeeman {
private:
  int Fcount; // Number of fields
  Zeeman** z; // Zeeman field routine pointer array
public:
  MultiZeeman(int argc,const char* argp[]);
  // argp's are number of routines, and then routine names & parameters.
  // Each name is preceded by the number of parameters for that routine.
  ~MultiZeeman(void);
  int SetCoords(Grid2D* grid,int imax,int jmax);
  int SetNomField(double newMs,const ThreeVector& newBfield);
  int GetLocalhField(int i,int j,ThreeVector& localfield) const;
};

#endif // ZEEMAN_H



