# FILE: class.tcl
#
# Defines a command 'em_class' which can be used to define additional
# commands.

# DOME:
#
# undefine should remove support commands -- part of cmdpkg/class 
#	clarifications and cleanup
#
# Background loading
#
# Make 'upvar' and 'uplevel' correct for the extra stack levels
# in method and proc invocations
#
# Migrate em_cmdpkg to an em_class
#
# Permit multiple 'em' versions.
#
# Add 'define' and 'undefine' (better names) to control the
#	initializing and cleanup of commons.
#
# Add boilerplate procs 'about' and 'help'
#	Extract documentation in class compiler step
#
# Perhaps a method to write a reproduction script
#
# error handler - to standardize output and shorten code
#
# factor out common code into utility procs
#
# Automatic update of contents/addenda files. - multiple commands per file.
#
# 'unset $this' can be disasterous.  Prevent?  Warn? Clean-up?
#	Could avoid with different name for command and state array.
#	Would take a lot of tracking - postpone.  Do as part of
#	recoding of em.
#
# Improve compiler
#
# Implement Tcl8.0 version w/namespaces.
#
# Hook to report global litter
#
# Template capability

# Only load em_class once (bootstrap protection)
if {[string match {} [info procs em_class]]} {

proc em_class {name definition} {
  # Lookup the extension and cmdpkg of this class, according to the
  # absolute file name of the file which includes the class definition.
  set pe [_em_cls_Sgetextension]
  set thisfile [file tail [info script]]
  set cp [_em_cls_Sgetcmdpkg $name $pe $thisfile]
  upvar #0 $cp cv
  # Can't break into extension here unless bootstrap is changed.
  set thisfile [file join [$pe directory] $thisfile]
  if {[string match "" $cp]} {
    if {[$pe cget -develop]} {
      error "
'em_class' error:
	Can't evaluate class definition for '$name' 
	in $thisfile .
	There is no corresponding command package in the command
	package database.  

	Did you forget to add a filename to the contents file:
	[file join [$pe directory] [$pe cget -contentsFile]] ?"
    } else {
      error "
'em_class' error:
	Can't evaluate class definition for '$name' 
	in $thisfile .
	There is no corresponding command package in the command
	package database.  This probably means the extension
	'[$pe cget -name]' was not installed properly."
    }
  }

  # Check the name
  if {[string compare $name #auto] && [string compare $name $cv(_command)]} {
    # DOME: Perhaps hack on contents file to resolve conflict?
    error "
'em_class' error:
	Class name mismatch in file 
	$thisfile .

	Call to em_class gives name: $name
	Command package database gives name: $cv(_command)"
  }

  # If the class already exists, don't define and initialize it again.
  # Instead, declare an error -- the auto-loader shouldn't be loading a
  # class which has already been loaded.
  if {![string match {} [info commands $name]]} {
    error "
'em_class' error:
	Can't evaluate class definition for '$name' 
	in $thisfile .

	The class name '$name' conflicts with an existing
	command in this Tcl interpreter (perhaps due to a
	previous evaluation of this class definition?)."
  }

  # Be sure the global array of the command package is in fact global
  # and create alias to it for convenience
  global $cv(_globalarray)
  upvar #0 $cv(_globalarray) cmdvars

  # Prevent conflicts in the global namespace
  if {([info exists cmdvars] || [llength [info commands $cv(_ibase)*]])
      && ![string match em_extension $name]} {
    if {[string match em_array $name] \
        || [string match em_extension $name] \
        || [string match em_log $name] \
        || [string match em_widgetlog $name] \
        || [string match em_channellog $name] } {
      # Allow redefinition of 'em_array' and 'em_extension' as part 
      # of bootstrap.
    } else {
      error "
'em_class' error:
	The global variables and procs to be created by defining the
	class '$name' would conflict with existing global
	variables or procs.

	Definition of class '$name' abandoned."
    }
  }

  # Parse the class definition into a sequence of commands
  if {[catch {_em_cls_Sparseintocommands $definition commands} msg]} {
    error "
'em_class' error:
	Syntax error in '$name' class definition body 
	in $thisfile:

	$msg"
  }
  
  # The command defined by this package.
  proc $name {proc args} "_em_cls_Hclass $cp \$proc \$args"
  
  # For the commands which are truly boilerplate, It makes no sense
  # to manufacture many identical procs.  Instead have each call one
  # implementation.

  # Cache the boiler plate commands so we don't pay the cost of
  # constructing them for every class.
  global _em_cls_boilercommands
  if {[info exists _em_cls_boilercommands]} {
    foreach {bpfx pfx} [list _em_cls_BC $cv(_cmpfx) _em_cls_BI $cv(_impfx) \
        _em_cls_BPC $cv(_pcmpfx) _em_cls_BPI $cv(_pimpfx)] {
      foreach {bc formallist definition} $_em_cls_boilercommands($bpfx) {
        ;proc $pfx$bc $formallist $definition
      }
    }
  } else {
    foreach {bpfx pfx} [list _em_cls_BC $cv(_cmpfx) _em_cls_BI $cv(_impfx) \
        _em_cls_BPC $cv(_pcmpfx) _em_cls_BPI $cv(_pimpfx)] {
      set boilercommands [info commands $bpfx*]
      foreach cmd $boilercommands {
        regexp ^${bpfx}(.+)$ $cmd match bc
        set definition [list $cmd]
        set formallist {}
        set arglist [info args $cmd]
        foreach arg $arglist {
          append definition " \$$arg"
          if {[info default $cmd $arg value]} {
            lappend formallist [list $arg $value]
          } else {
            lappend formallist $arg
          }
        }
        if {[string match args $arg]} {
          set definition [concat eval $definition]
        }
        lappend _em_cls_boilercommands($bpfx) $bc $formallist $definition
        ;proc $pfx$bc $formallist $definition
      }
    }
  }

  # The default constructor (empty)
  ;proc $cv(_hidpfx)constructor {this} {
    return {}
  }

  # The default destructor (empty)
  ;proc $cv(_hidpfx)destructor {this} {}

  # initialize variable lists to be empty
  foreach pat {--- --+ -+- -++ +-- +-+ ++- +++} {
    set cmdvars($pat) {}
  }

  # initialize list of variable and common names
  set cmdvars(#firstname) {{} #lastname}
  set cmdvars(#lastname) {#firstname {}}

  # Evaluate those commands to define the class.  Add the class id (cp).
  foreach cmd $commands {
    set keyword [lindex $cmd 0]
    if {[catch {_em_cls_D$keyword $cp [lrange $cmd 1 end]} msg]} {
      if {![string match {} [info commands _em_cls_D$keyword]]} {
        catch {$name undefine}
        error "
'em_class' error:
	Syntax error detected:

	$msg

	in $thisfile,
	in '$name' class definition body,
	in the following text:

$cmd"
      } else {
        catch {$name undefine}
        error "
'em_class' error:
	Syntax error detected:

	Unknown keyword: $keyword

	in $thisfile,
	in '$name' class definition body,
	in the following text:

$cmd"
      }
    }
  }

  # Possibility of setting values of commons here from values
  # passed in  -- could support template classes!

  # Initialize all the commons in order
  set vcname [lindex $cmdvars(#firstname) 1]
  while {![string match #lastname $vcname]} {
    if {[regexp ^>(.+)$ $vcname match cc]} {
      # $vcname is the name of a common ...
      if {[info exists cmdvars(=$cc)]} {
        # ... for which there is a default value ...
        if {[catch {set cmdvars(_$cc) \
            [_em_cls_HevaluateDefaultValue $cp $cmdvars(=$cc)]} msg]} {
          catch {$name undefine}
          error "
'em_class' error:
        Can't complete definition of class '$name'.
	Error setting common '$cc' to default value '$cmdvars(=$cc)':

	$msg"
        }
      }
    }
    set vcname [lindex $cmdvars($vcname) 1]
  }

  # Check that all const commons are initialized, and use trace
  # to force them to keep their constant value from now on.
  foreach cc [concat $cmdvars(-++) $cmdvars(+++)] {
    if {[info exists cmdvars(_$cc)]} {
      set cmdvars(~$cc) $cmdvars(_$cc)
      trace variable $cv(_globalarray)(_$cc) wu \
          [list _em_cls_Hreadonlycommon $cp $cc]
    } else {
      catch {$name undefine}
      error "
'em_class' error:
	No value given to const common '$cc' in class '$name'"
    }
  }

  # Initialize the hidden class variables
  if {![info exists cmdvars(#nextinstance)]} {
    set cmdvars(#nextinstance) 0
  } else {
    error "
'em_class' error:
	Internal programming error: #nextinstance pre-initialized"
  }
}

#
# Support routines for the em_class command
#

#
# Boiler plate code which is same for all classes
#

;proc _em_cls_Hclass {cp proc arglist} {
  upvar #0 $cp cv
  # First try to eval as a public class proc
  if {[catch {eval $cv(_cmpfx)$proc $cp $arglist} pubresult]} {
    # Either there was no public proc, or it raised an error
    #
    # Check access to private procs
    if {![catch {info level -2} caller] && [regexp \
        ^\($cv(_impfx)|$cv(_pimpfx)|$cv(_cmpfx)|$cv(_pcmpfx)|$cv(_hidpfx)) \
        [lindex $caller 0]]} {
      # We have acccess to private procs
      #
      # Try to eval as a private class proc
      if {[catch {eval $cv(_pcmpfx)$proc $cp $arglist} privresult]} {
        # Either there was no private proc, or it raised an error
        if {[string match {} [info commands $cv(_pcmpfx)$proc]]} {
          # There was no such private class proc
          # Possibly there was a public one (we haven't checked yet).
          if {[string match {} [info commands $cv(_cmpfx)$proc]]} {
            # No such public (or private) proc.
            # At this point we might want to provide for automatic completion
            # of unique proc prefixes for interactive convenience.  For now,
            # just report the error.
            error "
'$cv(_command)' usage:
        $cv(_command) <class_proc> ?args?

Invalid call:
	$cv(_command) $proc $arglist

Invalid class proc: $proc

Valid class procs:
	[$cv(_command) procs * all]"
          } else {
            # The public proc exists, but it raised an error
            # Determine if it was a usage mistake, or an error internal to
            # the public proc body.
            if {[_em_cls_HusageCheck $cv(_cmpfx)$proc [concat $cp $arglist] \
                usageargs]} {
              error "
'$cv(_command) $proc' error:
	$pubresult"
            } else {
              error "
'$cv(_command) $proc' usage:
	$cv(_command) $proc [lrange $usageargs 1 end]

Invalid call:
	$cv(_command) $proc $arglist"
            }
          }
        } else {
          # The private proc exists, but it raised an error
          # Determine if it was a usage mistake, or an error internal to
          # the private proc body.
          if {[_em_cls_HusageCheck $cv(_pcmpfx)$proc [concat $cp $arglist] \
              usageargs]} {
            error "
'$cv(_command) $proc' error:
	$privresult"
          } else {
            error "
'$cv(_command) $proc' usage:
	$cv(_command) $proc [lrange $usageargs 1 end]

Invalid call:
	$cv(_command) $proc $arglist"
          }
        }
      } else {
        # No error in the private proc, return its result
        return $privresult
      }
    } else {
      # No private access.  Determine proper error message for failure
      # of public class proc invocation

      if {[string match {} [info commands $cv(_cmpfx)$proc]]} {
        # No such public proc.
        # At this point we might want to provide for automatic completion
        # of unique proc prefixes for interactive convenience.  For now,
        # just report the error.
        error "
'$cv(_command)' usage:
        $cv(_command) <class_proc> ?args?

Invalid call:
	$cv(_command) $proc $arglist

Invalid class proc: $proc

Valid class procs:
	[$cv(_command) procs * public]"
      } else {
        # The public proc exists, but it raised an error
        # Determine if it was a usage mistake, or an error internal to
        # the public proc body.
        if {[_em_cls_HusageCheck $cv(_cmpfx)$proc [concat $cp $arglist] \
            usageargs]} {
          error "
'$cv(_command) $proc' error:
	$pubresult"
        } else {
          error "
'$cv(_command) $proc' usage:
	$cv(_command) $proc [lrange $usageargs 1 end]

Invalid call:
	$cv(_command) $proc $arglist"
        }
      }
    }
  } else {
    # No error in the public proc, return its result
    return $pubresult
  }
}

;proc _em_cls_Hinstance {this method arglist} {
  upvar #0 $this state
  upvar #0 $state(#cmdpkg) cv
  # First try to eval as a public instance method
  if {[catch {eval $cv(_impfx)$method $this $arglist} pubresult]} {
    # Either there was no public method, or it raised an error
    #
    # Check access to private methods
    if {![catch {info level -2} caller] && [regexp \
        ^\($cv(_impfx)|$cv(_pimpfx)|$cv(_cmpfx)|$cv(_pcmpfx)|$cv(_hidpfx)) \
        [lindex $caller 0]]} {
      # We have acccess to private methods
      #
      # Try to eval as a private instance method
      if {[catch {eval $cv(_pimpfx)$method $this $arglist} privresult]} {
        # Either there was no private method, or it raised an error
        if {[string match {} [info commands $cv(_pimpfx)$method]]} {
          # There was no such private instance method
          # Possibly there was a public one (we haven't checked yet).
          if {[string match {} [info commands $cv(_impfx)$method]]} {
            # No such public (or private) method.
            # At this point we might want to provide for automatic completion
            # of unique method prefixes for interactive convenience.  For now,
            # just report the error.
            error "
'<$cv(_command) instance>' usage:
        \$instance <instance_method> ?args?

Invalid call:
	\$instance $method $arglist

Invalid instance method: $method

Valid instance methods:
	[$cv(_command) methods * all]"
          } else {
            # The public method exists, but it raised an error
            # Determine if it was a usage mistake, or an error internal to
            # the public method body.
            if {[_em_cls_HusageCheck $cv(_impfx)$method [concat $this $arglist] \
                usageargs]} {
              error "
'<$cv(_command) instance> $method' error:
	$pubresult"
            } else {
              error "
'<$cv(_command) instance> $method' usage:
	\$instance $method [lrange $usageargs 1 end]

Invalid call:
	\$instance $method $arglist"
            }
          }
        } else {
          # The private method exists, but it raised an error
          # Determine if it was a usage mistake, or an error internal to
          # the private method body.
          if {[_em_cls_HusageCheck $cv(_pimpfx)$method [concat $this $arglist] \
              usageargs]} {
            error "
'<$cv(_command) instance> $method' error:
	$privresult"
          } else {
            error "
'<$cv(_command) instance> $method' usage:
	\$instance $method [lrange $usageargs 1 end]

Invalid call:
	\$instance $method $arglist"
          }
        }
      } else {
        # No error in the private method, return its result
        return $privresult
      }
    } else {
      # No private access.  Determine proper error message for failure
      # of public instance method invocation
      if {[string match {} [info commands $cv(_impfx)$method]]} {
        # No such public method.
        # At this point we might want to provide for automatic completion
        # of unique method prefixes for interactive convenience.  For now,
        # just report the error.
        error "
'<$cv(_command) instance>' usage:
        \$instance <instance_method> ?args?

Invalid call:
	\$instance $method $arglist

Invalid instance method: $method

Valid instance methods:
	[$cv(_command) methods * public]"
      } else {
        # The public method exists, but it raised an error
        # Determine if it was a usage mistake, or an error internal to
        # the public method body.
        if {[_em_cls_HusageCheck $cv(_impfx)$method [concat $this $arglist] \
            usageargs]} {
          error "
'<$cv(_command) instance> $method' error:
	$pubresult"
        } else {
          error "
'<$cv(_command) instance> $method' usage:
	\$instance $method [lrange $usageargs 1 end]

Invalid call:
	\$instance $method $arglist"
        }
      }
    }
  } else {
    # No error in the public method, return its result
    return $pubresult
  }
}

;proc _em_cls_BPCglobalname {cp varname} {
  upvar #0 $cp cv
  upvar #0 $cv(_globalarray) cmdvars
  # Search through all the commons for any named $varname
  foreach pat {+-+ +++} {
    if {[lsearch -exact $cmdvars($pat) $varname] >= 0} {
      return $cv(_globalarray)(_$varname)
    }
  }
  error "'$varname' is not the name of a common in the
	definition of the class '$cv(_command)'."
}

# Perhaps do something more sophisticated here?
#	Like... actually maintain a list.
#       This implementation depends on adherence to the naming conventions
;proc _em_cls_BCinstances {cp} {
  upvar #0 $cp cv
  set cand [info commands $cv(_ibase)*]
  set ret {}
  foreach i $cand {
    if {[regexp $cv(_ibase)\[0-9\]+ $i]} {
      lappend ret $i
    }
  }
  return $ret
}

;proc _em_cls_BCmethods {cp {pattern *} {access callable}} {
  # Depends on naming conventions.  Should migrate to actual maintenance
  # Break into cmdpkg
  upvar #0 $cp cv
  set publicmethods {}
  foreach rawcmd [info commands $cv(_impfx)$pattern] {
    regexp ^$cv(_impfx)(.+)$ $rawcmd match im
    lappend publicmethods $im
  }
  set privatemethods {}
  foreach rawcmd [info commands $cv(_pimpfx)$pattern] {
    regexp ^$cv(_pimpfx)(.+)$ $rawcmd match im
    lappend privatemethods $im
  }
  switch -- $access {
    public {
      return [lsort $publicmethods]
    }
    private {
      return [lsort $privatemethods]
    }
    all {
      return [lsort [concat $privatemethods $publicmethods]]
    }
    callable {
      if {![catch {info level -4} caller] && [regexp \
          ^\($cv(_impfx)|$cv(_pimpfx)|$cv(_cmpfx)|$cv(_pcmpfx)|$cv(_hidpfx)) \
          [lindex $caller 0]]} {
        return [lsort [concat $privatemethods $publicmethods]]
      } else {
        return [lsort $publicmethods]
      }
    }
    default {
      error "Unrecognized value '$access' for 'access' argument."
    }
  }
}

;proc _em_cls_BCprocs {cp {pattern *} {access callable}} {
  # Depends on naming conventions.  Should migrate to actual maintenance
  # Break into cmdpkg
  upvar #0 $cp cv
  set publicprocs {}
  foreach rawcmd [info commands $cv(_cmpfx)$pattern] {
    regexp ^$cv(_cmpfx)(.+)$ $rawcmd match cm
    lappend publicprocs $cm
  }
  set privateprocs {}
  foreach rawcmd [info commands $cv(_pcmpfx)$pattern] {
    regexp ^$cv(_pcmpfx)(.+)$ $rawcmd match cm
    lappend privateprocs $cm
  }
  switch -- $access {
    public {
      return [lsort $publicprocs]
    }
    private {
      return [lsort $privateprocs]
    }
    all {
      return [lsort [concat $privateprocs $publicprocs]]
    }
    callable {
      if {![catch {info level -4} caller] && [regexp \
          ^\($cv(_impfx)|$cv(_pimpfx)|$cv(_cmpfx)|$cv(_pcmpfx)|$cv(_hidpfx)) \
          [lindex $caller 0]]} {
        return [lsort [concat $privateprocs $publicprocs]]
      } else {
        return [lsort $publicprocs]
      }
    }
    default {
      error "Unrecognized value '$access' for 'access' argument."
    }
  }
}

;proc _em_cls_BCundefine {cp} {
  $cp undefine
}

;proc _em_cls_BCextension {cp} {
  $cp keys @extension
}

# Class method which creates a new instance.
#	Calls on the constructor for any special instructions
;proc _em_cls_BCnew {cp args} {
  # Break into cmdpkg for efficiency.  Use...
  upvar #0 $cp cv
  upvar #0 $cv(_globalarray) cmdvars
  # ...instead of
  # upvar #0 [$cp globalarray] cmdvars
  # Make a global array to hold the instance state.
  # Break into cmdpkg for efficiency.  Use...
  set this $cv(_ibase)$cmdvars(#nextinstance)
  # ...instead of
  # set this [$cp ibase]$cmdvars(#nextinstance)
  incr cmdvars(#nextinstance)
  global $this
  upvar #0 $this state
  if {[info exists $this]} {
    error "Internal inconsistency.

	Attempt to create two instances with same name: $this

	Second instance not created."
  }
 
  # From here on, any exits from this routine need to call 
  # '$this delete' to clean-up half-constructed instances,
  # in particular to remove the proc defined next.

  # Define the instance command
  proc $this {method args} "_em_cls_Hinstance $this \$method \$args"

  # save command package info
  array set state "#cmdpkg $cp"

  # Initialize all the variables in order
  # Not really necessary.  Could reduce overhead by stripping out list
  # structure.
  set vcname [lindex $cmdvars(#firstname) 1]
  while {![string match #lastname $vcname]} {
    if {[regexp ^<(.+)$ $vcname match vn]} {
      # $vcname is the name of a variable ...
      if {[info exists cmdvars(=$vn)]} {
        # ... for which there is a default value ...
        set state(_$vn) $cmdvars(=$vn)
# Don't run validation scripts on default values.  Let programmer have
# responsibility of making default value sensible (and the freedom of
# not making it sensible).
#       if {![string match {} [info commands $cv(_hidpfx)@$vn]] \
#          && [catch {$cv(_hidpfx)@$vn $this} emsg]} {
#         error "Error setting variable '$vn' to default value '$cmdvars(=$vn)':
#
#	$emsg"
#       }
      }
    }
    set vcname [lindex $cmdvars($vcname) 1]
  }

  # Try to evaluate the constructor
  if {[catch {eval $cv(_hidpfx)constructor $this $args} msg]} {
    # The constructor failed.
    # We don't want to leave half-constructed instances lying around,
    # so delete the instance before returning the error

    if {[catch {$this delete} deletereturn]} {
      # The destructor also failed.  Save its error message for reporting
      # later.
      set deletemsg "
	An attempt was made to delete the new $cv(_command) instance.

$deletereturn"
    } else {
      # The destructor completed normally.  Save a message to report that.
      set deletemsg "
	The new $cv(_command) instance was deleted."
    }
    # Determine what error message to report.
    if {[_em_cls_HusageCheck $cv(_hidpfx)constructor [concat $this $args] \
      usageargs]} {
      # Usage was OK.  Report error in the constructor.
      error "Error in $cv(_command) constructor:
	$msg
	$deletemsg"
    } else {
      # report usage error
      error "
'$cv(_command) new' usage:
	$cv(_command) new [lrange $usageargs 1 end]

Invalid call:
	$cv(_command) new $args
	$deletemsg"
    }
  }

  # The variable 'msg' holds the return value from the constructor
  #
  # If the constructor doesn't return a value (that is, it returns the
  # default empty string), then 'new' should return the value of the
  # 'this' pointer it's been building up.  This is the usual case.
  #
  # However, if the constructor returns a value, then 'new' will return
  # that value as the value of the 'this' pointer.  This mechanism allows
  # constructors to be written so that a call to 'new' doesn't return a
  # pointer to a new instance, but a pointer to an existing instance.
  # This is useful for implementing classes in which the objects have
  # identity, and there should be no identical copies floating around.
  # Of course, in these cases, there's a (possibly half-constructed)
  # instance pointed to by 'this' which must be disposed of.  That
  # task is left to the constructor.  It creates the need for cleanup,
  # so it is stuck with the burden of that cleanup.
  #
  if {![string match {} $msg]} {
    return $msg
  } else {
    # Check that all const variables are initialized, and use trace
    # to force them to keep their constant value from now on.
    foreach ccv [concat $cmdvars(-+-) $cmdvars(++-)] {
      if {[info exists state(_$ccv)]} {
        set state(~$ccv) $state(_$ccv)
        trace variable ${this}(_$ccv) wu \
            [list _em_cls_Hreadonlyvariable $this $ccv]
      } else {
        catch {$this delete}
        error "Constructor completed without setting const variable '$ccv'

	The new $cv(_command) instance was deleted."
      }
    }
    return $this
  }
}

;proc _em_cls_BPIglobalname {this varname} {
  upvar #0 $this state
  upvar #0 $state(#cmdpkg) cv
  upvar #0 $cv(_globalarray) cmdvars
  # Search through all the variable for any named $varname
  foreach pat {--- -+- +-- ++-} {
    if {[lsearch -exact $cmdvars($pat) $varname] >= 0} {
      return ${this}(_$varname)
    }
  }
  if {[catch {[$this class] globalname $varname} result]} {
    error "'$varname' is not the name of a variable or common in the
	definition of the class '$cv(_command)'."
  } else {
    return $result
  }
}

;proc _em_cls_BIclass {this} {
  upvar #0 $this state
  upvar #0 $state(#cmdpkg) cv
  return $cv(_command)
}

;proc _em_cls_BIconfigure {this args} {
# Given no arguments, return list of triples, one for each public variable.
# 	1st element:	-$varname
#	2nd element:	<default value>
#	3rd element:	<current value>
#
# Given one argument, it must be of the form -$varname, where varname is
# a public variable.  If not, return an error.  If so, return the triple 
# for that varname.
#
# Given more than one argument, it must be a list of option value pairs
# For each pair, if the variable is public, set that variable to
# the given value, and run the configuration script associated with the
# variable.  Const variable may only be set by configure when the call is
# made within the constructor.
  upvar #0 $this state
  set cp $state(#cmdpkg)
  # Break into cmdpkg.
  upvar #0 $cp cv
  upvar #0 $cv(_globalarray) cmdvars
  # upvar #0 [$cp globalarray] cmdvars

  set numargs [llength $args]
  if {!$numargs} {
    set retval {}
    foreach pv [lsort [concat $cmdvars(---) $cmdvars(-+-)]] {
      set triple -$pv
      if {[info exists cmdvars(=$pv)]} {
        lappend triple $cmdvars(=$pv)
      } else {
        lappend triple <undefined>
      }
      if {[info exists state(_$pv)]} {
        lappend triple $state(_$pv)
      } else {
        lappend triple <undefined>
      }
      lappend retval $triple
    }
    return $retval
  }
  if {$numargs == 1} {
    if {[string match - [string index $args 0]]} {
      set pv [string range $args 1 end]
      if {[lsearch -exact [concat $cmdvars(---) $cmdvars(-+-)] $pv] >= 0} {
        set triple -$pv
        if {[info exists cmdvars(=$pv)]} {
          lappend triple $cmdvars(=$pv)
        } else {
          lappend triple <undefined>
        }
        if {[info exists state(_$pv)]} {
          lappend triple $state(_$pv)
        } else {
          lappend triple <undefined>
        }
        return $triple
      }
    }
    set ncopts {}
    foreach ncopt [lsort $cmdvars(---)] {
      lappend ncopts -$ncopt
    }
    set copts {}
    foreach copt [lsort $cmdvars(-+-)] {
      lappend copts -$copt
    }
    error "Invalid option: $args

usage:
	\$instance configure
	\$instance configure <option>
	\$instance configure ?<option> <value> ...?

	The last form for setting public variable values may be used to set
	const public variables only when called from a class constructor.

Invalid call:
	\$instance configure $args

Const options: $copts
Non-const options: $ncopts"
  }
  # There are more than one argument
  #
  # Use a foreach construction so that the order of setting values is
  # given by the order of the argument list.
  set index 0
  foreach {opt val} $args {
    incr index 2
    # Check option validity: Start with -
    if {[string match - [string index $opt 0]]} {
      set pv [string range $opt 1 end]
      # Check that we have a value to assign
      if {$index <= $numargs} {
        # Check option validity: Refer to public variable
        # Case 1: non-const variable -- can always modify.
        if {[lsearch -exact $cmdvars(---) $pv] >= 0} {
          # Save any old value in case something goes wrong
          if {[info exists state(_$pv)]} {
            set oldval $state(_$pv)
          } else {
            catch {unset oldval}
          }
          # Set the value.  
          set state(_$pv) $val
          # Run the configuration script, if any.
          if {![string match {} [info procs $cv(_hidpfx)@$pv]] \
              && [catch {$cv(_hidpfx)@$pv $this} msg]} {
            # The configuration script raised an error.
            # Restore the former value of the variable...
            if {[info exists oldval]} {
              set state(_$pv) $oldval
            } else {
              catch {unset state(_$pv)}
            }
            # ... and report the error.
            error "Error setting variable '$pv' to value '$val':

	$msg"
          } else {
            # Successfully set public non-const variable to new value.
            # Move on to the next option.
            continue
          }

        # Case 2: const variable -- only modify when called from constructor.
        } elseif {[lsearch -exact $cmdvars(-+-) $pv] >= 0} {
          # Check that we're called from the constructor.  A direct call 
          # from the constructor is probably wisest, but there's no 
          # requirement to be that restrictive.  By controlling access 
          # based on whether the readonly trace has been activated, we can 
          # allow the configure method to initialize public const variables
          # anytime between the creation of the instance command (early in 
          # the execution of the 'new' class method) and the setting of
          # readonly traces (late in the 'new' class method).
          # The only user code that executes in that interval is the 
          # constructor (and anything it calls).
          #
          # CAUTION:  Any command in the constructor which turns over
          # control to the event loop could cause trouble.  The
          # entire 'new' method: internal housekeeping 
          # plus constructor should be one non-interrupted operation.
          if {![llength [trace vinfo state(_$pv)]]} {
            # Save any old value in case something goes wrong
            if {[info exists state(_$pv)]} {
              set oldval $state(_$pv)
            } else {
              catch {unset oldval}
            }
            # Set the value.  
            set state(_$pv) $val
            # Run the configuration script, if any.
            if {![string match {} [info procs $cv(_hidpfx)@$pv]] \
                && [catch {$cv(_hidpfx)@$pv $this} msg]} {
              # The configuration script raised an error.
              # Restore the former value of the variable...
              if {[info exists oldval]} {
                set state(_$pv) $oldval
              } else {
                catch {unset state(_$pv)}
              }
              # ... and report the error.
              error "Error initializing const variable '$pv' to value '$val':

	$msg"
            } else {
              # Successfully set public const variable to new value.
              # Move on to the next option.
              continue
            }
          } else {
            set msg "Const option '$opt' outside the constructor"
          }

        # Case 3: Not a public variable at all -- ERROR!
        } else {
          set msg "Invalid option: '$opt'"
        }
      } else {
        set msg "Value for option '$opt' missing"
      }
    } else {
      set msg "Invalid option: '$opt'"
    }
    set ncopts {}
    foreach ncopt [lsort $cmdvars(---)] {
      lappend ncopts -$ncopt
    }
    set copts {}
    foreach copt [lsort $cmdvars(-+-)] {
      lappend copts -$copt
    }
    error "$msg

usage:
	\$instance configure
	\$instance configure <option>
	\$instance configure ?<option> <value> ...?

	The last form for setting public variable values may be used to set
	const public variables only when called from a class constructor.

Invalid call:
	\$instance configure $args

Const options: $copts
Non-const options: $ncopts"
  }
  return {}
}

;proc _em_cls_BIcget {this args} {
  upvar #0 $this state
  upvar #0 $state(#cmdpkg) cv
  upvar #0 $cv(_globalarray) cmdvars
  if {[llength $args] == 1} {
    if {[string match - [string index $args 0]]} {
      set pv [string range $args 1 end]
      if {[lsearch -exact [concat $cmdvars(---) $cmdvars(-+-)] $pv] >= 0} {
        if {[info exists state(_$pv)]} {
          return $state(_$pv)
        } else {
          return <undefined>
        }
      }
    }
    set validopts {}
    foreach opt [lsort [concat $cmdvars(---) $cmdvars(-+-)]] {
      lappend validopts -$opt
    } 
    error "Invalid option: $args

usage:
	\$instance cget <option>

Invalid call:
	\$instance cget $args

Valid options: $validopts"
  }
  error "\$instance cget <option>

Invalid call:
	\$instance cget $args"
}

;proc _em_cls_BIdelete {this} {
  upvar #0 $this state
  upvar #0 $state(#cmdpkg) cv
  set destructorerror 0
  if {[catch {$cv(_hidpfx)destructor $this} msg]} {
    # Some error occurred trying to evaluate the destructor.
    # The instance might be in some half-destructed state.
    # Go ahead and destroy the parts of the instance which the
    # boiler plate code is responsible for
    set destructorerror 1
  }
  # Try to complete destruction, error or not.
  catch {rename $this {}}
  catch {unset state}
  # Then, if there was an error, report it.
  if {$destructorerror} {
    error "Error in $cv(_command) destructor:
	$msg

	Depending on the error, some portions of the destructor may not
	have been evaluated, and some internal data within the instance
	may not have been cleared.  A best effort was made to complete the
	instance deletion, and instance methods may no longer be invoked
	on that instance."
  }
}

#
# Routines which build class structure from class definition
#

;proc _em_cls_Dprivate {cp arglist} {
  set cmd [lindex $arglist 0]
  if {[catch {_em_cls_G$cmd +? $cp [lrange $arglist 1 end]} msg]} {
    if {[string match {} [info commands _em_cls_G$cmd]]} {
      error "'private' keyword cannot be applied to '$cmd'"
    } else {
      error $msg
    }
  }
}

;proc _em_cls_Dpublic {cp arglist} {
  set cmd [lindex $arglist 0]
  if {[catch {_em_cls_G$cmd -? $cp [lrange $arglist 1 end]} msg]} {
    if {[string match {} [info commands _em_cls_G$cmd]]} {
      error "'public' keyword cannot be applied to '$cmd'"
    } else {
      error $msg
    }
  }
}

;proc _em_cls_Dconst {cp arglist} {
  set cmd [lindex $arglist 0]
  if {[catch {_em_cls_G$cmd ?+ $cp [lrange $arglist 1 end]} msg]} {
    if {[string match {} [info commands _em_cls_G$cmd]]} {
      error "'const' keyword cannot be applied to '$cmd'"
    } else {
      error $msg
    }
  }
}

# Variables are private by default
;proc _em_cls_Dvariable {cp rgs} {
  _em_cls_Gvariable +- $cp $rgs
}

# Commons are private by default
;proc _em_cls_Dcommon {cp rgs} {
  _em_cls_Gcommon +- $cp $rgs
}

# Methods are public by default
;proc _em_cls_Dmethod {cp rgs} {
  _em_cls_Gmethod -? $cp $rgs
}

# Procs are public by default
;proc _em_cls_Dproc {cp rgs} {
  _em_cls_Gproc -? $cp $rgs
}

;proc _em_cls_Dconstructor {cp rgs} {
  if {[llength $rgs] == 2} {
    # Possibly very slightly more efficient.  Use...
    # foreach {arglist body} $args {}
    foreach {arglist body} $rgs {}
    # ...instead of
    # set arglist [lindex $args 0]
    # set body [lindex $args 1]
  } else {
    error "constructor usage:

	constructor <argument list> <body>"
  }
  # Break into command package for efficiency.  Use ...
  upvar #0 $cp cv
  set realname $cv(_hidpfx)constructor
  # ...instead of
  # set realname [$cp hidpfx]constructor
  set realarglist [concat this $arglist]
  global _em_class
  set realbody $_em_class(methodPreamble)
  append realbody $body
  # Be sure constructor returns a value
  append realbody {return {}}
  ;proc $realname $realarglist $realbody
}

;proc _em_cls_Ddestructor {cp rgs} {
  if {[llength $rgs] == 1} {
    set body [lindex $rgs 0]
  } else {
    error "destructor usage:

	destructor <body>"
  }
  # Break into command package for efficiency.  Use ...
  upvar #0 $cp cv
  set realname $cv(_hidpfx)destructor
  # ...instead of
  # set realname [$cp hidpfx]destructor
  set realarglist this
  global _em_class
  set realbody $_em_class(methodPreamble)
  append realbody $body
  ;proc $realname $realarglist $realbody
}

;proc _em_cls_Gprivate {qual cp rgs} {
  set cmd [lindex $rgs 0]
  set newqual +[string index $qual 1]
  if {[catch {_em_cls_G$cmd $newqual $cp [lrange $rgs 1 end]} msg]} {
    if {[string match {} [info commands _em_cls_G$cmd]]} {
      error "'private' keyword cannot be applied to '$cmd'"
    } else {
      error $msg
    }
  }
}

;proc _em_cls_Gpublic {qual cp rgs} {
  set cmd [lindex $rgs 0]
  set newqual -[string index $qual 1]
  if {[catch {_em_cls_G$cmd $newqual $cp [lrange $rgs 1 end]} msg]} {
    if {[string match {} [info commands _em_cls_G$cmd]]} {
      error "'public' keyword cannot be applied to '$cmd'"
    } else {
      error $msg
    }
  }
}

;proc _em_cls_Gconst {qual cp rgs} {
  set cmd [lindex $rgs 0]
  set newqual [string index $qual 0]+
  if {[catch {_em_cls_G$cmd $newqual $cp [lrange $rgs 1 end]} msg]} {
    if {[string match {} [info commands _em_cls_G$cmd]]} {
      error "'const' keyword cannot be applied to '$cmd'"
    } else {
      error $msg
    }
  }
}

;proc _em_cls_Gvariable {qual cp rgs} {
  set numargs [llength $rgs]
  if {$numargs == 1} {
    foreach name $rgs {}
  } elseif {$numargs == 2} {
    foreach {name script} $rgs {}
  } elseif {$numargs == 3} {
    foreach {name equalsign default} $rgs {}
  } elseif {$numargs == 4} {
    foreach {name equalsign default script} $rgs {}
  } else {
    error "variable usage:

	?<modifiers>? variable <name> ?= <default_value>? ?<validation_script>?"
  }
  if {[info exists equalsign] && [string compare = $equalsign]} {
    error "variable usage:

	?<modifiers>? variable <name> ?= <default_value>? ?<validation_script>?"
  }
  if {[string match {\?} [string index $qual 0]]} {
    set qual +[string index $qual 1]
  }
  if {[string match {\?} [string index $qual 1]]} {
    set qual [string index $qual 0]-
  }
  # Break into cmdpkg for efficiency
  upvar #0 $cp cv
  upvar #0 $cv(_globalarray) cmdvars
  _em_cls_Sclearvar $cp $name
  lappend cmdvars($qual-) $name
  set pred [lindex $cmdvars(#lastname) 0]
  set cmdvars(<$name) [list $pred #lastname]
  set cmdvars($pred) [lreplace $cmdvars($pred) 1 1 <$name]
  set cmdvars(#lastname) [lreplace $cmdvars(#lastname) 0 0 <$name]
  # Default value
  if {[info exists default]} {
    set cmdvars(=$name) $default
  }
  # Validation script
  if {[info exists script]} {
    if {[string match - [string index $qual 0]]} {
      # Break into cmdpkg for efficiency
      set realname $cv(_hidpfx)@$name
      set realarglist this
      global _em_class
      set realbody $_em_class(methodPreamble)
      append realbody $script
      ;proc $realname $realarglist $realbody
    } else {
      error "Error in declaration of variable '$name':

	Only public variables take validation scripts"
    }
  }
}

;proc _em_cls_Gcommon {qual cp rgs} {
  set numargs [llength $rgs]
  if {!$numargs} {
    error "Missing name in common declaration"
  }
  set name [lindex $rgs 0]
  if {$numargs > 2} {
    error "Too many arguments in declaration of common '$name'"
  } 
  if {[string match - [string index $qual 0]]} {
    error "'public' keyword cannot be applied to 'common'"
  }
  if {[string match {\?} [string index $qual 0]]} {
    set qual +[string index $qual 1]
  }
  if {[string match {\?} [string index $qual 1]]} {
    set qual [string index $qual 0]-
  }
  # Break into cmdpkg for efficiency. Use ..
  upvar #0 $cp cv
  upvar #0 $cv(_globalarray) cmdvars
  # ... instead of
  # upvar #0 [$cp globalarray] cmdvars
  _em_cls_Sclearvar $cp $name
  lappend cmdvars($qual+) $name
  set pred [lindex $cmdvars(#lastname) 0]
  set cmdvars(>$name) [list $pred #lastname]
  set cmdvars($pred) [lreplace $cmdvars($pred) 1 1 >$name]
  set cmdvars(#lastname) [lreplace $cmdvars(#lastname) 0 0 >$name]
  if {$numargs > 1} {
    set cmdvars(=$name) [lindex $rgs 1]
  }
}

# The next two routines are mini-compilers
;proc _em_cls_Gmethod {qual cp rgs} {
  set numargs [llength $rgs]
  if {$numargs == 3} {
    foreach {name arglist body} $rgs {}
  } elseif {$numargs == 4} {
    foreach {name arglist varlist body} $rgs {}
  } else {
    error "method usage:

	?<access>? method <name> <argument list> ?<variable list>? <body>"
  }
  if {[string match + [string index $qual 1]]} {
    error "'const' keyword cannot be applied to 'method'"
  }
  # methods are public by default
  if {[string match {\?} [string index $qual 0]]} {
    set qual -[string index $qual 1]
  }
  # select prefix to enforce access restrictions
  # Break into cmdpkg for efficiency
  upvar #0 $cp cv
  if {[string match - [string index $qual 0]]} {
    set realname $cv(_impfx)$name
  } else {
    set realname $cv(_pimpfx)$name
  }
  set realarglist [concat this $arglist]
  if {[info exists varlist]} {
    set realbody {
      global $this 
      set _ [set ${this}(#cmdpkg)]
      global $_
      set class [set ${_}(_command)]
    }
    # Break into cmdpkg for efficiency
    set cpga $cv(_globalarray)
    upvar #0 $cpga ccv
    set seencommon 0
    foreach varname $varlist {
      if {[catch {set ccv(<$varname)}]} {
        if {[catch {set ccv(>$varname)}]} {
          error "'$varname' in variable list is not a variable or common name"
        } else {
          if {!$seencommon} {
            append realbody "global $cpga\n"
            set seencommon 1
          }
          append realbody "upvar #0 ${cpga}(_$varname) $varname\n"
        }
      } else {
        append realbody "upvar #0 \${this}(_$varname) $varname\n"
      }
    }
    append realbody "catch {unset _}\n"
  } else {
    global _em_class
    set realbody $_em_class(methodPreamble)
  }
  append realbody $body
  catch {rename $cv(_impfx)$name {}}
  catch {rename $cv(_pimpfx)$name {}}
  ;proc $realname $realarglist $realbody
}

;proc _em_cls_Gproc {qual cp rgs} {
  set numargs [llength $rgs]
  if {$numargs == 3} {
    foreach {name arglist body} $rgs {}
  } elseif {$numargs == 4} {
    foreach {name arglist varlist body} $rgs {}
  } else {
    error "proc usage:

	?<access>? proc <name> <argument list> ?<variable list>? <body>"
  }
  if {[string match + [string index $qual 1]]} {
    error "'const' keyword cannot be applied to 'proc'"
  }
  if {[string match {\?} [string index $qual 0]]} {
    set qual -[string index $qual 1]
  }
  # select prefix to enforce access restrictions
  # Break into cmdpkg for efficiency
  upvar #0 $cp cv
  if {[string match - [string index $qual 0]]} {
    set realname $cv(_cmpfx)$name
  } else {
    set realname $cv(_pcmpfx)$name
  }
  # prepend formal argument '_', meaning 'cp', to argument list
  set realarglist [concat _ $arglist]
  if {[info exists varlist]} {
    set realbody {
      global $_
      set class [set ${_}(_command)]
    }
    set cpga $cv(_globalarray)
    upvar #0 $cpga ccv
    set seencommon 0
    foreach varname $varlist {
      if {[catch {set ccv(>$varname)}]} {
        error "'$varname' in variable list is not a common name"
      } else {
        if {!$seencommon} {
          append realbody "global $cpga\n"
          set seencommon 1
        }
        append realbody "upvar #0 ${cpga}(_$varname) $varname\n"
      }
    }
    append realbody "catch {unset _}\n"
  } else {
    global _em_class
    set realbody $_em_class(procPreamble)
  }
  append realbody $body
  catch {rename $cv(_cmpfx)$name {}}
  catch {rename $cv(_pcmpfx)$name {}}
  ;proc $realname $realarglist $realbody
}

;proc _em_cls_Sclearvar {cp varname} {
  upvar #0 $cp cv
  upvar #0 $cv(_globalarray) cmdvars
  if {[catch {set cmdvars(<$varname)}]} {
    # No variable by that name
    if {[catch {set cmdvars(>$varname)}]} {
      # No common by that name
      return
    } else {
      foreach {pred succ} $cmdvars(>$varname) {}
      set cmdvars($pred) [lreplace $cmdvars($pred) 1 1 $succ]
      set cmdvars($succ) [lreplace $cmdvars($succ) 0 0 $pred]
      unset cmdvars(>$varname)
      foreach pat {+-+ +++} {
        set i [lsearch -exact $cmdvars($pat) $varname]
        if {$i >= 0} {
          set cmdvars($pat) [lreplace $cmdvars($pat) $i $i]
          # IF the rest of the em_class code is bug-free,
          # each variable name is in at most one list.  Once we've
          # found and cleared it, we can end our search
          break
        }
      }
    }
  } else {
    foreach {pred succ} $cmdvars(<$varname) {}
    set cmdvars($pred) [lreplace $cmdvars($pred) 1 1 $succ]
    set cmdvars($succ) [lreplace $cmdvars($succ) 0 0 $pred]
    unset cmdvars(<$varname)
    foreach pat {--- -+- +-- ++-} {
      set i [lsearch -exact $cmdvars($pat) $varname]
      if {$i >= 0} {
        set cmdvars($pat) [lreplace $cmdvars($pat) $i $i]
        # IF the rest of the em_class code is bug-free,
        # each variable name is in at most one list.  Once we've
        # found and cleared it, we can end our search
        break
      }
    }
  }
  catch {unset cmdvars(=$name)}
  catch {rename $cv(_hidpfx)@$varname {}}
}

;proc _em_cls_Sparseintocommands {body resultName} {
  # Try to replace this with a call to 'eval' for efficiency

  # parse the body into a sequence of commands
  upvar $resultName commands

  set commands {}
  set chunk ""
  foreach line [split $body "\n"] {
    append chunk $line
    if {[info complete $chunk]} {
      # $chunk ends in a complete Tcl command, and none of the
      # newlines within it end a complete Tcl command.  If there
      # are multiple Tcl commands in $chunk, they must be separated
      # by semi-colons.
      set cmd ""
      foreach part [split $chunk ";"] {
        append cmd $part
        if {[info complete $cmd]} {
          set cmd [string trimleft $cmd]
          # Drop empty commands and comments
          if {![string match {} $cmd] && ![string match #* $cmd]} {
            lappend commands $cmd
          }
          if {[string match #* $cmd]} {
              set cmd "#;"
          } else {
              set cmd ""
          }
        } else {
          # No complete command yet.  Put the semicolon back and continue
          append cmd ";"
        }
      }
      set chunk ""
    } else {
      # No end of command yet.  Put the newline back and continue
      append chunk "\n"
    }
  }
  if {![string match {} [string trimright $chunk]]} {
    error "Can't interpret body as a sequence of commands.

	Incomplete command: $chunk"
  }
}

;proc _em_cls_HevaluateDefaultValue {@0 @2} {
  # Argument '@2' holds the default value for a class member of type ('common' 
  # or 'variable') held by the argument '@1'.  This proc returns the 
  # value to actually set the member to.  This is needed because '@2' may 
  # take the form of a script which needs to be evaluated, and we want 
  # the same access to class members within that default value script as 
  # is available within method or proc bodies, as appropriate.  Hence, the 
  # extra stack level to insulate all the extra references we have to 
  # create from the surrounding code in procs which call this one.
  #
  # The strange variable and argument names are also needed so we don't
  # clash with variable names in the default value script.
  #
  # BTW, this is quite wasteful for the simple case when '@2' just 
  # holds a value.  A better mechanism should be sought.
  upvar #0 _em_class @
  set _ ${@0}
  eval [set @(procPreamble)]
  string trim [subst ${@2}]
}

;proc _em_cls_HusageCheck {cmd testarglist usageargsName} {
  upvar $usageargsName usageargs
  # Check that the list $testarglist is suitable for
  # passing as the arguments to $cmd.
  set arglist [info args $cmd]

  # Determine the minimum number of arguments required to make
  # $cmd happy.  Also construct a usage string.
  set minargs 0
  set usageargs {}
  set done 0
  foreach a $arglist {
    if {(($minargs == ([llength $arglist] - 1)) && [string match args $a])} {
      # Last argument named 'args'
      lappend usageargs args
    } elseif {[info default $cmd $a dummy]} {
      lappend usageargs "?$a?"
      set done 1
    } else {
      lappend usageargs "$a"
      if {$done} {
        error "Internal programming error:
	Malformed formal argument list for command '$cmd'"
      } else {
        incr minargs
      }
    }
  }
  # Usage problem if too many arguments (more given than requested,
  # unless last one requested is 'args') or if too few arguments.
  if {(([llength $testarglist] > [llength $arglist]) 
      && ([string compare args [lindex $arglist end]]))
      || ([llength $testarglist] < $minargs)} {
    return 0
  }
  return 1
}

;proc _em_cls_Hreadonlyvariable {this varname name1 name2 ops} {
  global $this
  if {[array exists $this]} {
    set ${this}(_$varname) [set ${this}(~$varname)]
    if {[string match u $ops]} {
      trace variable ${this}(_$varname) wu \
          [list _em_cls_Hreadonlyvariable $this $varname]
    }
    error "
	Attempt to set const variable '$varname'."
  }
}

;proc _em_cls_Hreadonlycommon {cp cname name1 name2 ops} {
  upvar #0 $cp cv
  global $cv(_globalarray)
  if {[array exists $cv(_globalarray)]} {
    set $cv(globalarry)(_$cname) [set $cv(_globalarray)(~$cname)]
    if {[string match u $ops]} {
      trace variable $cv(_globalarray)(_$cname) wu \
          [list _em_cls_Hreadonlycommon $cp $cname]
    }
    error "
	Attempt to set const common '$cname'."
  }
}

global _em_class
array set _em_class {
methodPreamble {
  global $this 
  set _ [set ${this}(#cmdpkg)]
  global $_
  set class [set ${_}(_command)]
  set = [set ${_}(_globalarray)]
  global ${=}
  foreach + [concat [set ${=}(---)] [set ${=}(-+-)] \
                    [set ${=}(+--)] [set ${=}(++-)]] {
    upvar #0 ${this}(_${+}) ${+}
  }
  foreach + [concat [set ${=}(--+)] [set ${=}(-++)] \
                    [set ${=}(+-+)] [set ${=}(+++)]] {
    upvar #0 ${=}(_${+}) ${+}
  }
  catch {unset _ = +}
}
procPreamble {
  global $_
  set class [set ${_}(_command)]
  set = [set ${_}(_globalarray)]
  global ${=}
  foreach + [concat [set ${=}(--+)] [set ${=}(-++)] \
                    [set ${=}(+-+)] [set ${=}(+++)]] {
    upvar #0 ${=}(_${+}) ${+}
  }
  catch {unset _ = +}
}
}

}

;proc _em_cls_Sgetextension {} {
  em_extension lookup [file join [pwd] [string trimright [file dirname \
      [info script]] ./]]
}

;proc _em_cls_Sgetcmdpkg {name pe thisfile} {
  em_cmdpkg lookup @extension $pe @file $thisfile
}
