# FILE: account.tcl
#
# The server process for an OOMMF account.
#
# It listens on the port name by its first command line argument, and 
# directs communications among all OOMMF threads owned by this account.
#

if {([llength $argv] > 2) || ([llength $argv] == 0)} {
    error "usage: account.tcl <service_port> ?<creator_port>?"
}

global master Omf_export Omf_export_list Omf_export_mtime
set master(version) 0.0

# Set up for auto-loading
set master(relhdir) [string trimright [file dirname [info script]] ./]
set master(abshdir) [file join [pwd] $master(relhdir)]
set master(pdir) [file dirname $master(abshdir)]
set master(libdir) [file dirname $master(pdir)]
set master(root) [file dirname [file dirname $master(libdir)]]

if {[lsearch $auto_path $master(libdir)] < 0} {
    lappend auto_path $master(libdir)
}

# Support libraries
package require em 0
package require net 0

# Log info
set master(log) [em_extension ApplicationLog]

proc FatalError {msg} {
    global master
    $master(log) Append $msg
    if {$master(registered)} {
        Deregister
        return
    }
    exit
}

# Multiple connections share one protocol, and one set of globals
# Define protocol
set master(protocol) [net_protocol new \
        -name [list OOMMF account protocol $master(version)]]
$master(protocol) AddMessage start programs {} {
    global Omf_export_list
    if {[CheckExportTimes]} {
	SourceExportFiles  ;# Reread .omfExport.tcl files if necessary
    }
    foreach {id cmd} $Omf_export_list { lappend namelist $id }
    return [list start [concat 0 $namelist]]
}
$master(protocol) AddMessage start launch {program} {
    global Omf_export portmap
    if {[CheckExportTimes]} {
	SourceExportFiles  ;# Reread .omfExport.tcl files if necessary
    }
    if {[info exists Omf_export($program)]} {
        if {[catch {
                eval exec $Omf_export($program) < [em_extension NullFile] \
                    >>& [em_extension NullFile] &
                } threadid]} {
            return [list start [list 1 Launch failed: $threadid]]
        }
# I don't remember why I did this registration on launch.  With the
# new forking behavior of omfsh though, it causes double registration,
# so I've removed it.
#        array set portmap [list $threadid [list $program {}]]
        return [list start [list 0 $threadid]]
    }
    return [list start [list 1 Can't launch $program: no such program]]
}
$master(protocol) AddMessage start threads {} {
    global portmap master
    set ret {}
    foreach thread [array names portmap] {
        if {![string match {} [lindex $portmap($thread) 1]]} {
            lappend ret [list [lindex $portmap($thread) 0] $thread \
                    [lindex $portmap($thread) 1]]
        }
    }
$master(log) Append "Returning: $ret"
    return [list start [concat 0 $ret]]
}
$master(protocol) AddMessage start lookup {thread} {
    global portmap
    if {[info exists portmap($thread)]} {
        set port [lindex $portmap($thread) 1]
        if {![string match {} $port]} {
            return [list start [list 0 $port]]
        }
    }
    return [list start [list 1 $thread not registered]]
}
$master(protocol) AddMessage start register {thread alias port} {
    global portmap expire_event master
    if {[info exists portmap($thread)]} {
        set existingport [lindex $portmap($thread) 1]
        if {[string match {} $existingport]} {
            set ret [list start [list 0 [array set portmap [list \
                    $thread [lreplace $portmap($thread) 1 1 $port]]]]]
        } else {
            return [list start [list 1 \
                    $thread already registered at port $existingport]]
        }
    } else {
        set ret [list start [list 0 [array set portmap [list $thread \
                [list $alias $port]]]]]
    }
    if {[info exists expire_event] && [array size portmap]} {
        after cancel $expire_event
        unset expire_event
    }
    em_eventHandler Generate NewThread all
    $master(log) Append "portmap: [array get portmap]"
    return $ret
}
$master(protocol) AddMessage start deregister { thread port } {
    # Ought to call back for security -- and that will require
    # event-driven query-handling.
    global portmap expire_event master
    if {![info exists portmap($thread)]} {
        return [list start [list 1 $thread not registered]]
    }
    if {![string match $port [lindex $portmap($thread) 1]]} {
        return [list start [list 1 permission denied]]
    }
    set ret [list close [list 0 [unset portmap($thread)]]]
    if {![array size portmap]} {
        set expire_event [after 900000 [list FatalError \
                "OOMMF account server expired"]]
    }
    $master(log) Append "portmap: [array get portmap]"
    return $ret
}

$master(protocol) AddMessage start bye {} {
    return [list close "Bye!"]
}

# A rather ugly hack -- have all connections handle the NewThread event
# by sending out a newthread message.  Would be better to support
# this within the protocol somehow.
em_eventHandler new net_connection Ready AddNewThreadHandler
proc AddNewThreadHandler {} {
    global nthandler master
    foreach conn [net_connection instances] {
        if {![info exists nthandler($conn,nt)]} {
$master(log) Append "NewThread handler for connection $conn"
            set nthandler($conn,nt) [em_eventHandler new all NewThread \
                    [list $conn Tell newthread]]
            set nthandler($conn,del) [em_eventHandler new $conn delete \
                    [list DeleteNewThreadHandler $conn]]
        }
    }
}
proc DeleteNewThreadHandler { conn } {
    global nthandler master
    $nthandler($conn,nt) delete
    $nthandler($conn,del) delete
$master(log) Append "NewThread handler deleted for connection $conn"
    unset nthandler($conn,nt)
    unset nthandler($conn,del)
}

# Define and start server 
# add -allow and -deny options to control access
# We have to do a manual registration with host server
set master(registered) 0
set master(server) [net_server new -protocol $master(protocol) \
        -log $master(log) -register 0]
set serviceport [lindex $argv 0]
if {[catch {$master(server) Start $serviceport} master(msg)]} {
    FatalError "Can't start OOMMF account server on port $serviceport:
	$master(msg)."
}
set master(port) [$master(server) Port]

# Our account name
set master(account) [net_account DefaultAccountName]
# Our home directory
if {[info exists env(HOME)]} {
    set master(home) $env(HOME)
} else {
    set master(home) [pwd]
}

# Get export list -- read into global array
set master(exportFiles) {}
if {[file readable [file join $master(home) .omfExport.tcl]]} {
    lappend master(exportFiles) [file join $master(home) .omfExport.tcl]
}
if {[file readable [file join $master(root) .omfExport.tcl]]} {
    lappend master(exportFiles) [file join $master(root) .omfExport.tcl]
}

proc CheckExportTimes {} {
    # Check .omfExport file modification times against the 
    # modification time when last sourced.  Returns 1 if any
    # of the files have a different mtime, 0 otherwise
    global Omf_export_mtime
    foreach file [array names Omf_export_mtime] {
	if {$Omf_export_mtime($file) != [file mtime $file]} {
	    return 1
	}
    }
    return 0
}

proc SourceExportFiles {} {
    # Reads in .omfExport files
    global master Omf_export Omf_export_list Omf_export_mtime

    catch {unset Omf_export}
    catch {unset Omf_export_list}
    catch {unset  Omf_export_mtime}

    foreach file $master(exportFiles) {
	if {![info exists Omf_export] && ![info exists Omf_export_list]} {
	    $master(log) Append "Sourcing $file ..."
	    if {[catch {uplevel #0 [list source $file]} msg]} {
		$master(log) Append "Error: $msg"
		error $msg
	    }
	}
	set Omf_export_mtime($file) [file mtime $file]
    }
    if {![info exists Omf_export] && ![info exists Omf_export_list]} {
	error "Neither Omf_export nor Omf_export_list\
		set in $master(exportFiles)"
    }
    if {![info exists Omf_export]} {
	# Fill Omf_export array from Omf_export_list
	foreach {id action} $Omf_export_list {
	    set Omf_export($id) $action
	}
    }
    if {![info exists Omf_export_list]} {
	# Construct Omf_export_list from Omf_export array
	foreach id [array names Omf_export] {
	    lappend Omf_export_list $id $Omf_export($id)
	}
    }
    $master(log) Append "Exporting: $Omf_export_list"
}

SourceExportFiles

# The host we register with
set master(host) [net_host new -hostname localhost -log $master(log)]
em_eventHandler new $master(host) delete [list FatalError "host died"]
em_eventHandler new $master(host) Ready Register
proc Register {} {
    global master
    $master(log) Append "Sending registration message to host"
    set qid [$master(host) Send register $master(account) $master(port)]
    em_eventHandler new $master(host) Reply$qid RegisterReply
    em_eventHandler new $master(host) Timeout$qid RegisterTimeout
}
proc RegisterTimeout {} {
    FatalError "Timed out waiting to complete registration"
}
proc RegisterReply {} {
    global master
    set reply [$master(host) Get]
    switch -- [lindex $reply 0] {
        0 {
            CallbackCreator
        }
        default {
            FatalError "Registration error: [lrange $reply 1 end]"
        }
    }
}
proc CallbackCreator {} {
    global argv master
    set master(registered) 1
    if {[llength $argv] == 2} {
        # Inform my creator that my server is running.
        set port [lindex $argv 1]
        if {[catch {socket localhost $port} s]} {
            FatalError "Unable to call back $port: $s"
        }
        $master(log) Append "Called back $port"
        close $s
    }
}
proc Deregister {} {
    global master
    $master(log) Append "Sending deregistration message to host"
    if {[catch {
            $master(host) Send deregister $master(account) $master(port)
            } qid]} {
        # Can't send message -- host is probably dead.
        $master(log) Append "Dying without deregistering!"
        exit
    }
    em_eventHandler new $master(host) Reply$qid DeregisterReply
    em_eventHandler new $master(host) Timeout$qid DeregisterTimeout
}
proc DeregisterReply {} {
    global master
    set reply [$master(host) Get]
    switch -- [lindex $reply 0] {
        0 {
            $master(log) Append "Deregistered.  Exiting clean."
            exit
        }
        default {
            $master(log) Append \
                "Deregistration error: [lrange $reply 1 end]
	Exiting dirty."
            exit
        }
    }
}
proc DeregisterTimeout {} {
    $master(log) Append "Deregistration timeout.  Exiting dirty."
    exit
}
set expire_event [after 900000 [list FatalError "OOMMF account server expired"]]

vwait master(forever)

