#!/bin/sh
# FILE: mmGraph.tcl
#
# The control widget and network wrapping for an instance of the
# omf_graphwin class.
#
# This comment continues on in Tcl, but not in sh \
exec wish "$0" ${1+"$@"}

set oommf(progname) "mmGraph"

# App global variables:
#    _mmgraph(ptbufsize) = Maximum number of data lines stored in graph widget

# This code can be source'd from inside another script by setting
# no_net and no_event_loop as desired before sourcing this file.
# If no_net is set true, use the proc "AddTriples { triples }" to add
# data.

if {![info exists no_net]} {
    if {[string match "--nonet" [lindex $argv 0]]} {
	set no_net 1
    } else {
	set no_net 0      ;# Default is use net
    }
}

if {![info exists no_event_loop]} {
    set no_event_loop 0   ;# Default is to enter the event loop
}

# Set up for auto-loading
set master(relhdir) [string trimright [file dirname [info script]] ./]
set master(abshdir) [file join [pwd] $master(relhdir)]
set master(pdir) [file dirname $master(abshdir)]
set master(libdir) [file dirname $master(pdir)]
if {[lsearch $auto_path $master(libdir)] < 0} {
    lappend auto_path $master(libdir)
}

# Support libraries
package require Tk
package require em
package require oommf

# Death
proc Die { win } {
    if {![string match "." $win]} { return }  ;# Child destroy event
    global protocol
    if {[info exists protocol]} {
	$protocol delete; after 2000 exit
    } else {
	exit
    }
}
bind . <Destroy> {+Die %W}

# Save data routines

proc SaveFile { filename } {
    global savefilename
    set numfmt "%-15.8g"   ;# For table body
    set strfmt "%-15s"      ;# For column headers
    if {[catch {open $filename "a+"} fileid]} {
	global oommf
	OMF_NonFatalError \
		"$oommf(progname): Unable to open $filename for appending"
	return
    }
    global menu_select data_value data_index data_unit

    # Extract output table column labels, as selected by displayed graphs
    set column_list {}
    set column_index {}
    if {[string match {} $menu_select(x)]} {
	return ;# No abscissa selected
    }
    lappend column_list $menu_select(x)
    lappend column_index $data_index($menu_select(x))
    foreach ordinate $data_index(all) {
	if {[info exists menu_select(y,$ordinate)] && \
		$menu_select(y,$ordinate) } {
	    # Label $ordinate is turned on in y-axis menu
	    lappend column_list $ordinate
	    lappend column_index $data_index($ordinate)
	}
    }

    # Build column headers
    set column_labels {}
    foreach col $column_list {
	set unit $data_unit($col)
	if {[string match {} $unit]} {
	    lappend column_labels "$col"
	} else {
	    lappend column_labels "$col ($unit)"
	}
    }
    # Compress and strip newlines
    regsub -all "\[ \r\t\n\]" $column_labels { } column_header

    # Write column header
    puts -nonewline $fileid "#"
    foreach h $column_header {
	puts -nonewline $fileid "[format $strfmt [list $h]] "
    }
    puts $fileid {} ;# newline

    # Write out all rows of table
    for {set i $data_value(count,start)} {$i<$data_value(count,end)} {incr i} {
	set record $data_value($i)
	foreach col $column_index {
	    puts -nonewline $fileid \
		    " [format $numfmt [lindex $record $col]]"
	}
	puts $fileid {}   ;# newline
    }
    close $fileid
    set savefilename $filename
}

proc SaveFileCallback { widget } {
    SaveFile [$widget GetFilename]
    return {}
}

proc DialogCallback { widget actionid args } {
    # Re-enable menubutton when the dialog widget is destroyed
    if {[string match DELETE $actionid]} {
        eval [join $args]
	return
    }

    # Call appropriate action proc
    set errmsg "ERROR (proc DialogCallback): Invalid actionid: $actionid"
    if {[string match SAVE $actionid]} {
	set errmsg [SaveFileCallback $widget]
    }

    return $errmsg
}

proc LaunchDialog { menuname itemlabel action } {
    if { [string match "disabled" [$menuname entrycget $itemlabel -state]] } {
	return   ;# Item not enabled
    }
    global oommf
    if {[string match SAVE $action]} {
	set dialogbox [omf_mmpesavedlg new -callback DialogCallback \
		-dialog_title "$oommf(progname):Save File" \
		-selection_title "Append ODT Output To..." \
		-filter "*.odt" \
		-delete_callback_arg \
		[list $menuname entryconfigure $itemlabel -state normal]]
    } else {
	OMF_NonFatalError "Invalid action code: $action" \
		"Error in LaunchDialog proc"
	return
    }

    # Set icon
    OMF_SetIcon [$dialogbox cget -winpath]

    # Disable relevant menubutton
    $menuname entryconfigure $itemlabel -state disabled
}

# Menu bar
set menubar [frame .mb -borderwidth 2 -relief raised]

set file_select [menubutton $menubar.file -text "File" \
	-menu $menubar.file.menu -underline 0]
set filemenu [menu $file_select.menu -tearoff 0]
$filemenu add command -label "Save" -underline 0 \
	-command { SaveFile $savefilename } -state disabled
proc MenuFileSaveSetState { var elt op } {
    global filemenu
    if {[string match "u" $op]} {
	$filemenu entryconfigure "Save" -state disabled
    } else {
	$filemenu entryconfigure "Save" -state normal
    }
}
trace variable savefilename wu MenuFileSaveSetState

$filemenu add command -label "Save As..." -underline 5 \
	-command { LaunchDialog $filemenu "Save As..." "SAVE" }
$filemenu add command -label "Clear data" -underline 0 \
	-command {ClearDataInteractive}
$filemenu add separator
$filemenu add command -label "Exit" -command { Die . } -underline 1

set xaxis_select [menubutton $menubar.xaxis -text "X-Axis" \
	-menu $menubar.xaxis.menu -underline 0]
set xmenu [menu $xaxis_select.menu -tearoff 1]

set yaxis_select [menubutton $menubar.yaxis -text "Y-Axis" \
	-menu $menubar.yaxis.menu -underline 0]
set ymenu [menu $yaxis_select.menu -tearoff 1]

set options_select [menubutton $menubar.options -text "Options" \
	-menu $menubar.options.menu -underline 0]
set optionsmenu [menu $options_select.menu -tearoff 0]
set showkey 1
$optionsmenu add checkbutton -label "Key" -underline 0 \
		-variable showkey -offvalue 0 -onvalue 1 \
		-command {$graph SetKeyState $showkey}
$optionsmenu add command -label "Point Buffer" -underline 0 \
	-command {SetBufferSizeInteractive} 
$optionsmenu add command -label "Rescale" -underline 0 \
	-command {$graph Rescale} 

pack $file_select $xaxis_select $yaxis_select $options_select -side left
pack $menubar -side top -fill x

# Buffer quota code
set _mmgraph(ptbufsize) 0  ;# Default value 0 => no limit

proc PtBufOverflowCheck {} {
    global _mmgraph data_value
    if {$_mmgraph(ptbufsize)<1} { return } ;# No limit
    set bufstart $data_value(count,start)
    set bufend $data_value(count,end)
    set bufoverflow [expr $bufend-$bufstart-$_mmgraph(ptbufsize)]
    if {$bufoverflow<1} { return }  ;# No overflow
    if {$bufoverflow==1} {
	# Normal overflow case (single record)
	unset data_value($bufstart)
	incr data_value(count,start)
    } else {
	# Large overflow
	set newbufstart [expr $bufstart+$bufoverflow]
	for {set i $bufstart} {$i<$newbufstart} {incr i} {
	    unset data_value($i)
	}
	set data_value(count,start) $newbufstart
    }
}

proc SetBufferSize { newsize } {
    global _mmgraph
    if {[catch {expr round($newsize)} newsize] || $newsize<1} {
	set newsize 0
    }
    set _mmgraph(ptbufsize) $newsize
    PtBufOverflowCheck
}

proc SetBufferSizeInteractive {} {
    global oommf _mmgraph
    set result [OMF_PromptDialog  "$oommf(progname): Point Buffer Size" \
	    "Enter maximum number of data records to store\n in the graph\
	    widget backend (0 => no limit):" $_mmgraph(ptbufsize)]
    if {[lindex $result 1]==0} {
	SetBufferSize [lindex $result 0]
    }
}


# Proc to add new labels to the x & y axis menus.  It is not currently
# possible to remove a label once added.  A global array menu_select
# stores the state of the axis menu select boxes:
#    menu_select(x)        = name of selected label, or {} if none
#    menu_select(y,$label) = 1 if box associated with label is selected,
#                            0 otherwise.
set menu_select(x) {}
proc AppendAxesSelect { new_labels } {
    global xmenu ymenu menu_select
    # Add new entries
    foreach label $new_labels {
	set menu_select(y,$label) 0
	regsub -all "\[ \r\t\n\]" $label { } compressed_label
	$xmenu add radiobutton -label $compressed_label \
		-variable menu_select(x) -value $label  \
		-command CreateGraph
	$ymenu add checkbutton -label $compressed_label \
		-variable menu_select(y,$label) \
		-offvalue 0 -onvalue 1 \
		-command "ChangeCurve \{$label\}"
    }
}

# Graph
set graph_frame [frame .gf -borderwidth 0]
set graph [omf_graphwin new $graph_frame]
$graph configure -autodraw 1 -showkey $showkey
pack [$graph cget -winpath] -side top -fill both -expand 1
pack $graph_frame -side top -fill both -expand 1
em_eventHandler new $graph delete { after 2000 exit }

# Proc to update graph with latest data input triples
proc UpdateGraph { index } {
    global menu_select data_value data_index graph

    if {[catch {set data_value($index)} record]} { return } ;# Out-of-bounds

    set abscissa $menu_select(x)
    if {[string match {} $abscissa]} {return} ;# No abscissa selected
    set xindex $data_index($abscissa)
    set xvalue [lindex $record $xindex]
    if {[string match {} $xvalue]} {return} ;# No abscissa value at this index

    foreach ordinate $data_index(all) {
	if {!$menu_select(y,$ordinate)} { continue }  ;# Skip curve
	set yindex $data_index($ordinate)
	if {![string match {} [set yvalue [lindex $record $yindex]]]} {
	    $graph AddPoints $ordinate $xvalue $yvalue
	}
    }
    return
}

# Proc to add or delete a curve from the graph, as determined by the
# value of menu_select(y,$ordinate) (1=>display, 0=>don't display),
# or as forced by the value of the import "display".
proc ChangeCurve { ordinate {display {}}} {
    global menu_select data_value data_index data_unit graph
    $graph DeleteCurve $ordinate ;# Remove any previously curve by this name

    if {![string match {} $display]} {
	set menu_select(y,$ordinate) $display
    } elseif {![info exists menu_select(y,$ordinate)]} {
	return  ;# Safety
    }
    if {!$menu_select(y,$ordinate)} {
	# Curve turned off.  If this is last curve, delete y-label.
	set no_curves 1
	foreach curve $data_index(all) {
	    if {$menu_select(y,$curve)} {
		set no_curves 0
		break
	    }
	}
	if {$no_curves} {
	    $graph SetYLabel {}
	    $graph DrawAxes
	}
	return
    }
    # Otherwise, turn curve on...
    if [string match {} $menu_select(x)] { return 0 } ;# No abscissa selected
    set abscissa $menu_select(x)
    # If y-label is not yet set, build one from current curve.
    set ylabel [lindex [$graph GetLabels] 1]
    if {[string match {} $ylabel]} {
	set yunit $data_unit($ordinate)
	if {[string match {} $yunit]} { set yunit "Non Dim" }
	$graph SetYLabel "$yunit"
	$graph DrawAxes
    }
    # Write curve
    set data_list {}
    $graph NewCurve $ordinate
    set xindex $data_index($abscissa)
    set yindex $data_index($ordinate)
    for {set i $data_value(count,start)} {$i<$data_value(count,end)} {incr i} {
	set record $data_value($i)
	if {[string match {} [set xvalue [lindex $record $xindex]]] || \
		[string match {} [set yvalue [lindex $record $yindex]]]} {
	    continue
	}
	append data_list " $xvalue $yvalue"
    }
    eval {$graph AddPoints $ordinate} $data_list
    return
}

# Proc to create a complete new graph, erasing any previously existing
# graph.  Return value is the number of curves in new graph.
proc CreateGraph {} {
    global menu_select data_value data_index data_unit graph
    $graph ResetGraph    ;# Remove any previously existing graph
    if [string match {} $menu_select(x)] { return 0 } ;# No abscissa selected
    set abscissa $menu_select(x)
    set xunit " ($data_unit($abscissa))"
    if {[string match " ()" $xunit]} { set xunit {} }
    $graph SetXLabel "$abscissa$xunit"
    set curve_count 0
    foreach ordinate $data_index(all) {
	if {![info exists menu_select(y,$ordinate)]} { continue } ;# Safety
	if {! $menu_select(y,$ordinate) } { continue }        ;# Skip curve
	ChangeCurve $ordinate
	incr curve_count
    }
    $graph DrawAxes
    return $curve_count
}

# Data storage
#   Data is stored in two global arrays: data_value and data_index.
# The first is an array of lists indexed by import order, e.g.,
#                data_value(4)
# Each element is a list of values.  The data_index array provides a
# mapping from label names to position inside the data_value lists.
# This position is constant for each label across all data_value entries.
# A data_value list entry is {} if no data field corresponding to that
# entry was entered on that import record.  Also, some data_value() lists
# may be longer than others.  If data_index(label) points past the end
# of a data_value() list, then the corresponding value is {} i.e.,
# missing.
#   The data_value array also contains two elements,
# data_value(count,start) and data_value(count,end) which denote the first
# and last data_value record indices stored in the array, i.e.,
#      data_value(count,start) <= i < data_value(count,end)
# Similarly, the data_index(count) element contains the length of the
# longest data_value() list.  data_index(all) is an ordered list of
# the data labels.
#   There is currently no way to remove a label once it has been added.
#   A previous version of this code stored each data item in an array
# element (as opposed to grouping each record together in a list).  For
# 10 element data records, data import appears to be about 20% faster the
# old way, but storage costs are about 4 times higher (about 1K per record
# as opposed to about 250 bytes per record).  Moreover, it is probably
# possible to streamline the new code somewhat.
#
#   The return value for the import StoreDataTable routine is a list of all
# new labels (if any).  Each call to "StoreDataTable" represents 1 record.
# If a labels occurs more than once in an import record, only the last
# instance is stored.
#
set data_value(count,start) 0
set data_value(count,end) 0
set data_index(count) 0   ;# Current maximum length of data_value lists.
set data_index(all) {}
proc StoreDataTable { triples } {
    global data_value data_index data_unit
    set new_labels {}
    foreach trip $triples {
	set label [lindex $trip 0]
	set unit  [lindex $trip 1]
	set value [lindex $trip 2]
	if {[catch {set data_index($label)} index]} {
	    # New label
	    set index $data_index(count)
	    set data_index($label) $index
	    lappend data_index(all) $label
	    set data_unit($label) $unit
	    incr data_index(count)
	    lappend new_labels $label
	}
	set new_value($index) $value    ;# Store temporarily in an array
    }
    # Construct data list
    set new_list {}
    for {set index 0} {$index<$data_index(count)} {incr index} {
	if {[catch {lappend new_list $new_value($index)}]} {
	    lappend new_list {}      ;# Missing element
	}
    }
    set recno $data_value(count,end) ;  incr data_value(count,end)
    set data_value($recno) $new_list
    PtBufOverflowCheck
    UpdateGraph $recno
    return $new_labels
}

proc ClearData {} {
    global data_value graph
    unset data_value
    set data_value(count,start) 0
    set data_value(count,end) 0
    $graph ResetData
}

proc ClearDataInteractive {} {
    global oommf
    set response [OMF_ModalDialog  "$oommf(progname): Clear Data" warning \
	    "Are you *sure* want to erase all data in stored\
	    in this graph widget?" \
	    {} {} 1 "Yes" "No"]
    if { $response == 0 } { ClearData }
}

if {!$no_net} {
    # Protocol
    package require net
    set protocol [net_protocol new -name "OOMMF DataTable protocol 0.0"]
    $protocol AddMessage start HasGui {} {
	return [list start [list 0 0]]
    }
    $protocol AddMessage start DataTable { triples } {
	set new_labels [StoreDataTable $triples]
	if {![string match {} $new_labels]} {
	    # New labels added
	    AppendAxesSelect $new_labels
	}
    }

    set server [net_server new -protocol $protocol -alias mmGraph -log em_nop]
    $server Start 0
} else {
    proc AddTriples { triples } {
	set new_labels [StoreDataTable $triples]
	if {![string match {} $new_labels]} {
	    # New labels added
	    AppendAxesSelect $new_labels
	}
    }
}

wm title . mmGraph<[pid]>

# Enter event loop
if {!$no_event_loop} {
    vwait omf_forever
}
