# FILE: oommf.tcl
#
# This file is evaluated by interpreters on which the C function
# Oommf_Init is evaluated.  It performs many common initialization
# tasks for the applications of the OOMMF project.
#
# Last modified on: $Date: 1998/02/25 21:14:52 $
# Last modified by: $Author: dgp $ 
#
########################################################################
# Don't source this file more than once
########################################################################
if { [info exists oommf(_oommf_tcl)] } { return }
set oommf(_oommf_tcl) "oommf.tcl has been sourced"

########################################################################
# Check the versions of the Tcl/Tk packages loaded in the interpreter
########################################################################
# OOMMF applications need Tcl 7.5 or later
if {[catch {package require Tcl 8}]} {
    package require Tcl 7.5
}
# Check whether this application loads Tk ...
set oommf(tk) 0
foreach _ [info loaded ""] {
    if {[string match Tk [lindex $_ 1]]} {
        set oommf(tk) 1
        break
    }
}
# ... and if so, whether Tk 4.1 or later is loaded
if {$oommf(tk)} {
    if {[catch {package require Tk 8}]} {
        package require Tk 4.1
    }
}

########################################################################
# Set up the console and standard channels
########################################################################
# Verify that some configuration variables have values
if {![info exists DEBUGLVL]} { set DEBUGLVL 1 }
set oommf(platform) "unknown"
if {[info exists tcl_platform]} {
    set oommf(platform) [string tolower [string trim $tcl_platform(platform)]]
}

# Process command line: do we open a console window? 
set oommf(console) [expr {$DEBUGLVL && $oommf(tk)}]
set idx 0
foreach _ $argv {
    if {[string match -- $_]} {
        break
    }
    switch -exact -- $_ {
        -console {
            set oommf(console) 1
            set argv [lreplace $argv $idx $idx]
        }
        -noconsole {
            set oommf(console) 0
            set argv [lreplace $argv $idx $idx]
        }
        default {
            incr idx
        }
    }
}
unset idx

if {$oommf(console)} {
    # If console requested, try to start one
    set _ [file join [file dirname [info script]] console.tcl]
    if {[file readable $_]} {
        source $_
    } else {
        error "Can't create console: File not readable: $_"
    }
}

########################################################################
# Re-define some basic procs, which were already defined in 
# oommfInit.cc .  This lets us source this file from wish.
proc OMF_PackageName {} {return Oommf}
proc OMF_FileNameBase {} {string tolower [OMF_PackageName]}
proc OMF_NameOfExecutable {} {
  set tclanswer [info nameofexecutable]
  set savedir [pwd]
  cd [file dirname $tclanswer]
  set myanswer [pwd]
  cd $savedir
  return [file join $myanswer [file tail $tclanswer]]
}

# Make a list of locations to search for the OOMMF project's Tcl shell:
if {![info exists oommf(shell)]} {
    set oommf(shell) omfsh	;# Safety
}
if {![info exists oommf(shellStaticSearchPath)]} {
    set oommf(shellStaticSearchPath) {}	;# Safety
}
# 1. The same directory as the currently running executable.
set oommf(shellSearchPath) [file dirname [OMF_NameOfExecutable]]
# 2. ../../../oommf/bin/ relative to currently running executable
lappend oommf(shellSearchPath) [file join [file dirname [file dirname \
        [file dirname [OMF_NameOfExecutable]]]] oommf bin]
# 3. Also search the static search path
set oommf(shellSearchPath) [concat $oommf(shellSearchPath) \
        $oommf(shellStaticSearchPath)]

foreach dir $oommf(shellSearchPath) {
    if {[file executable [file join $dir $oommf(shell)]]} {
        set oommf(shell) [file join $dir $oommf(shell)]
        break
    }
}

# Default HTML Browser (assumes source directory structure)
# (also assumes [info script] returns absolute pathname)
set oommf(HTMLBrowser) [list $oommf(shell) -fork [file join [file dirname \
        [file dirname [file dirname [file dirname [info script]]]]] mmhelp \
        mmhelp.tcl]]

# Add the directory named by the OOMMF_PACKAGE_PATH environment variable
# to the path searched by the auto-loader.
#
# This allows users to replace Oommf's support packages with their own 
# replacement implementations.
catch {lappend auto_path $env([string toupper [OMF_PackageName]]_PACKAGE_PATH)}

# Add the directory containing default implementations Oommf's support
# packages to the path searched by the auto_loader.
if {[info exists omf_library] && \
	[lsearch -exact $auto_path $omf_library] < 0} {
    lappend auto_path $omf_library
}

# The default package unknown handlers included in versions of Tcl earlier
# than 8.0 have a great weakness. If there is an error in *any* pkgIndex.tcl
# file in the current search path of the package unknown command, that
# error will bring the application to a halt -- even if the error is in
# an index file provided by an entirely different application!  The default
# package unknown handler included in Tcl 8.0 is far more robust, and is
# compatible back to Tcl 7.5.  It is reproduced below and made the 
# package unknown handler for OOMMF applications. 
#
# Another benefit to this is the avoidance of the inconsistency between
# Tcl 7.5 and Tcl 7.6+ about what constitutes the search path for the
# package unknown handler.
#
# NOTE: Because this *replaces* the default package unknown handler,
# it is possible that a specially customized installation of Tcl in
# which a specialized package unkown handler has already been defined
# and is required to handled the customized environment may make that
# customization mechanism broken by this strategy.  Too bad.  OOMMF
# goes best with plain Tcl/Tk.
#
if {$tcl_version < 8} {
  set _omf_g(defaultPkgUnknown) [package unknown]

  proc _OMF_PkgUnknown {name version {exact {}}} {
    global auto_path tcl_platform env dir

    if ![info exists auto_path] {
	return
    }
    if {[info exists dir]} {
	set save_dir $dir
    }
    for {set i [expr [llength $auto_path] - 1]} {$i >= 0} {incr i -1} {
	# we can't use glob in safe interps, so enclose the following
	# in a catch statement
	catch {
	    foreach file [glob -nocomplain [file join [lindex $auto_path $i] \
		    * pkgIndex.tcl]] {
		set dir [file dirname $file]
		if [catch {source $file} msg] {
		    tclLog "error reading package index file $file: $msg"
		}
	    }
        }
	set dir [lindex $auto_path $i]
	set file [file join $dir pkgIndex.tcl]
	# safe interps usually don't have "file readable", nor stderr channel
	if {[interp issafe] || [file readable $file]} {
	    if {[catch {source $file} msg] && ![interp issafe]}  {
		tclLog "error reading package index file $file: $msg"
	    }
	}
	# On the Macintosh we also look in the resource fork 
	# of shared libraries
	# We can't use tclMacPkgSearch in safe interps because it uses glob
	if {(![interp issafe]) && ($tcl_platform(platform) == "macintosh")} {
	    set dir [lindex $auto_path $i]
	    tclMacPkgSearch $dir
		foreach x [glob -nocomplain [file join $dir *]] {
		    if [file isdirectory $x] {
			set dir $x
			tclMacPkgSearch $dir
		    }
		}
	}
    }
    if {[info exists save_dir]} {
	set dir $save_dir
    } else {
	unset dir
    }
  }
  package unknown _OMF_PkgUnknown
}

# Patch a bug MJD found in the Tk 4.2 distribution
if { [info exists tk_version] && ($tk_version == 4.2) } {
    # The distribution tkMenuFind source seems to be broken: it has
    # comments in an illegal position inside the 'switch' statement.
    # The following code is from Tk 4.2, but with the comments moved.
    proc tkMenuFind { w char } {
	global tkPriv
	set char [string tolower $char]
	
	foreach child [winfo child $w] {
	    switch [winfo class $child] {
		Menubutton {
		    set char2 [string index [$child cget -text]  \
			    [$child cget -underline]]
		    if {([string compare $char \
			    [string tolower $char2]] == 0) \
			    || ($char == "")} {
			if {[$child cget -state] != "disabled"} {
			    return $child
			}
		    }
		}
		default {
		    # The tag above used to be "Frame", but it was changed so
		    # that the code would work with Itcl 2.0, which apparently
		    # uses other classes of widgets to hold menubuttons.
		    set match [tkMenuFind $child $char]
		    if {$match != ""} {
			return $match
		    }
		}
	    }
	}
	return {}
    }
}

# Provide an approximation of the subcommands of the Tcl command 'file' which 
# were introduced in Tcl 7.6 to Unix platforms.  Some of the unix command
# behavior may differ from the Tcl file specification, especially when
# overwriting or deleting read-only files or directories.  If is probably
# safer in general to use the "-force" option for copy, delete and rename
# subcommands.
proc OMF_File {subcommand args} {
    global tcl_version tcl_platform
    if {$tcl_version < 7.6} {
        if {[string match unix $tcl_platform(platform)]} {
	    # If passed, change "-force" to "-f"
	    if {[string match "-force" [lindex $args 0]]} {
		set args [lreplace $args 0 0 "-f"]
	    }
            switch $subcommand {
                copy {
                    uplevel exec cp $args >@ stdout 2>@ stderr
                }
                delete {
                    uplevel exec rm $args >@ stdout 2>@ stderr
                }
                mkdir {
                    uplevel exec mkdir $args >@ stdout 2>@ stderr
                }
                rename {
                    uplevel exec mv $args >@ stdout 2>@ stderr
                }
		default {
		    error "OMF_File command request $subcommand not\
			    supported in Tcl $tcl_version"
		}
            }
        } else {
	    error "OOMMF software on non-Unix platforms requires\
		    Tcl 7.6 or later"
        }
    } else {
        uplevel file $subcommand $args
    }
}

#####################################################################
# Utility routines

###################################################################
### Routine to search for a file.  On success, the return value ###
### is usable path to that file.  The empty string is returned  ###
### on failure.                                                 ###
###################################################################
proc OMF_FindFile { filename {filetest {file readable}}} {
    # Set "filetest" to null string to disable test
    set maxlevel 10  ;# Don't go down more than this many subdirectories

    # Search for $filename in script and lower directories
    set level 0  ;  set dirlist [list [file dirname [info script]]]
    while { $level<$maxlevel && [llength $dirlist]>0 } {
	incr level
	set newdirlist {}
	foreach dir $dirlist {
	    # Search for file in directory $dir
	    set fullname [file join $dir $filename]
	    if {[file exists $fullname]} {
		if {[string match {} $filetest] || \
			[eval $filetest [list $fullname]]} {
		    return $fullname
		}
	    }
	    if {$level<$maxlevel} {
		# If not found, build list of subdirectories
		foreach subdir [glob -nocomplain -- [file join $dir "*"]] {
		    if [file isdirectory $subdir] {
			lappend newdirlist $subdir
		    }
		}
	    }
	}
	set dirlist $newdirlist
    }
    # If we get to here, then file was not found
    return {}
}


###################################################################
### Routine to changes an import pathname as necessary to       ###
### an absolute pathname, using [pwd].                          ###
###################################################################
proc OMF_MakeAbsolutePathname { filename } {
    set pathtype [file pathtype $filename]
    if {[string match "absolute" $pathtype]} {
	return $filename   ;# Already in absolute form
    }
    if {[string match "relative" $pathtype]} {
	return [file join [pwd] $filename]
    }
    # Otherwise assume volumerelative, and attach to current working volume
    set templist [file split $filename]
    set templist [lreplace $templist 0 0 [lindex [file split [pwd]] 0]]
    return [eval file join $templist]
}


###################################################################
### Routine that attempts to return a unique filename.  A fatal ###
### is raised on failure, otherwise the purported unique name   ###
### is returned.                                                ###
###################################################################
set oommfstatic(OMF_TempName_count) 0 ;# Static counter for OMF_TempName proc
set oommfstatic(OMF_TempName_maxcount) 1000 ;# Max no. of filenames to try
proc OMF_TempName { {basedir {}} {baseprefix {_}} {suffix {}}} {
    global oommf OMF_TempName_count OMF_TempName_maxcount
    global oommfstatic

    if {[string match {} $basedir]} {
	# Search for temporary file directory
	set basedir "."       ;# In case all else fails
	if { [string compare $oommf(platform) windows] == 0 } {
	    if {[info exists env(TEMP)] && [file writable $env(TEMP)]} {
		set basedir $env(TEMP)
	    } elseif {[info exists env(TMP)] && [file writable $env(TMP)]} {
		set basedir $env(TMP)
	    } elseif {[info exists env(TMPDIR)] && \
		    [file writable $env(TMPDIR)]} {
		set basedir $env(TMPDIR)
	    } elseif {[file isdirectory "C:/TEMP"] && \
		    [file writable "C:/TEMP"]} {
		set basedir "C:/TEMP"
	    } elseif {[file isdirectory "C:/"] && [file writable "C:/"]} {
		set basedir "C:/"
	    }
	} else { ;# Assume Unix
	    if {[info exists env(TMP)] && [file writable $env(TMP)]} {
		set basedir $env(TMP)
	    } elseif {[info exists env(TMPDIR)] && \
		    [file writable $env(TMPDIR)]} {
		set basedir $env(TMPDIR)
	    } elseif {[info exists env(TEMP)] && \
		    [file writable $env(TEMP)]} {
		set basedir $env(TEMP)
	    } elseif {[file isdirectory "/tmp"] && \
		    [file writable "/tmp"]} {
		set basedir "/tmp"
	    }
	}
    }

    set basename [file join $basedir "$baseprefix-[info hostname]-[pid]-"]
    set count_start $oommfstatic(OMF_TempName_count)
    while {1} {
	set workname "$basename$oommfstatic(OMF_TempName_count)$suffix"
	incr oommfstatic(OMF_TempName_count)
	if {$oommfstatic(OMF_TempName_count) \
		>= $oommfstatic(OMF_TempName_maxcount)} {
	    set oommfstatic(OMF_TempName_count) 0
	}
	if {![file exists $workname]} { break }
	if {$oommfstatic(OMF_TempName_count) == $count_start} {
	    OMF_FatalError "Unable to find unique temporary name of \
		    form $workname after $OMF_TempName_maxcount tries."
	}
    }
    return $workname
}

###############################################
### Routine to associate icon with a window ###
### Under Windows this is a nop.            ###
###############################################
set oommf(icon) [OMF_FindFile "omficon.xbm"]
set oommf(iconmask) [OMF_FindFile "omfmask.xbm"]
proc OMF_SetIcon { win {iconname {}} {iconmaskname {}} } {
    global oommf
    if { [string compare $oommf(platform) windows] == 0 || \
	    ![winfo exists $win] || !$oommf(tk)} {
	return
    }
    set win [winfo toplevel $win]
    if { "$win" == "." } {
	set iconwin ".root:iconwindow"
    } else {
	set iconwin "$win:iconwindow"
    }

    if {[string match {} $iconname]} {
	set iconname $oommf(icon)
    }
    if {[string match {} $iconmaskname]} {
	set iconmaskname $oommf(iconmask)
    }

    set iconname [OMF_FindFile $iconname]
    set iconmaskname [OMF_FindFile $iconmaskname]

    if {[string match {} $iconname] || [string match {} $iconmaskname]} {
	return         ;# Give up
    }

    # Setup plain icon
    wm iconbitmap $win @$iconname
    wm iconmask $win @$iconmaskname

    # Setup fancier icon (not all wm's support this)
    if [winfo exists $iconwin] {
	destroy $iconwin
	update idletasks
    }
    if {![winfo exists $iconwin]} {
	toplevel $iconwin
	label $iconwin.label
	pack $iconwin.label
	$iconwin.label configure -bitmap @$iconname -bg yellow -fg black
	if {[catch {wm iconwindow $win $iconwin}] || \
		[string compare $iconwin [wm iconwindow $win]]!=0 || \
		[update idletasks ; \
		expr [winfo viewable $win] && [winfo viewable $iconwin]] || \
		[string compare "icon" [wm state $iconwin]]!=0 } {
	    # wm iconwindow didn't take
	    catch {wm iconwindow $win {}}
	    destroy $iconwin
	} else {
	    # wm iconwindow *did* take.  Setup auto-destruct, taking care
	    # to ignore child window <Destroy> events
	    bind $win <Destroy> "+
	    if \[string compare $win %W\] continue
	    destroy $iconwin
	    "
	}
    }
    # Here's an alternate way, by loading the image from a file:
    #    set imagetoken [image create photo -file oommficon.gif]
    #    toplevel $iconwin
    #    label $iconwin.label -image $imagetoken
    #    pack $iconwin.label
    #    wm iconwindow $win $iconwin
    #    unset imagetoken
}

########################
### Call stack trace ###
########################
proc OMF_CallStack {} {
    set history {}
    for {set n [expr [info level]-1]} {$n>0} {incr n -1} {
        append history "LEVEL $n: [info level $n]\n\n"
    }
    return $history
}

#####################################################################
### Proc to position a child window relative to a parent.  The    ###
### imports xrat & yrat are the position of the upper lefthand    ###
### corner of the child relative to the ulh corner of the parent. ###
### Bug: if the parent geometry has negative position parameters, ###
### then the aforementioned corners will not be ulh.  Also, this  ###
### may be wm-specific.                                           ###
#####################################################################
proc OMF_PositionChild { child {parent .} {xrat .25} {yrat .25} } {
    set parentpos [wm geometry $parent]
    set junk ""
    regexp {^([0-9]*)x?([0-9]*)([-\+][0-9]+)([-\+][0-9]+)} \
            "$parentpos" junk parwidth parheight parposx parposy
    if { "$junk" != "" } then {
        if { "$parposx" >= 0 } then {
            set childposx [expr round($parposx + $xrat*$parwidth)]
        } else {
            set childposx [expr round($parposx - $xrat*$parwidth)]
        }
        # Note: Negative entries measure from lower & right window corner
        if { "$parposy" >= 0 } then {
            set childposy [expr round($parposy + $yrat*$parheight)]
        } else {
            set childposy [expr round($parposy - $yrat*$parheight)]
        }
        # We can write some fancier code (like centering) at a later date
        # -mjd 19-Jan-1997
        if { $childposx >= 0 } then { set childposx "+$childposx" }
        if { $childposy >= 0 } then { set childposy "+$childposy" }
        wm geometry $child "$childposx$childposy"
    }
}


##############################################
### Fatal error handler dialog box (modal) ###
##############################################
set oommf(error_dialog,instance) 0
set oommf(dialog,width) 12c
proc OMF_FatalError { msg {title {}} {exitcode 1} } {
    global oommf
    if {!$oommf(tk)} {
	puts stderr $msg ; flush stderr
	exit $exitcode
    }
    if {[string match {} "$title"]} {
        global ProgName
        if [info exists ProgName] { set title "$ProgName: " }
        append title "Fatal Error"
    }
    set instance $oommf(error_dialog,instance) ;# for errors inside errors.
    incr oommf(error_dialog,instance)
    set callstack [OMF_CallStack]
    set width $oommf(dialog,width)
    set window [toplevel .omfErrorDialog$instance]
    wm group $window .
    wm title $window "$title"
    OMF_PositionChild $window

    frame $window.top
    frame $window.bottom
    label $window.top.bitmap -bitmap error
    pack $window.top.bitmap -side left -anchor w -padx 2m
    message $window.top.msg -text "$msg"
    $window.top.msg configure -width $width
    pack $window.top.msg -side left -fill both -expand 1

    if {[info exists oommf(error_dialog,done$instance)]} {
        unset oommf(error_dialog,done$instance)  ;# Safety
    }
    set btn0 [button $window.bottom.btn0 \
            -text Die \
            -command "set oommf(error_dialog,done$instance) 1"]
    pack $btn0 -side left -expand 1
    set btn1 [button $window.bottom.btn1 \
            -text "Call Stack" \
            -command "OMF_ModalDialog {Call Stack} {} \
            \"$callstack\" {} {} 0"]
    pack $btn1 -side left -expand 1
    wm protocol $window WM_DELETE_WINDOW \
            "set oommf(error_dialog,done$instance) 2"
    pack $window.top $window.bottom -side top -fill both -expand 1

    update idletasks
    focus $window
    catch {tkwait visibility $window; grab $window}
    # Note: It appears dangerous to put a "grab" on a window
    #  before it is visible.  (Errant mouse clicks???)

    OMF_SetIcon $window
    catch {tkwait variable oommf(error_dialog,done$instance)}
    unset oommf(error_dialog,done$instance)
    grab release $window
    destroy $window
    exit $exitcode
}


###################################################
### Nonfatal warning message dialog box (modal) ###
###################################################
proc OMF_NonFatalError { msg {title {}} {exitcode 1} } {
    if {"$title"==""} {
        global ProgName
        if [info exists ProgName] { set title "$ProgName: " }
        append title "Nonfatal Error"
    }
    if {[OMF_ModalDialog $title warning $msg {} {} 0 \
	    "Continue" "Die"]==1} {
	exit $exitcode
    }
    return 0
}

################################
# Modal message dialog box
#
# Note: Tk built-in bitmaps include error, info, question,
#  questhead and warning.
set oommf(modal_dialog,instance) 0
proc OMF_ModalDialog { title bitmap message {width ""} {scrollbars 0} \
        {defaultbtn ""} args } {
    global oommf
    if {!$oommf(tk)} {
	puts stderr $message ; flush stderr
	if {[string match {} $defaultbtn]} {return 0}
	return $defaultbtn
    }
    set instance $oommf(modal_dialog,instance)
    incr oommf(modal_dialog,instance)
    if {"$width"==""} {set width $oommf(dialog,width)}
    if {"$scrollbars"==""} {set scrollbars 0}
    set window [toplevel .omfModalDialog$instance]
    wm group $window .
    wm title $window "$title"
    OMF_PositionChild $window

    frame $window.top    -bd 10
    frame $window.bottom -bd 10
    if { "$bitmap" != "" } then {
        label $window.top.bitmap -bitmap $bitmap
        pack $window.top.bitmap -side left -anchor w -padx 2m
    }
    message $window.top.msg -text "$message"
    $window.top.msg configure -width $width
    pack $window.top.msg -side left -fill both -expand 1
    set btncount 0
    if {[info exists oommf(modal_dialog,btncode$instance)]} {
        unset oommf(modal_dialog,btncode$instance)  ;# Safety
    }
    foreach btnlabel $args {
        set btn($btncount) [button $window.bottom.btn$btncount \
                -text $btnlabel \
                -command "set oommf(modal_dialog,btncode$instance) $btncount"]
	bind $btn($btncount) <Key-Return> "$btn($btncount) invoke"
        pack $btn($btncount) -side left -expand 1 -padx 5
        incr btncount 1
    }
    if {$btncount==0} {
        set btn($btncount) [button $window.bottom.btn$btncount -text "OK" \
                -command "set oommf(modal_dialog,btncode$instance) $btncount"]
        pack $btn($btncount) -side left -expand 1 -padx 5
        incr btncount 1
    }
    wm protocol $window WM_DELETE_WINDOW \
            "set oommf(modal_dialog,btncode$instance) 0"
    pack $window.top $window.bottom -side top -fill both -expand 1

    if {"$defaultbtn"!="" && $defaultbtn>=0 && $defaultbtn<$btncount} {
        # Code to make a "default" frame around default button.
        set fout [frame $window.bottom.defaultouter -bd 2 -relief sunken]
        lower $fout
        pack $fout -before $btn($defaultbtn) \
                -side left -expand 1 -padx 5
        pack forget $btn($defaultbtn)
        pack $btn($defaultbtn) -in $fout -padx 4 -pady 4
        # Give this button the keyboard focus
        focus $btn($defaultbtn)
    }

    update idletasks
    focus $window
    catch {tkwait visibility $window; grab $window }
    # Note: It appears dangerous to put a "grab" on a window
    #  before it is visible.  (Errant mouse clicks???)
    # Also, the sequence (including order) of
    #  'update idletasks ; focus $window ; catch ...'
    # seems to protect against binding reentrancy from the
    # user clicking too many times, or hitting keys too fast.

    OMF_SetIcon $window
    catch {tkwait variable oommf(modal_dialog,btncode$instance) }
    set returncode $oommf(modal_dialog,btncode$instance)
    unset oommf(modal_dialog,btncode$instance)
    grab release $window
    destroy $window
    return $returncode
}

################################
# Modal propmt dialog box 
#   Returns a 2 element list.  The first element is the return
# string, the second element is either 0 or 1; 0 indicates
# an "OK" return, 1 implies a "Cancel" return.  In the latter
# case the "defaultvalue" will always be returned as the first
# element
#
proc OMF_PromptDialog { title message {defaultvalue {}} \
	{width {}} {checkscript {}} } {
    # "checkscript" not yet implemented
    global oommf _oommf
    if {!$oommf(tk)} { return [list $defaultvalue 1] }
    if {[string match {} $width]} {set width $oommf(dialog,width)}
    set window [toplevel .omfPromptDialog]
    wm group $window .
    wm title $window "$title"
    OMF_PositionChild $window

    frame $window.top    -bd 10
    frame $window.bottom -bd 10
    set msg [message $window.top.msg -text $message -width $width]
    set value [entry $window.top.entry]
    $value insert 0 $defaultvalue

    pack $msg -side top -fill both -expand 1
    pack $value -side top -fill x -expand 1

    set btnOK [button $window.bottom.btnOK -text "OK" \
	    -command "set _oommf(prompt_dialog,btn) 0"]
    set btnCancel [button $window.bottom.btnCancel -text "Cancel" \
	    -command "set _oommf(prompt_dialog,btn) 1"]
    pack $btnOK $btnCancel -side left -expand 1 -padx 5

    wm protocol $window WM_DELETE_WINDOW \
            "set _oommf(prompt_dialog,btn) 0"

    pack $window.top $window.bottom -side top -fill both -expand 1

    # Extra bindings
    bind $value <Key-Return> "$btnOK invoke"
    bind $value <Key-Escape> \
	    "$value delete 0 end ; $value insert 0 \{$defaultvalue\}"
    bind $btnOK <Key-Return> "$btnOK invoke"
    bind $btnCancel <Key-Return> "$btnCancel invoke"

    $value selection range 0 end
    focus $value

    update idletasks
    focus $window
    catch {tkwait visibility $window; grab $window }
    # Note: It appears dangerous to put a "grab" on a window
    #  before it is visible.  (Errant mouse clicks???)
    # Also, the sequence (including order) of
    #  'update idletasks ; focus $window ; catch ...'
    # seems to protect against binding reentrancy from the
    # user clicking too many times, or hitting keys too fast.

    OMF_SetIcon $window
    catch {tkwait variable _oommf(prompt_dialog,btn) }
    if {$_oommf(prompt_dialog,btn)==0} {
	# OK invoked
	set result [list [$value get] 0]
    } else {
	# Cancel invoked
	set result [list $defaultvalue 1]
    }
    unset _oommf(prompt_dialog,btn)

    grab release $window
    destroy $window
    return $result
}

# Add parent directory to auto_path, if not already there
set __libdir [file join [pwd] \
	[string trimright [file dirname [file dirname [info script]]] ./]]
if {[lsearch -exact $auto_path $__libdir]<0} {
    lappend auto_path $__libdir
}
unset __libdir

# Now that the package command conforms to Tcl 8.0+ behavior, use it.
package require em
package require oommf

proc OMF_LaunchHTMLBrowser {url} {
    global oommf
    eval exec $oommf(HTMLBrowser) $url &
}

# Have background error handler log the error msg *and* notify user.
proc bgerror { msg } {
    [em_extension ApplicationLog] Append $msg
    OMF_NonFatalError $msg 
}

########################################################################
# If the application has a Tk main window, set up an icon (background)
########################################################################
if {$oommf(tk)} {
    after idle OMF_SetIcon .
}

